#!/usr/bin/perl

use strict;

# uncomment one env only
my $envType = "datasetEnv";

# dataset env
my $dataSetName = "/Users/vahdat/ESC/data/vahdat/50d4c/EM/50d4c.dlm";
my $labelsProvided = 1;
#my $labelSetName = "../data/ARCENE/arcene_label.txt";
my $dataSetDim = 50;
my $dataSetSize = 1289;

my $attrWiseClusCentFileName   = "/Users/vahdat/ESC/data/vahdat/50d4c/EM/50d4c-AttributeWiseCentroids";
my $attrWiseClusLabelsFileName = "/Users/vahdat/ESC/data/vahdat/50d4c/EM/50d4c-AttributeWiseLabels";
my $noneligibleAttrFileName	= "/Users/vahdat/ESC/data/vahdat/50d4c/EM/50d4c-NoneligibleAttributes";
my $noneligibleClusFileName	= "/Users/vahdat/ESC/data/vahdat/50d4c/EM/50d4c-Noneligible1DCentroids";

my $numLevels = 2;
my $numObjectives0 = 2;
my $numObjectives1 = 2;

my $doDistortion = 1;
my $doConnectivity = 1;
my $doAttrCount = 0;
my $doClusterCount = 0;

# Point population parameters
my $doSubsampling = 1;
my $pointPopSize = 10;

# Host (team, clustering solution) level parameters
my $hostPopSize = 10;
my $minSymbionts = 2;
my $maxSymbionts = 5;

# Host level variation probabilities
# if a fixed mutation rate is desired, set both values equal. 
my $pc_intiial = 0.1;
my $pc_final = 1.0;
# if a fixed mutation rate is desired, set both values equal. 
my $pm_intiial = 0.0;
my $pm_final = 1.0;

my $psd = 0.1;
my $psa = 0.1;
my $pss = 0.1;
my $pmlm = 1.0; # Not used

# Symbiont (cluster) level parameters
my $minAttributes = 2;
my $maxAttributes = 10;

# Symbiont level variation probabilities
my $pad = 0.1;
my $paa = 0.1;
my $pas = 0.1;
my $plc = 0.1;

# number of generations
my $numGenerations0 = 0;
my $numGenerations1 = 5;

# esc params
my $exe = "esc";
my $useAttrProbLrng = 0; # do we use dynamic probabilities for attributes?
my $attrProbLrngRate = 0.001; # if yes what's the learning rate?
my $seed = 1; # first seed number, last seed number is the input from command line 
my $statMod = 100;

my $maxAttempts = 20;

my $epsilon = 0.001;
my $infinity = 1000000.00;

# 0 to use all the population (recommended in case first level doesn't have a MOEA), 
# 1 to use only the unique inds of the PF, and 2 to use all inds of the PF (recommanded if first level uses a MOEA).
my $symbsToUse = 0;


if(scalar(@ARGV) != 3)
{
	die "usage: createInputs.pl prefix numRuns perBatch";
}

my $prefix = $ARGV[0];
my $numRuns = $ARGV[1];
my $perBatch = $ARGV[2];

print "prefix: $prefix\n";
print "numRuns: $numRuns\n";
print "perBatch: $perBatch\n";

my $run;
my $argFile;
my $nextSeed;

open(RUN, ">$prefix" . ".run") || die "cannot open RUN $prefix.run";

for($run = 0; $run < $numRuns; $run++)
{
	$nextSeed = $seed+ $run;

	$argFile = "$prefix.$nextSeed.arg";
	
	open(ARG, ">$argFile") || die "cannot open $argFile";

	#print ARG "envType $envType\n";
	#print ARG "\n";

	print ARG "dataSetName $dataSetName\n";
	print ARG "labelsProvided $labelsProvided\n";
	#print ARG "labelSetName $labelSetName\n";
	print ARG "dataSetDim $dataSetDim\n";
	print ARG "dataSetSize $dataSetSize\n";
	print ARG "\n";

	print ARG "attrWiseClusCentFileName $attrWiseClusCentFileName\n";
	print ARG "attrWiseClusLabelsFileName $attrWiseClusLabelsFileName\n";
	print ARG "noneligibleAttrFileName $noneligibleAttrFileName\n";
	print ARG "noneligibleClusFileName $noneligibleClusFileName\n";
	print ARG "\n";

	print ARG "numLevels $numLevels\n";
	print ARG "numObjectives0 $numObjectives0\n";
	print ARG "numObjectives1 $numObjectives1\n";
	print ARG "\n";
	
	print ARG "doDistortion $doDistortion\n";
	print ARG "doConnectivity $doConnectivity\n";
	print ARG "doAttrCount $doAttrCount\n";
	print ARG "doClusterCount $doClusterCount\n";
	print ARG "\n";
	
	print ARG "doSubsampling $doSubsampling\n";
	print ARG "pointPopSize $pointPopSize\n";
	print ARG "\n";
	
	print ARG "hostPopSize $hostPopSize\n";
	print ARG "minSymbionts $minSymbionts\n";
	print ARG "maxSymbionts $maxSymbionts\n";
	
	print ARG "pc_intiial $pc_intiial\n";
	print ARG "pc_final $pc_final\n";
	print ARG "pm_intiial $pm_intiial\n";
	print ARG "pm_final $pm_final\n";
	
	print ARG "psd $psd\n";
	print ARG "psa $psa\n";
	print ARG "pss $pss\n";
	print ARG "pmlm $pmlm\n";
	print ARG "\n";
	
	print ARG "minAttributes $minAttributes\n";
	print ARG "maxAttributes $maxAttributes\n";
	
	print ARG "pad $pad\n";
	print ARG "paa $paa\n";
	print ARG "pas $pas\n";
	print ARG "plc $plc\n";
	print ARG "\n";
	
	print ARG "numGenerations0 $numGenerations0\n";
	print ARG "numGenerations1 $numGenerations1\n";
	print ARG "\n";
	
	print ARG "useAttrProbLrng $useAttrProbLrng\n";
	print ARG "attrProbLrngRate $attrProbLrngRate\n";
	print ARG "seed $nextSeed\n";
	print ARG "statMod $statMod\n";
	print ARG "\n";

	print ARG "epsilon $epsilon\n";
	print ARG "infinity $infinity\n";
	print ARG "maxAttempts $maxAttempts\n";
	print ARG "\n";
	
	print ARG "symbsToUse $symbsToUse\n";
	print ARG "\n";
	
	close(ARG);
	
	if($nextSeed % $perBatch != 0)
	{
		print RUN "./$exe $argFile 1> $prefix.$nextSeed.std 2> $prefix.$nextSeed.err &\n";
	}
	else
	{
		print RUN "./$exe $argFile 1> $prefix.$nextSeed.std 2> $prefix.$nextSeed.err\n";
	}
}

close(RUN);
