
//#include "esc.hpp"
#include "types.hpp"
#include "sorting.hpp"
#include "splitting.hpp"

#include <algorithm>
#include <vector>
#include <utility>
#include <map>
#include <iterator>
#include <limits.h>
#include <float.h>

#if SORTING==FASTSORTING

/* #define DOCHECKING */

extern esc *model;
void checksplit(vector<solution *> *L,vector<solution *> *H, unsigned int M, 
				unsigned int indexL, unsigned int indexH, char mess[]);

#define PARTITIONSPLITTING 1
#define SORTSPLITTING 2

/* toggles the splitting method in fast sorting */
#define SPLITTING PARTITIONSPLITTING

using std::map;
using std::pair;

/***********************************************************************************************************/

bool cmp_2(const solution *s1, const solution *s2)
{
	return (s1->twoDcmp(*s2));
}

/***********************************************************************************************************/

bool mapcomp(pair<double, unsigned int> *p, double key)
{
	return(p->first < key);
}

/***********************************************************************************************************/

bool solution::twoDcmp(const solution& s2) const
{
	return(objectives[0]<s2.objectives[0] || (objectives[0]==s2.objectives[0] && objectives[1]<s2.objectives[1]));
}

/***********************************************************************************************************/

void sortOnFirstObjective(vector<solution *> &V)
{
	sort(V.begin(),V.end(),cmp_2);
}

/***********************************************************************************************************/

void population::twoDsort()
{
	sortedfronts=vector<front>();
	
	sortOnFirstObjective(popsorted);
	
	int numfronts=1;
	
	sortedfronts.push_back(front(popsorted[0],numfronts));
	
	vector<double> T;
	T.push_back((*popsorted[0]).getObj(1));
	
	vector<double> Tx;
	Tx.push_back((*popsorted[0]).getObj(0));
	
	for(unsigned int i=1;i<doublepopsize;i++) {
		
		if( (*popsorted[i]).getObj(1)<T[numfronts-1] || ( (*popsorted[i]).getObj(0)==Tx[numfronts-1] && (*popsorted[i]).getObj(1)==T[numfronts-1]) ) {
			// solution belongs in an existing front.
			
			// find the right front
			vector<double>::iterator f=lower_bound(T.begin(),T.end(),(*popsorted[i]).getObj(1));
			
			// if the two y-values are identical, maybe it's the next front
			while( (*popsorted[i]).getObj(1)==(*f) && 
				  Tx[f-T.begin()]<(*popsorted[i]).getObj(0) ) f++;
			
			// insert popsorted[i] in front *f
			sortedfronts[f-T.begin()].push_back(popsorted[i]);
			
			*f=(*popsorted[i]).getObj(1);
			Tx[f-T.begin()]=(*popsorted[i]).getObj(0);
			T[f-T.begin()]=(*popsorted[i]).getObj(1); 
			
		} else {
			// new front must be created
			numfronts++;
			sortedfronts.push_back(front(popsorted[i],numfronts-1));
			
			// update T and M
			T.push_back((*popsorted[i]).getObj(1));
			Tx.push_back((*popsorted[i]).getObj(0));
		}
	}
}

/***********************************************************************************************************/

void population::sortOnFronts()
{
	if (model->_numObjectives > 2)
	{
		vector<solution *> S;
		
		S.reserve(doublepopsize);
		for(vector<solution>::iterator i=pop.begin(); i != pop.end(); i++)
		{
			i->setFrontNumber(1);
			//FN START - 10/24/2006 - Added the following 2 lines and commented out the 3rd line to fix a debug error
			solution *h = &*i;
			S.push_back(h);
			//S.push_back(i);
			//FN END 
		}
		
		ND_helper_A(S, model->_numObjectives, true);
		
		sortedfronts = vector<front>();
		
		for(unsigned int i=0; i < doublepopsize; i++)
		{
			while((int)sortedfronts.size() < pop[i].getFrontNumber())
				sortedfronts.push_back(front(sortedfronts.size()+1));
			
			sortedfronts[pop[i].getFrontNumber()-1].push_back(&pop[i]);
		}
	}
	else
	{
		twoDsort();
	}
	
	unsigned int s=0;
	unsigned int i=0;
	unsigned int p=0;
	
#if ATTEMPT_IDENTICAL
	while(s < doublepopsize) 
#else
		while(s < popsize) 
#endif
		{
			// XXX      sortedfronts[i].calcCrowdingDistances();
			
			if(s+sortedfronts[i].size() > popsize)
			{
				front backup = sortedfronts[i];
				
				sortedfronts[i].sortOnCD();
				
				for(unsigned int j=0; j < sortedfronts[i].size(); j++)
					popsorted[p++] = &sortedfronts[i][j];
				
				sortedfronts[i] = backup; // restore sorting on objectives
			}
			else
				for(unsigned int j=0; j < sortedfronts[i].size(); j++)
					popsorted[p++] = &sortedfronts[i][j];
			
			s += sortedfronts[i++].size();
		}
	
#if !ATTEMPT_IDENTICAL
	while(s < doublepopsize)
	{
		for(unsigned int j=0; j < sortedfronts[i].size(); j++)
			popsorted[p++] = &sortedfronts[i][j];
		
		s += sortedfronts[i++].size();
	}
#endif
	
	// Added by A.V.
	paretofront = sortedfronts[0];
}

/***********************************************************************************************************/

// The many-dimensional algorithm follows:

// called from helper_A. Assumes s2 has not been observed inferior
// in the remaining objectives
bool solution::MdomA(const solution &s2, unsigned int M)
{
	bool better = false;
	
	for(vector<double>::size_type i=0; i < M; i++)
	{
		if(objectives[i] > s2.objectives[i])
			return(false);
		
		if(objectives[i] < s2.objectives[i])
			better = true;
	}
	
	return(better);
}

/***********************************************************************************************************/

// called from helper_B. Assumes s2 has been observed inferior
// in the remaining objectives
bool solution::MdomB(const solution &s2, unsigned int M)
{
	for(vector<double>::size_type i=0; i < M; i++)
		if(objectives[i] > s2.objectives[i])
			return(false);
	
	return(true);
}

/***********************************************************************************************************/

void cleanup_stairs(map<double, unsigned int> &stair,double xval)

{
	// search through stairs from start and upwards, removing elements with 
	// a lower frontnumber
	
	map<double, unsigned int>::iterator start = stair.find(xval);
	map<double, unsigned int>::iterator slut;
	
	if(start!=stair.end()) {
		unsigned int frontnum= start->second;
		start++;
		
		if(start!=stair.end() && start->second<=frontnum) {
			slut=start;slut++;
			
			while(slut!=stair.end() && slut->second<=frontnum) slut++;
			
			stair.erase(start,slut);
		}
	}
}

/***********************************************************************************************************/

/*
 void checkstairs(map<double, unsigned int > &stairs, char tekst[])
 {
 map<double, unsigned int >::iterator i;
 
 bool flawed=false;
 unsigned int f=0;
 for(i=stairs.begin();i!=stairs.end();i++) {
 if(i->second<=f)
 flawed=true;
 f=i->second;
 }
 
 if(flawed) {
 cout << "Fejl i trappe: " << tekst << endl;
 for(i=stairs.begin();i!=stairs.end();i++)
 cout << i->first << ":" << i->second << endl;
 exit(1);
 } else {
 cout << "trappe ok:" << endl;
 for(i=stairs.begin();i!=stairs.end();i++)
 cout << i->first << ":" << i->second << endl;
 }
 }
 */

/***********************************************************************************************************/

void population::ND_helper_A(vector<solution *> &S, unsigned int M, bool dosort)
{
	if(S.size()>1) {
		if(S.size()==2) {
			// compare the two solutions on M objectives.
			if(S[0]->MdomA(*S[1],M))
				S[1]->updateFront(*S[0]);
			else 
				if(S[1]->MdomA(*S[0],M))
					S[0]->updateFront(*S[1]);
		} else {
			if(M==2) {
				// two-dimensional case: do sweepline algorithm.
				map<double, unsigned int > stairs;
				double lasto0,lasto1;
				unsigned int lastfront;
				
				sortOnFirstObjective(S);
				
				vector<solution *>::iterator si=S.begin();
				
				// insert first element of S in datastructures
				
				stairs[(*si)->getObj(1)]=(*si)->getFrontNumber();
				lasto0=(*si)->getObj(0);
				lasto1=(*si)->getObj(1);
				lastfront=(*si)->getFrontNumber();
				
				si++;
				
				// Traverse S from left to right.
				while( si!=S.end() ) {
					
					// find position of si within stairs and update frontnumber
					map<double, unsigned int>::iterator pos;
					pos=stairs.lower_bound((*si)->getObj(1));
					if(pos==stairs.end()) pos--;
					if( pos->first>(*si)->getObj(1) && pos!=stairs.begin() )
						pos--;
					
					if( (*si)->getObj(1)>pos->first || 
					   ((*si)->getObj(1)==pos->first && 
						((*si)->getObj(0)!=lasto0 || (*si)->getObj(1)!=lasto1) ) ) {
						   (*si)->updateFront(pos->second+1);
					   } else
						   if( (*si)->getObj(0)==lasto0 && (*si)->getObj(1)==lasto1 ) 
							   (*si)->updateFront(lastfront);
					
					
					// Insert in stairs	  
					
					if( pos->first <= (*si)->getObj(1) && (int)pos->second < (*si)->getFrontNumber()
					   || pos->first>(*si)->getObj(1) ) {
						// The new element must be inserted in stairs
						stairs[(*si)->getObj(1)]=(*si)->getFrontNumber();
						cleanup_stairs(stairs,(*si)->getObj(1));
						lasto0=(*si)->getObj(0);
						lasto1=(*si)->getObj(1);
						lastfront=(*si)->getFrontNumber();
					}
					si++;
				}
			} else {     
				// recursive case
				
#if SPLITTING==SORTSPLITTING  
				if(dosort) {
					setCurrentObj(M-1);
					sort(S.begin(),S.end(),cmp_1);
				} 
				
				if(S[0]->getObj(M-1)!=S[S.size()-1]->getObj(M-1)) {
					unsigned int cutinx=S.size()/2;
					if(S[cutinx]->getObj(M-1)!=S[S.size()-1]->getObj(M-1)) {
						while(S[cutinx+1]->getObj(M-1)==S[cutinx]->getObj(M-1))
							cutinx++;
						cutinx++;
					} else 
						while(S[cutinx-1]->getObj(M-1)==S[cutinx]->getObj(M-1))
							cutinx--;
					
					vector<solution *> L,H;
					L.reserve(cutinx); 	  
					H.reserve(S.size()-cutinx); 	  
					
					copy(S.begin(),S.begin()+cutinx,back_inserter(L));
					copy(S.begin()+cutinx,S.end(),back_inserter(H));
					
					double maxL=L[0]->getObj(M-1);
					for(vector<solution *>::iterator i=L.begin();i!=L.end();i++)
						maxL=max(maxL,(*i)->getObj(M-1));
					
					double minH=H[0]->getObj(M-1);
					for(vector<solution *>::iterator i=H.begin();i!=H.end();i++)
						minH=min(minH,(*i)->getObj(M-1));
					
					if(maxL>=minH) {
						cout << "FORKERT! " << maxL << "   " << minH << endl;
						exit(10);
					}
					
					vector<solution *> HH=H;
					ND_helper_A(L,M,false);
					ND_helper_B(L,H,M-1,true);
					ND_helper_A(HH,M,false);
				} else {
					
					ND_helper_A(S,M-1,true);
				}
				
#elif SPLITTING==PARTITIONSPLITTING
				unsigned int cutinx;  
				if(splitA(S,M-1,cutinx)) {
#ifdef DOCHECKING
					checksplit(&S,NULL,M-1,cutinx,0,"splitA");
#endif
					
					vector<solution *> L,H;
					L.reserve(cutinx); 	  
					H.reserve(S.size()-cutinx); 	  
					
					copy(S.begin(),S.begin()+cutinx,back_inserter(L));
					copy(S.begin()+cutinx,S.end(),back_inserter(H));
					
					double maxL=L[0]->getObj(M-1);
					for(vector<solution *>::iterator i=L.begin();i!=L.end();i++)
						maxL=max(maxL,(*i)->getObj(M-1));
					
					double minH=H[0]->getObj(M-1);
					for(vector<solution *>::iterator i=H.begin();i!=H.end();i++)
					{
						//minH=min(minH,(*i)->getObj(M-1));
						//Added to fix debug error  
						
						if (minH>(*i)->getObj(M-1)) minH=((*i)->getObj(M-1));
					}
					
					if(maxL>=minH) {
						cout << "FORKERT! " << maxL << "   " << minH << endl;
						exit(1);
					}
					
					ND_helper_A(L,M,false);
					ND_helper_B(L,H,M-1,true);
					ND_helper_A(H,M,false);
				} else {
					
#ifdef DOCHECKING
					vector<solution *>::iterator i=S.begin();
					vector<solution *>::iterator j=S.begin();
					i++;
					while(i!=S.end()) {
						if( (*i)->getObj(M-1)!=(*j)->getObj(M-1) ) {
							cout << "OOOPPS!" << endl;
							
							for(vector<solution *>::iterator k=S.begin();k!=S.end();k++) 
								cout << (*k)->getObj(M-1) << endl;
							exit(1);
						}
						i++;j++;
					}
#endif  	       
					ND_helper_A(S,M-1,true);
				}
#else
#error "PARTITIONING not set!"
#endif	
			}
		}
	}
}

/***********************************************************************************************************/

void population::ND_helper_B(vector<solution *> &L, vector<solution *> &H, unsigned int M, bool dosort)
{
	if(L.size()!=0 && H.size()!=0) {
		if(L.size()==1) {
			// compare all elements in H to L[0]
			for(vector<solution *>::iterator i=H.begin();i!=H.end();i++)
				if(L[0]->MdomB(**i,M))
					(*i)->updateFront(*L[0]);
		} else 
			if(H.size()==1) {
				// compare all elements in L to H[0]
				for(vector<solution *>::iterator i=L.begin();i!=L.end();i++)
					if((*i)->MdomB(*H[0],M)) 
						H[0]->updateFront(**i);
			} 
			else 
				if(M==2) {
					// two-dimensional case: do sweepline algorithm
					
					map<double, unsigned int > stairs;
					
					sortOnFirstObjective(L);
					sortOnFirstObjective(H); 
					
					
					vector<solution *>::iterator li=L.begin();
					vector<solution *>::iterator hi=H.begin();
					
					// fast forward through H until *li dominates *hi
					while( hi!=H.end() && cmp_2(*hi,*li) ) 
						hi++;
					
					// insert first element of L in datastructures
					
					stairs[(*li)->getObj(1)]=(*li)->getFrontNumber();
					li++;
					
					// Traverse L and H from left to right. (merge-like)
					while( hi!=H.end() ) {
						while(li!=L.end()&& !cmp_2(*hi,*li)) {//takes advantage of lazy eval!
							// the next solution comes from L.
							
							map<double, unsigned int>::iterator pos;
							
							pos=stairs.lower_bound((*li)->getObj(1));
							if( pos->first>(*li)->getObj(1) && pos!=stairs.begin() || pos==stairs.end() )
								pos--;
							
							if( (int)pos->second < (*li)->getFrontNumber() || 
							   (pos==stairs.begin() && pos->first > (*li)->getObj(1)) ) {
								// The new element must be inserted in stairs
								stairs[(*li)->getObj(1)]=(*li)->getFrontNumber();
								cleanup_stairs(stairs,(*li)->getObj(1));
							} 
							li++;
						}
						
						// the next solution comes from H.
						// find position of hi within stairs and update frontnumber
						map<double, unsigned int>::iterator pos;
						pos=stairs.lower_bound((*hi)->getObj(1));
						if( pos->first > (*hi)->getObj(1) && pos!=stairs.begin() || pos==stairs.end() )
							pos--;
						
						if((*hi)->getObj(1) >= pos->first) 
							(*hi)->updateFront(pos->second+1);
						
						hi++;
					}
				}
				else { 
					// non-base case. Do recursion
					
#if SPLITTING==SORTSPLITTING  
					setCurrentObj(M-1);
					if(dosort) {
						sort(L.begin(),L.end(),cmp_1);
						sort(H.begin(),H.end(),cmp_1);
					}
					
					if(L[L.size()-1]->getObj(M-1)<H[0]->getObj(M-1))
						ND_helper_B(L,H,M-1,true);
					else {  
						unsigned int splitLL,splitLH,splitHH,splitHL;
						
						if(L[0]->getObj(M-1)!=L[L.size()-1]->getObj(M-1)) {
							splitLL=L.size()/2;
							if(L[splitLL]->getObj(M-1)!=L[0]->getObj(M-1)) {
								while(L[splitLL-1]->getObj(M-1)==L[splitLL]->getObj(M-1))
									splitLL--;
							} else {
								while(L[splitLL]->getObj(M-1)==L[splitLL+1]->getObj(M-1))
									splitLL++;
								splitLL++;
							}
							splitLH=(lower_bound(H.begin(),H.end(),L[splitLL],cmp_1)-H.begin());
						} else {
							splitLL=0;splitLH=0;
						}
						
						if(H[0]->getObj(M-1)!=H[H.size()-1]->getObj(M-1)) {
							splitHH=(H.size()+1)/2;
							if(H[splitHH]->getObj(M-1)!=H[H.size()-1]->getObj(M-1)) {
								while(H[splitHH]->getObj(M-1)==H[splitHH+1]->getObj(M-1))
									splitHH++;
								splitHH++;
							} else 
								while(H[splitHH-1]->getObj(M-1)==H[splitHH]->getObj(M-1)) {
									splitHH--;
								}
							splitHL=(lower_bound(L.begin(),L.end(),H[splitHH],cmp_1)-L.begin());
						} else {
							splitHH=0;splitHL=0;
						}
						
						if( splitLL!=0 || splitHH!=0 || splitLH!=0 || splitHL!=0) {
							// choose the split that maximizes |L2|*|H1|
							
							vector<solution *> L1,H1,L2,H2;
							
							if( (L.size()-splitLL)*splitLH >= (L.size()-splitHL)*splitHH &&
							   (0<splitLL && splitLL<L.size())||(0<splitLH && splitLH<H.size())) {
								L1.reserve(splitLL);
								copy(L.begin(),L.begin()+splitLL,back_inserter(L1));
								L2.reserve(L.size()-splitLL);
								copy(L.begin()+splitLL,L.end(),back_inserter(L2));	  
								H1.reserve(splitLH);
								copy(H.begin(),H.begin()+splitLH,back_inserter(H1));
								H2.reserve(H.size()-splitLH);
								copy(H.begin()+splitLH,H.end(),back_inserter(H2));	  
							} else {
								L1.reserve(splitHL);
								copy(L.begin(),L.begin()+splitHL,back_inserter(L1));
								L2.reserve(L.size()-splitHL);
								copy(L.begin()+splitHL,L.end(),back_inserter(L2));	  
								H1.reserve(splitHH);
								copy(H.begin(),H.begin()+splitHH,back_inserter(H1));
								H2.reserve(H.size()-splitHH);
								copy(H.begin()+splitHH,H.end(),back_inserter(H2));	  
							}
							vector<solution *> HH2=H2;
							ND_helper_B(L1,H1,M,false);
							
							ND_helper_B(L1,H2,M-1,true);
							
							ND_helper_B(L2,HH2,M,false);
						} else {
							// all elements in L are identical in obj M, 
							// and all elements in H are identical in obj M. 
							// If the elements of L have a smaller (M-1)-value
							// than H, we call recursively, otherwise we don't 
							// need to do anything.
							if(L[0]->getObj(M-1)<=H[0]->getObj(M-1))
								ND_helper_B(L,H,M-1,true);
						}
					}
#elif SPLITTING==PARTITIONSPLITTING
					unsigned int splitL,splitH;
					if(splitB(L,H,M-1,splitL,splitH)) {
#ifdef DOCHECKING
						checksplit(&L,&H,M-1,splitL,splitH,"splitB");
#endif
						
						vector<solution *> L1,H1,L2,H2;
						L1.reserve(splitL);
						copy(L.begin(),L.begin()+splitL,back_inserter(L1));
						L2.reserve(L.size()-splitL);
						copy(L.begin()+splitL,L.end(),back_inserter(L2));	  
						H1.reserve(splitH);
						copy(H.begin(),H.begin()+splitH,back_inserter(H1));
						H2.reserve(H.size()-splitH);
						copy(H.begin()+splitH,H.end(),back_inserter(H2));	  
						
						ND_helper_B(L1,H1,M,false);
						
						ND_helper_B(L1,H2,M-1,true);
						
						ND_helper_B(L2,H2,M,false);
					} else {
						// all elements in L are identical in obj M, 
						// and all elements in H are identical in obj M. 
						// If the elements of L have a smaller (M-1)-value
						// than H, we call recursively, otherwise we don't 
						// need to do anything.
						
#ifdef DOCHECKING
						vector<solution *>::iterator i=L.begin();
						vector<solution *>::iterator j=L.begin();
						i++;
						while(i!=L.end()) {
							if( (*i)->getObj(M-1)!=(*j)->getObj(M-1) ) {
								cout << "LOOOPPS!" << endl;
								
								for(vector<solution *>::iterator k=L.begin();k!=L.end();k++) 
									cout << (*k)->getObj(M-1) << endl;
								exit(1);
							}
							i++;j++;
						}  
						
						i=H.begin();
						j=H.begin();
						i++;
						while(i!=H.end()) {
							if( (*i)->getObj(M-1)!=(*j)->getObj(M-1) ) {
								cout << "HOOOPPS!" << endl;
								
								for(vector<solution *>::iterator k=H.begin();k!=H.end();k++) 
									cout << (*k)->getObj(M-1) << endl;
								exit(1);
							}
							i++;j++;
						}
#endif  // dochecking  
						
						if(L[0]->getObj(M-1)<=H[0]->getObj(M-1))
							ND_helper_B(L,H,M-1,true);
					}
#endif
				}
	}
}


#endif

/***********************************************************************************************************/

