
#include "esc.hpp"
#include "types.hpp"
#include "misc.hpp"

#include <iostream>

//#include <iomanip>
//#include <math.h>
//#include <cstdlib>

using namespace std;

long seed;

// Global variable storing parameters read from the arguments file in string format.
map < string, string > args;

// Global variable storing parameters read from the arguments file in appropriate formats.
esc *model;

// Global variable to store all of solutions (symbionts and hosts) through time/levels.
vector < vector <solution *> > popHistory;

// Global variable to store indexes to individuals (symbionts) that will be used in next level.
vector <vector <int> > symbsToUse;

// Global variable to store number of references to individuals (symbionts) that will be used in next level.
vector <vector <int> > symbsRefs;
//vector <vector <int> > symbsRefsPrev;

/***********************************************************************************************************/

int main(int argc, char **argv)
{
	/* Argument must specify parameter file. */
	
	if(argc != 2)
		die(__FILE__, __FUNCTION__, __LINE__, "bad arguments");
	
	
	map < string, string > :: iterator ariter;
	
	
	/* Get arguments. */
	cout << readMap(argv[1], args) << " args read" << endl;
	
	/* Seed random number generator. */
	if((ariter = args.find("seed")) == args.end())
		die(__FILE__, __FUNCTION__, __LINE__, "cannot find arg seed");
	
	srand48(stringToInt(ariter->second));
	
	model = new esc(args);
	model->go();
	
	cout << endl << endl;
	
	double sys1, usr1, sys2, usr2;

	// Get the starting time
	if(getusage(sys1, usr1) == false)
		die(__FILE__, __FUNCTION__, __LINE__, "cannot get first rusage");
	
//	for(model->_lev = 0; model->_lev < model->_numLevels; model->_lev++) 
//	{
	
	model->_lev = 0;
	model->_numGenerations = model->_numGenerations0;
	model->_numObjectives = model->_numObjectives0;
	
	// Initialize current level population.
	population p;//p(model->_lev)
	
	
	// Print the symbiont population before the evolutionary procss.
//	cout << endl << "main() => Lev = " << model->_lev << ", Symbiont Population Before Evolution!" << endl << endl;
//	p.print(0);
//	cout << endl;
	
	// Run the first evolutionary phase (symbiont level.)
	p.run();
	
	// Print the symbiont population after the evolutionary procss.
//	cout << endl << "main() => Lev = " << model->_lev << ", Symbiont Population After Evolution!" << endl << endl;
//	p.print(0);
//	cout << endl;
	
	// Finalizing level 0.
	p.finalizeLevel(model->_lev);
	
	// Next level (This can be done in finalizeLevel() function too.)
	model->_lev = 1;
	model->_numGenerations = model->_numGenerations1;
	model->_numObjectives = model->_numObjectives1;
	
	// Initialize current level population
	population p2;//p(model->_lev)
	
	// Print the host population before the evolutionary procss.
	cout << endl << "main() => Lev = " << model->_lev << ", Host Population Before Evolution!" << endl << endl;
	p2.print(0);
	cout << endl;
	
	// Run the second evolutionary phase (first host level.)
	p2.run();
	
	// Print the host population after the evolutionary procss.
	cout << endl << "main() => Lev = " << model->_lev << ", Host Population After Evolution!" << endl << endl;
	p2.print(0);
	cout << endl;
	
	// Finalizing level 1.
	p2.finalizeLevel(model->_lev);
	
//	}
	
	/* Record training time and print out time to train up to and including the current level. */
	if(getusage(sys2, usr2) == 0)
		die(__FILE__, __FUNCTION__, __LINE__, "cannot get last rusage");
	
	cout << endl << "main() => Time:"; 
	cout << " sys1 " << sys1 << " usr1 " << usr1 << " sum1 " << (sys1 + usr1);
	cout << " sys2 " << sys2 << " usr2 " << usr2 << " sum2 " << (sys2 + usr2);
	cout << " traintime = " << (sys2 + usr2) - (sys1 + usr1) << endl;
	
	if(model->_lev !=0 && model->_useAttrProbLrng)
	{
		cout << endl << "population::finalizeLevel() => Attribute probabilities:" << endl;
		for(int i=0; i < model->_dataSetDim; i++)
			cout << model->_attrProbs[i] << ", ";
	}
	cout << endl;
	
	cout << endl << "main() => Performance Measures for " << popHistory[2].size() << " solutions." << endl << endl;
	
	int temo;
	
	for(int i=0; i < (int)popHistory[2].size(); i++)
	{
		cout << "Solution (" << i << "): " << endl;
		
		popHistory[2][i]->print(0);
		
		for(int j=0; j < popHistory[2][i]->size(); j++)
		{
			temo = popHistory[2][i]->getGene().index1[j];
			popHistory[1][temo]->print(0);
		}
		
		if(model->_labelsProvided)
			popHistory[2][i]->performance_measures(1, 1);
		
		cout << endl;
	}

	cout << "main() => DONE" << endl;
	
//	cout << sizeof(popHistory)
	delete model;
	
	return 0;
}
