
//#include "esc.hpps"
#include "types.hpp"
#include "misc.hpp"
#include "permute.hpp"

#include <iostream>
#include <iomanip>
#include <fstream>
#include <vector>
#include <algorithm>

using namespace std;

// My declarations
//vector <int> attrsIndexed;
//vector <int> instsIndexed;
//int numAttrsIndexed;
//int numInstsIndexed;

extern esc *model;
extern vector < vector <solution *> > popHistory;
extern vector < vector <int> > symbsToUse;
extern vector < vector <int> > symbsRefs;
//extern vector < vector <int> > symbsRefsPrev;

// sumDistances is used for shared fitness evaluation
vector <double> sumDistances;


/***********************************************************************************************************/

void addToPopHistory(int lev, solution *parent)
{
	solution *temp = new solution(parent);
	
#ifdef DEBUGGING
	cout << "addToPopHistory => Print the new PARENT symbiont (level = " << lev << "). " << endl;
	temp->print(0);
#endif
	
	// Add/Clone the symb to global variable popHistory
	popHistory[lev].push_back(temp);
	
#ifdef DEBUGGING
	cout << "addToPopHistory => Print the new CLONED symbiont (level = " << lev << "). " << endl;
	popHistory[lev].back()->print(0);
#endif
	
	// Update symbsToUse and symbsRefs
	symbsToUse[lev].push_back(popHistory[lev].size()-1);
	symbsRefs[lev].push_back(1);
	
	// Update the number of symbionts available to the current level.
	model->_numSymbsToUse++;
	
#ifdef DEBUGGING
	cout << "addToPopHistory() => symbsToUse[" << lev << "].size() = " << symbsToUse[lev].size() << endl;
	cout << "addToPopHistory() => popHistory[" << lev << "].size() = " << popHistory[lev].size() << endl;
	cout << "addToPopHistory() => model->_numSymbsToUse = " << model->_numSymbsToUse << endl;
#endif

}
/***********************************************************************************************************/

population::population()//int l
//: level(l)
{
#ifdef DEBUGGING
	cout << "population::population() enter." << endl;
#endif
	
	popsize = model->_hostPopSize;
	doublepopsize = 2 * model->_hostPopSize;
	
	level = model->_lev;
	
	// This is to flag the evaluation function that this is the first time the population is being evaluated.
	//model->_gen = -1;
	
	pop = vector <solution> (2 * model->_hostPopSize);
	//pop.reserve(2 * model->_hostPopSize);
	
	popsorted = vector <solution *> (2 * model->_hostPopSize);
	//popsorted.reserve(2 * model->_hostPopSize);
	
	for(int i=0; i < 2 * model->_hostPopSize; i++)
	{
		//pop[i] = solution(level);
		
#ifdef DEBUGGING
		cout << "population::population() => pop[" << i << "].size() = " << pop[i].size() <<endl;
#endif
		
		pop[i].randomize();
		
		//cout << "population::population() => evaluating individual " << i << " ..." << endl;
		// -1 means that this evaluation is perfomred prior to evlutionary loop.
		//pop[i].evaluate(-1);
		//cout << "population::population() => evaluation individual " << i << " finished!" << endl;
		
		popsorted[i] = &pop[i];
	}
	
#ifdef DEBUGGING
	cout << "population::population() exit." << endl;
#endif
}

/***********************************************************************************************************/

int population::print(int t=0)
{
#ifdef DEBUGGING
	cout << "population::print() enter." << endl;
#endif
	
	if(t == 0)
	{
		for(int i=0; i < 2 * model->_hostPopSize; i++)
		{
			cout << "individual/solution (" << i << "): " << endl;
			(*popsorted[i]).print(0);
		}
		/*
		 for(vector<front>::size_type i=0; i < sortedfronts.size(); i++)
		 {
		 cout << "Front " << i+1 << ":" << endl;
		 sortedfronts[i].print(0); // front::print function
		 cout << endl;
		 }
		 */
	}
	
	else if(t == 1)
	{
		int temp;
		
		vector <int> geneSizeVect(model->_dataSetDim + 1, 0); // vector of zeros
		
		for(vector<solution *>::size_type i=0; i < popsorted.size(); i++)
		{
			temp = (*popsorted[i]).print(t);
			geneSizeVect[temp]++;
		}
		
		for(size_t k=1; k < geneSizeVect.size(); k++)
		{
			//cout << k << " " << geneSizeVect[k] << endl;
			cout << geneSizeVect[k] << " ";
		}
		cout << endl;
	}
	
#ifdef DEBUGGING
	cout << "population::print() exit." << endl;
#endif
	
	return (int)pop.size();
}

/***********************************************************************************************************/

void population::run()
{
	cout << "population::run() enter." << endl << endl;
	
	model->_gen = -1;
	
	// Perform the subsampling if _doSubsampling is set (no need here, this is done in esc::go())
	if(model->_doSubsampling == 1)
		model->RSS();
	
	// Evaluate the population, model->_gen equals -1
	evaluate(-1);
	
#ifdef DEBUGGING
	cout << "population::run() => popsorted before sortOnFronts():" << endl;
	print(0);
	cout << endl;
#endif
	
	// Sort the population on fronts after first evaluation.
	sortOnFronts();
	
#ifdef DEBUGGING
	cout << "population::run() => popsorted after sortOnFronts() and before going through evolutionary process." << endl;
	print(0);
	cout << endl;
#endif
	
	// paretofront is initialized manualy once. It will be re-assigned in finalizeGeneration() hereafter.
	paretofront = sortedfronts[0];
	
#ifdef DEBUGGING
	cout << "population::run() => paretofront:" << endl;
	paretofront.print(0);
#endif
	
	// Initialize the attributes and instances indexed by PF individuals after first evaluation.
	if(level == 0)
		attrsInstsOnPF();
	
	int numGens = 0;
	
	if(level == 0)
		numGens = model->_numGenerations0;
	else if(level > 0 && level < model->_numLevels)
		numGens = model->_numGenerations1;
	else
		die(__FILE__, __FUNCTION__, __LINE__, "Wrong level value!");


	// Start the evolutionary loop.
	for(model->_gen=0; model->_gen < numGens; model->_gen++)
	{
		
		cout << endl << "Generation: " << setw(2) << model->_gen << endl;
		cout << "===============" << endl;

#ifdef DEBUGGING
		cout << "population::run() => Population at gen " << model->_gen << " before breedGeneration()" << endl;
		print(0);
		cout << endl;
#endif
		
		// Reproduce new individuals.
		breedGeneration(model->_gen);
		
#ifdef DEBUGGING
		cout << "population::run() => Population at gen " << model->_gen << " after breedGeneration()" << endl;
		print(0);
		cout << endl;
#endif
		
		// Evaluation is done here since the evaluation call in population::breedGeneration() is commented out.
		evaluate(model->_gen);
		
		// Sort new individuals on fronts.
		sortOnFronts();
		
#ifdef DEBUGGING
		cout << "population::run() => Population at gen " << model->_gen << " after sortOnFronts()" << endl;
		print(0);
		cout << endl;
#endif
		
		// Get attributes and instances indxed by PF inds.
		if(level == 0)
			attrsInstsOnPF();
		
		// Finalize current generation.
		finalizeGeneration(level, model->_gen);
		
#ifdef DEBUGGING
		cout << "population::run() => Population at gen " << model->_gen << " after finalizeGeneration()" << endl;
		print(0);
		cout << endl;
#endif
		
	}
	
	cout << "population::run() exit." << endl;
}

/***********************************************************************************************************/

void population::evaluate(int gen)
{
#ifdef DEBUGGING
	cout << "population::evaluate() enter." << endl;
#endif
	
	sumDistances.clear();
	sumDistances.resize(model->_dataSetSize, -1);
	
	double dist, tempdist, sumdist;
	int temp_int;

	// If there's only one single shared-fitness to be calculated.
	if((level != 0) && (model->_numObjectives1 == 1))
	{
		for(int i=0; i < model->_dataSetSize; i++) // Iterates on all points of the dataset.
		{
			//cout << "population::evaluate() => Point " << i << endl;
			sumdist = 0;
			
			for(int j=0; j < 2 * model->_hostPopSize; j++) // Iterates on the whole population of hosts.
			{
				//cout << "population::evaluate() => host " << j << " = popsorted[" << j << "] :" << endl; 
				//popsorted[j]->print(0);
				dist = model->_infinity;
				tempdist = 0;
				
				for(int k=0; k < (int)popsorted[j]->size(); k++) // Iterates on the symbionts of a host.
				{
					temp_int = popsorted[j]->getGene().index1[k];
					
					//cout << "population::evaluate() => host " << j << " symb " << j << " = " << "popHistory[" << level << "][" << temp_int << "]" << ":" << endl;
					//popHistory[level][temp_int]->print(0);
					
					tempdist = popHistory[level][temp_int]->getDistances()[i];
					
					//cout << "population::evaluate() => point " << i << " host " << j << " symb " << k << " tempdist = " << tempdist << endl;
					
					if(tempdist < dist)
					{
						dist = tempdist;
					}
				}
				
				//cout << "population::evaluate() => point " << i << " host " << j << " dist = " << dist << endl;
				sumdist += dist;
				//cout << "population::evaluate() => point " << i << " sumdist so far = " << sumdist << endl;
			}
			
			//cout << "population::evaluate() => point " << i << " sumdist = " << sumdist << endl;
			sumDistances[i] = sumdist;
		}
	}
	
	
	// Note "2 * model->_hostPopSize" and not "model->_hostPopSize" because individuals hostPopSize+1 to 2*hostPopSize have
	// already been generated and individuals 1 to hostPopSize need to be evaluated against a different subset of instances.
	for(int i=0; i < 2 * model->_hostPopSize; i++)
	{
#ifdef DEBUGGING
		cout << "population::evaluate() => Evaluating ind[" << i << "]" << endl;
#endif
		
		//if(model->_activeSetModified || pop[i].modified)
		pop[i].evaluate(gen);
	}
	
#ifdef DEBUGGING
	cout << "population::evaluate() exit." << endl;
#endif
}

/***********************************************************************************************************/

void population::breedGeneration(int gen)
{
	// Function to reproduce popsize new individuals.

#ifdef DEBUGGING
	cout << "population::breedGeneration() enter." << endl;
#endif
	
	vector <solution *> fathers1 = permutation(popsorted, 0, popsize);
	vector <solution *> fathers2 = permutation(popsorted, 0, popsize);
	vector <solution *> mothers1 = permutation(popsorted, 0, popsize);
	vector <solution *> mothers2 = permutation(popsorted, 0, popsize);
	
	int tempcntr;
	double temp_pc, temp_pm, temp_pmlm;
	int symbIndx, symbVal;
	
	temp_pc = drand48();
	temp_pm = drand48();
	temp_pmlm = drand48();
	
#ifdef DEBUGGING
	// Print the popsorted, fathers1, fathers2, mothers1, and mothers2 vectors
	cout << "population::breedGeneration() => popsorted population" << endl;
	for(size_t i=0; i < popsorted.size(); i++)
		(*popsorted[i]).print(0);
	
	cout << "population::breedGeneration() => fathers1 population: " << endl;
	for(size_t i=0; i < fathers1.size(); i++)
		(*fathers1[i]).print(0);
	
	cout << "population::breedGeneration() => fathers2 population: " << endl;
	for(size_t i=0; i < fathers2.size(); i++)
		(*fathers2[i]).print(0);
	
	cout << "population::breedGeneration() => mothers1 population: " << endl;
	for(size_t i=0; i < mothers1.size(); i++)
		(*mothers1[i]).print(0);
	
	cout << "population::breedGeneration() => mothers2 population: " << endl;
	for(size_t i=0; i < mothers2.size(); i++)
		(*mothers2[i]).print(0);
#endif
	
	for(unsigned int i=0; i < popsize; i++)
	{
		solution *far, *mor, *parent;
		
		// temporary initialization of parent. will be re-initialized later.
		parent = fathers1[i];
		
		// Decrease the number of references to symbionts that are linked to the host to be replaced.
		if(level != 0)
			for(int j=0; j < (int) popsorted[i+popsize]->size(); j++)
			{
				tempcntr = popsorted[i+popsize]->getGene().index1[j];
				popHistory[level][tempcntr]->decrefs();
			}
		
#ifdef DEBUGGING
		// Print the child before crossover and mutation.
		cout << "population::breedGeneration() => Child to be replaced:" << endl;
		(*popsorted[i+popsize]).print(0);
#endif
		
		// Select father
#ifdef DEBUGGING
		cout << "population::breedGeneration() => fathers1[" << i << "] is:" << endl;
		(*fathers1[i]).print(0);
		cout << "population::breedGeneration() => fathers2[" << i << "] is:" << endl;
		(*fathers2[i]).print(0);
#endif
		
		// << operator only checks frontnumber and crowdingdistance, whereas >> checks the domination (ref to sorting.cpp)
		if((*fathers1[i]) << (*fathers2[i]))
		{
			//cout << "fathers1[" << i << "] dominated fathers2[" << i << "]. " << endl;
			far = fathers1[i];
		}
		else if((*fathers2[i]) << (*fathers1[i]))
		{
			//cout << "fathers2[" << i << "] dominated fathers1[" << i << "]. " << endl;
			far = fathers2[i];
		}
		else
		{
			// Tie breaker based on size; select the smaller ind
#ifdef DEBUGGING
			cout << "population::breedGeneration() => None of the father inds dominates the other!!!" << endl;
			cout << "fathers1[" << i << "]'s frontnumber is: " << (*fathers1[i]).getFrontNumber();
			cout << ", crowdingdistance is: " << (*fathers1[i]).getCrowdingDistance();
			cout << ", and size is: " << (*fathers1[i]).size() << endl;
			cout << "fathers2[" << i << "]'s frontnumber is: " << (*fathers2[i]).getFrontNumber();
			cout << ", crowdingdistance is: " << (*fathers2[i]).getCrowdingDistance();
			cout << ", and size is: " << (*fathers2[i]).size() << endl;
#endif
			
			if((*fathers1[i]).size() > (*fathers2[i]).size())
			{
				//cout << "fathers2[" << i << "] is smaller than fathers1[" << i << "]. " << endl;
				far = fathers2[i];
			}
			else
			{
				//cout << "fathers1[" << i << "] is smaller than or equal to fathers2[" << i << "]. " << endl;
				far = fathers1[i];
			}
		}
		
#ifdef DEBUGGING
		cout << "population::breedGeneration() => Ind " << i << "'s father is: " << endl;
		far->print(0);
#endif
		
		
		// Select mother
#ifdef DEBUGGING
		cout << "population::breedGeneration() => mothers1[" << i << "] is:" << endl;
		(*mothers1[i]).print(0);
		cout << "population::breedGeneration() => mothers2[" << i << "] is:" << endl;
		(*mothers2[i]).print(0);
#endif
		
		// << operator only checks frontnumber and crowdingdistance, whereas >> checks the domination (ref to sorting.cpp)
		if((*mothers1[i]) << (*mothers2[i]))
		{
			//cout << "mothers1[" << i << "] dominated mothers2[" << i << "]. " << endl;
			mor = mothers1[i];
		}
		else if((*mothers2[i]) << (*mothers1[i]))
		{
			//cout << "mothers2[" << i << "] dominated mothers1[" << i << "]. " << endl;
			mor = mothers2[i];
		}
		else
		{
			// Tie breaker based on size; select the smaller ind
#ifdef DEBUGGING
			cout << "population::breedGeneration() => None of the mother inds dominates the other!!!" << endl;
			cout << "mothers1[" << i << "]'s frontnumber is: " << (*mothers1[i]).getFrontNumber();
			cout << ", crowdingdistance is: " << (*mothers1[i]).getCrowdingDistance();
			cout << ", and size is: " << (*mothers1[i]).size() << endl;
			cout << "mothers2[" << i << "]'s frontnumber is: " << (*mothers2[i]).getFrontNumber();
			cout << ", crowdingdistance is: " << (*mothers2[i]).getCrowdingDistance();
			cout << " and size is: " << (*mothers2[i]).size() << endl;
#endif
			
			if((*mothers1[i]).size() > (*mothers2[i]).size())
			{
				//cout << "mothers2[" << i << "] is smaller than mothers1[" << i << "]. " << endl;
				mor = mothers2[i];
			}
			else
			{
				//cout << "mothers1[" << i << "] is smaller than or equal to mothers2[" << i << "]. " << endl;
				mor = mothers1[i];
			}
		}
		
#ifdef DEBUGGING
		cout << "population::breedGeneration() => Ind " << i << "'s mother is: " << endl;
		mor->print(0);
#endif
		
		// Select the best parent among father and mother. 
		// << operator only checks frontnumber and crowdingdistance, whereas >> checks the domination (ref to sorting.cpp)
		if((*far) << (*mor))
		{
			//cout << "population::breedGeneration() => Father selected as parent." << endl;
			parent = far;
		}
		else if((*mor) << (*far))
		{
			//cout << "population::breedGeneration() => Mother selected as parent." << endl;
			parent = mor;
		}
		else // Tie breaker based on size; select the smaller ind
		{
			if(far->size() < mor->size())
			{
				//cout << "Father is smaller, hence selected as parent." << endl;
				parent = far;
			}
			else
			{
				//cout << "Mother is smaller or equal, hence selected as parent." << endl;
				parent = mor;
			}
		}
		
#ifdef DEBUGGING
		cout << "population::breedGeneration() => Ind " << i << "'s parent is: " << endl;
		parent->print(0);
#endif



/*
		if(temp_pc < model->_pc)
		{
#ifdef DEBUGGING
			// Print the child before crossover and muttion.
			cout << "population::breedGeneration() => Child before crossover and mutation:" << endl;
			(*popsorted[i+popsize]).print(0);
#endif
			
			// Apply the crossover operator.
			(*popsorted[i+popsize]).crossover(far, mor);
			
#ifdef DEBUGGING
			// Print the child after crossover and before mutation.
			cout << "population::breedGeneration() => Child after crossover and before mutation:" << endl;
			(*popsorted[i+popsize]).print(0);
#endif
			
			if(temp_pm < model->_pm)
			{
				// Apply the mutation operator.
				(*popsorted[i+popsize]).mutation();
			}
			
#ifdef DEBUGGING
			// Print the child after crossover and mutation.
			cout << "population::breedGeneration() => Child after crossover and mutation:" << endl;
			(*popsorted[i+popsize]).print(0);
#endif

		}//if(temp_pc < model->_pc)
		
		else // only do the mutation
		{
			// Copy the best parent 
			*popsorted[i+popsize] = *parent;
			
#ifdef DEBUGGING
			// Print the child before muttion.
			cout << "population::breedGeneration() => Child before mutation (mutation only): " << endl;
			(*popsorted[i+popsize]).print(0);
#endif
			
			// Apply the mutation operator only.
			if(temp_pm < model->_pm)
				(*popsorted[i+popsize]).mutation();
			
#ifdef DEBUGGING
			// Print the child after muttion.
			cout << "population::breedGeneration() => Child after mutation (mutation only): " << endl;
			(*popsorted[i+popsize]).print(0);
#endif

		}//if(temp_pc < model->_pc)
 */



		if(level == 0)
		{
			// Clone the best parent.
			*popsorted[i+popsize] = *parent;
			
#ifdef DEBUGGING
			// Print the child before muttion.
			cout << "population::breedGeneration() => Child before mutation:" << endl;
			(*popsorted[i+popsize]).print(0);
#endif
			
			// Apply the mutation operator only.
			if(temp_pm < model->_pm)
				(*popsorted[i+popsize]).mutation();
			
#ifdef DEBUGGING
			// Print the child after mutation.
			cout << "population::breedGeneration() => Child after mutation:" << endl;
			(*popsorted[i+popsize]).print(0);
#endif
			
		}//if(level == 0)
		
		else if((level != 0))// && (temp_pmlm < model->_pmlm))
		{
			// Clone the host that goes through multi-level mutation (MLM).
			*popsorted[i+popsize] = *parent;
			
#ifdef DEBUGGING
			// Print the child before mutation.
			cout << "population::breedGeneration() => Child before fixed-level mutation:" << endl;
			(*popsorted[i+popsize]).print(0);
#endif
			
			// Do the single-level mutation for this ind.
			if(temp_pm < model->_pm)
				(*popsorted[i+popsize]).mutation();
			
#ifdef DEBUGGING
			// Print the child after mutation.
			cout << "population::breedGeneration() => Child after fixed-level mutation:" << endl;
			(*popsorted[i+popsize]).print(0);
#endif
			
#ifdef DEBUGGING
			cout << "population::breedGeneration() => ind before multi-level mutation : " << endl;
			(*popsorted[i+popsize]).print(0);
#endif
			
			// Prepare for multi-level mutation.
			// Pick the symbiont that goes through further modification.
			symbIndx = (int) (drand48() * (*popsorted[i+popsize]).size());
			
			// Get the index of the symbiont to be cloned.
			symbVal = (*popsorted[i+popsize]).getGene().index1[symbIndx];
			
#ifdef DEBUGGING
			cout << "population::breedGeneration() => symbIndx = " << symbIndx << endl;
			cout << "population::breedGeneration() => symbVal = " << symbVal << endl;
			cout << "population::breedGeneration() => The individual to be cloned:" << endl;
			popHistory[level][symbVal]->print(0);
#endif
			
#ifdef DEBUGGING
			cout << "population::breedGeneration() => numSymbsToUse before adding the symbiont = " << model->_numSymbsToUse << endl;
			
			cout << "population::breedGeneration() => symbsToUse before adding the symbiont (" << symbsToUse[level].size() << "):" << endl;
			for(int j=0; j < (int)symbsToUse[level].size(); j++)
				cout << symbsToUse[level][j] << " ";
			cout << endl;
			
			cout << "population::breedGeneration() => popHistory before adding the new symbiont to popHistory (" << popHistory[level].size() << "):" << endl;
			for(int j=0; j < (int)symbsToUse[level].size(); j++)
			{
				cout << "Solution " << j << ":" << endl;
				popHistory[level][symbsToUse[level][j]]->print(0);
			}
#endif
			
			// Add the symbiont to popHistory and update required variables.
			addToPopHistory(level, popHistory[level][symbVal]);
			
#ifdef DEBUGGING
			cout << "population::breedGeneration() => popsorted[i+popsize] symbiont BEFORE updating: " << endl;
			vector <int> tempvec1 = (*popsorted[i+popsize]).getGene().index1;
			for(int pha=0; pha < (int) tempvec1.size(); pha++)
				cout << tempvec1[pha] << ", ";
			cout << endl;
			
			cout << "population::breedGeneration() => symb index to be removed: " << symbVal << endl;
#endif
			
			// Updating the index of the host to point to the new cloned symbiont instead of the old symbiont.
			vector <int> tempvec = (*popsorted[i+popsize]).getGene().index1;
			// replace the old symb index with new symb index
			tempvec[symbIndx] = (int)popHistory[level].size()-1;
			sort(tempvec.begin(), tempvec.end());
			(*popsorted[i+popsize]).setGene1(tempvec);
			
			
#ifdef DEBUGGING
			vector <int> tempvec2 = (*popsorted[i+popsize]).getGene().index1;
			for(int pha=0; pha < (int) tempvec2.size(); pha++)
				cout << tempvec2[pha] << ", ";
			cout << endl;
#endif
			
#ifdef DEBUGGING
			cout << "population::breedGeneration() => Updating the index of the host to point to the new cloned symbiont instead of the old symbiont:" << endl;
			(*popsorted[i+popsize]).print(0);
			
			cout << "population::breedGeneration() => the symbiont to be mutated:" << endl;
			popHistory[level].back()->print(0);
#endif
			
			// Mutate the newly added symbiont.
			popHistory[level].back()->mutation();
			
			// Update attribute list of the solution.
			popHistory[level].back()->updateAttrs();
			
#ifdef DEBUGGING
			cout << "population::breedGeneration() => calling calcDistance()" << endl;
#endif
			
			// calculate distances between newly mutated symbiont and _activePoints
//			if(model->_doConnectivity)
				popHistory[level].back()->calcDistances();
			
			
#ifdef DEBUGGING
			cout << "population::breedGeneration() => the symbiont after mutation:" << endl; 
			popHistory[level].back()->print(0);
#endif
			
#ifdef DEBUGGING
			cout << "population::breedGeneration() => numSymbsToUse after adding the symbiont = " << model->_numSymbsToUse << endl;
			
			cout << "population::breedGeneration() => symbsToUse after adding the symbiont (" << symbsToUse[level].size() << "):" << endl;
			for(int j=0; j < (int)symbsToUse[level].size(); j++)
				cout << symbsToUse[level][j] << " ";
			cout << endl;
			
			cout << "population::breedGeneration() => popHistory after adding the new symbiont to popHistory (" << popHistory[level].size() << "):" << endl;
			for(int j=0; j < (int)symbsToUse[level].size(); j++)
			{
				cout << "Solution " << j << ":" << endl;
				popHistory[level][symbsToUse[level][j]]->print(0);
			}
#endif
			
		}
		
		else
			die(__FILE__, __FUNCTION__, __LINE__, "Wrong level value!");
		
		
		// Check the eligibility of the new offspring.
		if(!(*popsorted[i+popsize]).isEligible())
		{
			cout << "population::breedGeneration() => The new offspring is not eligible, replacing with the best parent." << endl;
			
			// if not eligible replace the offspring with its parent 
			*popsorted[i+popsize] = *parent;
		}
		
		// Evaluating the new individual after applying diversity operators.
		//(*popsorted[i+popsize]).evaluate(gen);
		
		// Increase the number of references by the new solution.
		if(level != 0)
			for(int j=0; j < (int) popsorted[i+popsize]->size(); j++)
			{
				tempcntr = popsorted[i+popsize]->getGene().index1[j];
				popHistory[level][tempcntr]->increfs();
			}
		
	}
	
#ifdef DEBUGGING
	cout << "population::breedGeneration() exit." << endl;
#endif
}

/***********************************************************************************************************/

// Compare each individual with all solutions on the pareto front to count 
// the number of indexed features and instances by those solutions
void population::attrsInstsOnPF()
{
#ifdef DEBUGGING
	cout << "population::attrsInstsOnPF() enter." << endl;
#endif
	
	//vector <int> temp; // Used to be a temporary buffer used in the main loop for adding attributes and instances
	vector <int> :: iterator it;
	
	// Determining indexed attributes by individuals on pareto front (starting from 2nd solution on PF)
	
	attrs_on_PF.clear();
	attrs_on_PF.reserve(model->_dataSetDim);
	//num_attrs_on_PF = 0;
	
	// Initializing the attrs_on_PF with attributes of the first solution on PF
	attrs_on_PF = paretofront[0].getAttrs();
	
#ifdef DEBUGGING
	cout << "population::attrsInstsOnPF() => attrs_on_PF.size() = " << attrs_on_PF.size() << endl;
	for(int j=0; j < (int)attrs_on_PF.size(); j++)
		cout << attrs_on_PF[j] << " ";
	cout << endl;
#endif
	
	for(int i=0; i < (int)paretofront.size(); i++)
	{
		// For some reason merge() function did not work!!!
		//merge(attrs_on_PF.begin(), attrs_on_PF.end(), paretofront[i].getAttrs().begin(), paretofront[i].getAttrs().end(), temp.begin());
		
		// Combining the attributes
		for(int j=0; j < (int)paretofront[i].getAttrs().size(); j++)
			attrs_on_PF.push_back(paretofront[i].getAttrs()[j]);
		
#ifdef DEBUGGING
		cout << "population::attrsInstsOnPF() => (after push_back) attrs_on_PF.size() = " << attrs_on_PF.size() << endl;
		for(int j=0; j < (int)attrs_on_PF.size(); j++)
			cout << attrs_on_PF[j] << " ";
		cout << endl;
#endif
		
		// Sorting the instances
		sort(attrs_on_PF.begin(), attrs_on_PF.end());
		
#ifdef DEBUGGING
		cout << "population::attrsInstsOnPF() => (after sort) attrs_on_PF.size() = " << attrs_on_PF.size() << endl;
		for(int j=0; j < (int)attrs_on_PF.size(); j++)
			cout << attrs_on_PF[j] << " ";
		cout << endl;
#endif
		
		// Removing the duplicate attributes
		it = unique(attrs_on_PF.begin(), attrs_on_PF.end());
		attrs_on_PF.resize(it - attrs_on_PF.begin());
		
#ifdef DEBUGGING
		cout << "population::attrsInstsOnPF() => (after removing duplicates) attrs_on_PF.size() = " << attrs_on_PF.size() << endl;
		for(int j=0; j < (int)attrs_on_PF.size(); j++)
			cout << attrs_on_PF[j] << " ";
		cout << endl;
#endif
		
	}
	
#ifdef DEBUGGING
	cout << "population::attrsInstsOnPF() => Final attributes of attrs_on_PF are (total is " << attrs_on_PF.size() << ")= ";
	for(int j=0; j < (int)attrs_on_PF.size(); j++)
		cout << attrs_on_PF[j] << " ";
	cout << endl;
#endif
	
	
	
	// Determining indexed instances by individuals on pareto front
	
	insts_on_PF.clear();
	insts_on_PF.reserve(model->_dataSetSize);
	//num_insts_on_PF = 0;
	
	// Initializing the insts_on_PF with instances of teh first solution on PF
	insts_on_PF = paretofront[0].getInsts();
	
#ifdef DEBUGGING
	cout << "population::attrsInstsOnPF() => insts_on_PF.size() = " << insts_on_PF.size() << endl;
	for(int j=0; j < (int)insts_on_PF.size(); j++)
		cout << insts_on_PF[j] << " ";
	cout << endl;
#endif
	
	for(int i=0; i < (int)paretofront.size(); i++)
	{
		// For some reason merge() function did not work!!!
		//merge(insts_on_PF.begin(), insts_on_PF.end(), paretofront[i].getInsts().begin(), paretofront[i].getInsts().end(), temp.begin());

		// Combining the instances
		for(int j=0; j < (int)paretofront[i].getInsts().size(); j++)
			insts_on_PF.push_back(paretofront[i].getInsts()[j]);
		
#ifdef DEBUGGING
		cout << "population::attrsInstsOnPF() => (after push_back) insts_on_PF.size() = " << insts_on_PF.size() << endl;
		for(int j=0; j < (int)insts_on_PF.size(); j++)
			cout << insts_on_PF[j] << " ";
		cout << endl;
#endif
		
		// Sorting the instances
		sort(insts_on_PF.begin(), insts_on_PF.end());
		
#ifdef DEBUGGING
		cout << "population::attrsInstsOnPF() => (after sort) insts_on_PF.size() = " << insts_on_PF.size() << endl;
		for(int j=0; j < (int)insts_on_PF.size(); j++)
			cout << insts_on_PF[j] << " ";
		cout << endl;
#endif
		
		// Removing the duplicate instances
		it = unique(insts_on_PF.begin(), insts_on_PF.end());
		insts_on_PF.resize(it - insts_on_PF.begin());
		
#ifdef DEBUGGING
		cout << "population::attrsInstsOnPF() => (after removing duplicates) insts_on_PF.size() = " << insts_on_PF.size() << endl;
		for(int j=0; j < (int)insts_on_PF.size(); j++)
			cout << insts_on_PF[j] << " ";
		cout << endl;
#endif
		
	}
	
#ifdef DEBUGGING
	cout << "population::attrsInstsOnPF() => Final instances of insts_on_PF are (total is " << insts_on_PF.size() << ")= ";
	for(int j=0; j < (int)insts_on_PF.size(); j++)
		cout << insts_on_PF[j] << " ";
	cout << endl;
#endif
	

#ifdef DEBUGGING
	cout << "population::attrsInstsOnPF() exit." << endl;
#endif
}

/***********************************************************************************************************/

void population::finalizeGeneration(int lev, int gen)
{
	// 1. Removes the symbionts with zero references from popHistory.
	// 2. Perform the subsampling (subset selection)
	// 3. Update the mutation rate
	
	// This line is now at the end of both sortOnFronts() functions in sorting.cpp and fastmanydimsort.cpp
	//paretofront = sortedfronts[0];
	
//	cout << "population::finalizeGeneration() => enter." << endl;
	
#ifdef DEBUGGING
	if(lev != 0)
	{
		cout << "population::finalizeGeneration() => numSymbsToUse before removing symbionts with zero references = " << model->_numSymbsToUse << endl;
		
		cout << "population::finalizeGeneration() => symbsToUse before removing symbionts with zero references (" << symbsToUse[lev].size() << "):" << endl;
		for(int i=0; i < (int)symbsToUse[lev].size(); i++)
			cout << symbsToUse[lev][i] << " ";
		cout << endl;
		
		cout << "population::finalizeGeneration() => symbsRefs before removing symbionts with zero references (" << symbsRefs[lev].size() << "):" << endl;
		for(int i=0; i < (int)symbsRefs[lev].size(); i++)
			cout << symbsRefs[lev][i] << " ";
		cout << endl;
		
		cout << "population::finalizeGeneration() => popHistory before removing symbionts with zero references (" << popHistory[lev].size() << "):" << endl;
		for(int i=0; i < (int)symbsToUse[lev].size(); i++)
		{
			cout << "popHistory[" << lev << "][" << symbsToUse[lev][i] << "] = " << endl;
			popHistory[lev][symbsToUse[lev][i]]->print(0);
		}
	}
#endif
	
	
	int tempRefs;
	int tempRefsPrev;
	double pmin = double(1 / (1.1*model->_dataSetDim));
	double pmax = double(1 - (model->_dataSetDim - 1)*pmin);
	
	// Remove the symbionts with zero references
	if(lev != 0)
	{
		
		// Copy symbsToUse to symbsToUsePrev and symbsRefs to symbsRefsPrev locally.
		vector <int> symbsToUsePrev (symbsToUse[lev].begin(), symbsToUse[lev].end());;
		vector <int> symbsRefsPrev (symbsRefs[lev].begin(), symbsRefs[lev].end());
		symbsRefs[lev].clear();
		
#ifdef DEBUGGING
		cout << "population::finalizeGeneration() => symbsToUse (" << symbsToUse[1].size() << "):" << endl;
		for(int i=0; i < (int)symbsToUse[1].size(); i++)
			cout << symbsToUse[1][i] << " ";
		cout << endl;
		
		cout << "population::finalizeGeneration() => symbsRefsPrev (" << symbsRefsPrev.size() << "):" << endl;
		for(int i=0; i < (int)symbsRefsPrev.size(); i++)
			cout << symbsRefsPrev[i] << " ";
		cout << endl;
#endif
		
		for(int i=0, k=0; i < (int)symbsToUse[lev].size(); i++, k++)
		{
			tempRefs = popHistory[lev][symbsToUse[lev][i]]->getRefs();
			symbsRefs[lev].push_back(tempRefs);
			
			// count references of an attribute and compare with it's frequency count in previous generation.
			if(model->_useAttrProbLrng)
			{
				if(i <= symbsToUse[lev].back())
				{
					//tempRefsPrev = symbsRefsPrev[symbsToUse[lev][i]];
					tempRefsPrev = symbsRefsPrev[i];
				}
				else
				{
					tempRefsPrev = 0;
					//cout << "new symbiont getting (previous) reference count = 0" << endl;
				}
				//cout << "population::finalizeGeneration() => symbsToUse[1][" << i << "] = " << symbsToUse[lev][i] << " tempRefs = " << tempRefs << ", tempRefsPrev = " << tempRefsPrev << endl;
				
				// If references to a symbiont has increased
				if(tempRefs > (int)tempRefsPrev)
				{
					vector <int> tempAttrs = popHistory[lev][symbsToUse[lev][i]]->getAttrs();
					
					//cout << "population::finalizeGeneration() => attributes to increase probability: ";
					
					for(int j=0; j < (int)tempAttrs.size(); j++)
					{
						//model->_attrProbs[tempAttrs[j]] += (double) (1.0 / model->_dataSetDim);
	//					model->_attrProbs[tempAttrs[j]] += (double) model->_attrProbLrngRate * (1.0 - model->_attrProbs[tempAttrs[j]]);
						model->_attrProbs[tempAttrs[j]] += (double) model->_attrProbLrngRate * (pmax - model->_attrProbs[tempAttrs[j]]);
						//cout << tempAttrs[j] << ", ";
					}
					//cout << endl;
				}
			}
			
			// If a symbiont is no longer referenced
			if(tempRefs <= 0)
			{
				
#ifdef DEBUGGING
				cout << "population::finalizeGeneration() => The Symbiont with zero refs is:" << endl;
				cout << "popHistory[" << lev << "][" << symbsToUse[lev][i] << "]:" << endl;
				popHistory[lev][symbsToUse[lev][i]]->print(0);
#endif
				
				// Free the memory allocated to the solution which won't be used any more.
				// This is commented since the memory allocation is not done with "new" instruction.
				if(i >= (int) 2*model->_hostPopSize)
				{
					//cout << "ind to delete is " << i << endl;
					
					delete popHistory[lev][symbsToUse[lev][i]];
					
#ifdef DEBUGGING
					cout << "population::finalizeGeneration() => popHistory after removing ONE symbiont with zero references:" << endl;
					for(int i=0; i < (int)symbsToUse[lev].size(); i++)
					{
						cout << "popHistory[" << lev << "][" << symbsToUse[lev][i] << "] = " << endl;
						popHistory[lev][symbsToUse[lev][i]]->print(0);
					}
#endif
				}
				

				// Decrease model->_numSymbToUse.
				model->_numSymbsToUse--;

				// Two ways to remove the solution from being indexed later on (We use the first method):
				// 1. Remove the solution pointer from popHistory 
				// 2. Remove the solution index from symbsToUse
				
				// Method 1:
				// Remove the index to the removed solution from the global variable symbsToUse.
				symbsToUse[lev].erase(symbsToUse[lev].begin() + k);
				symbsRefs[lev].erase(symbsRefs[lev].begin() + k);
				
				// k iterates the symbToUse[lev] vector whereas i iterates the popHistory[lev] vector, 
				// so k should be decreased after each deletion.
				k--;
				i--;
				
				//cout << "i = (after removing) " << i << endl;
				
				// Method 2:
				// Remove the solution pointer from the popHistory global variable.
				//popHistory[lev].erase(popHistory[lev].begin() + i);
				
				// Update the symbsToUse variable by removing the solution index and also 
				// symbsToUse[lev].erase(symbsToUse[lev].begin() + i);
				// symbsRefs[lev].erase(symbsToUse[lev].begin() + i);
				
				// And also reducing one from other cells.
				//for(int j=i; j < (int)symbsToUse[lev].size(); j++)
					//symbsToUse[lev][j]--;
				
				// Adjust the counter/iterator to point at the current solution after deleting a solution.
				//i--;
				
#ifdef DEBUGGING
				cout << "population::finalizeGeneration() => numSymbsToUse = " << model->_numSymbsToUse << endl;
				
				cout << "population::finalizeGeneration() => symbsToUse (" << symbsToUse[lev].size() << ") after deleting ONE symbiont:" << endl;
				for(int j=0; j < (int)symbsToUse[lev].size(); j++)
					cout << symbsToUse[lev][j] << " ";
				cout << endl;
				
				cout << "population::finalizeGeneration() => popHistory (" << symbsToUse[lev].size() << ") after deleting ONE symbiont:" << endl;
				for(int j=0; j < (int)symbsToUse[lev].size(); j++)
				{
					cout << "popHistory[" << lev << "][" << symbsToUse[lev][j] << "] = " << endl;
					popHistory[lev][symbsToUse[lev][j]]->print(0);
				}
#endif
				
			}
			
		}
		
		if(model->_useAttrProbLrng)
		{
			
#ifdef DEBUGGING
			cout << "population::finalizeGeneration() => Attribute Probabilities BEFORE normalization:" << endl;
			for(int j=0; j < (int)model->_attrProbs.size(); j++)
				cout << model->_attrProbs[j] << ", ";
			
			cout << endl;
#endif
			
			// normalize attrProbs vector
			double tempSum = 0;
			
			for(int j=0; j < (int)model->_attrProbs.size(); j++)
				tempSum += model->_attrProbs[j];
			
			//cout<< "tempSum = " << tempSum << endl;
			
			for(int j=0; j < (int)model->_attrProbs.size(); j++)
				model->_attrProbs[j] /= tempSum;
			
#ifdef DEBUGGING
			cout << "population::finalizeGeneration() => Attribute Probabilities AFTER normalization:" << endl;
			for(int j=0; j < (int)model->_attrProbs.size(); j++)
				cout << model->_attrProbs[j] << ", ";
			
			cout << endl;
#endif
			
		}
		
	}
	
	
	// Perform the subsampling
	/* Do the RSS subsampling if _doSubsampling is set, otherwise use all points (i.e. all points are active) */
	if(model->_doSubsampling == 1)
		model->RSS();
	else
	{
#ifdef DEBUGGING
		cout << "population::finalizeGeneration() => assigning all instances to _activePoints" << endl;
#endif
		model->_pointPopSize = model->_dataSetSize;
		
		// We don't need to re-initialize the whole data set after each generation if the subsampling is not performed.
		/*
		model->_activePoints.clear();
		model->_activePoints.resize(model->_dataSetSize);
		
		for(int i=0; i < (int) model->_dataSetSize; i++)
		{
			model->_activePoints[i] = i;
		}
		*/
	}
	
	// Calculate the distances of symbionts in use (determined by 'symbToUse') from _activePoints at the end of each generation.
	if(level==0)
		cout << "THIS PIECE OF CODE YET TO BE DONE!!!" << endl;
//		for(int i=0; i < model->_hostPopSize; i++)
//			if(model->_doConnectivity)
//				popHistory[lev+1][i]->calcDistances();
	else if(level==1)
	{
		// assigning _activePoints with all points for post-training performance measurements.
#ifdef DEBUGGING
		cout << "population::finalizeGeneration() => model->_activePoints: ";
		for(int i=0; i < model->_pointPopSize; i++)
			cout << model->_activePoints[i] << " ";
		cout << endl;
#endif
		
		for(int i=0; i < (int)symbsToUse[1].size(); i++)
//			if(model->_doConnectivity)
				popHistory[1][symbsToUse[1][i]]->calcDistances();
	}

	
	// Update variation operators probability rate
	model->_pc = model->_pc_final - (gen * ((model->_pc_final - model->_pc_initial) / (model->_numGenerations-1)));
	model->_pm = model->_pm_final - (gen * ((model->_pm_final - model->_pm_initial) / (model->_numGenerations-1)));
	
//#ifdef DEBUGGING
	cout << "population::finalizeGeneration() => model->_pc_initial = " << model->_pc_initial << endl;
	cout << "population::finalizeGeneration() => model->_pc_final = " << model->_pc_final << endl;
	cout << "population::finalizeGeneration() => model->_pc = " << model->_pc << " at generation " << gen << endl;
	cout << "population::finalizeGeneration() => model->_pm_initial = " << model->_pm_initial << endl;
	cout << "population::finalizeGeneration() => model->_pm_final = " << model->_pm_final << endl;
	cout << "population::finalizeGeneration() => model->_pm = " << model->_pm << " at generation " << gen << endl;
//#endif
	
	
//#ifdef DEBUGGING
	if(lev != 0)
	{
		cout << "population::finalizeGeneration() => numSymbsToUse after removing symbionts with zero references = " << model->_numSymbsToUse << endl;
		
		cout << "population::finalizeGeneration() => symbsToUse after removing symbionts with zero references (" << symbsToUse[lev].size() << "): " << endl;
		for(int i=0; i < (int)symbsToUse[lev].size(); i++)
			cout << symbsToUse[lev][i] << " ";
		cout << endl;
		
		cout << "population::finalizeGeneration() => symbsRefs after removing symbionts with zero references (" << symbsRefs[lev].size() << "):" << endl;
		for(int i=0; i < (int)symbsRefs[lev].size(); i++)
			cout << symbsRefs[lev][i] << " ";
		cout << endl;
		
/*		cout << "population::finalizeGeneration() => popHistory after removing symbionts with zero references (" << popHistory[lev].size() << "):" << endl;
		for(int i=0; i < (int)symbsToUse[lev].size(); i++)
		{
			cout << "popHistory[" << lev << "][" << symbsToUse[lev][i] << "] = " << endl;
			popHistory[lev][symbsToUse[lev][i]]->print(0);
		}
*/
	}
//#endif

//	cout << "population::finalizeGeneration() => exit." << endl;
}

/***********************************************************************************************************/

void population::finalizeLevel(int lev)
{
	// 1. Store pointers to solutions that will be used for next levels. 
	// 2. Calculate the distances of symbionts in use (determined by 'symbToUse') from ALL points at the end of each level.
	
	cout << "population::finalizeLevel() enter." << endl;
	
	// The first position of popHistory and symbsToUse are empty placeholders.
	if(lev == 0)
	{
		popHistory.push_back(vector <solution *>());
		symbsToUse.push_back(vector <int>());
		symbsRefs.push_back(vector <int>());
//		symbsRefsPrev.push_back(vector <int>());
	}
	
	// Not sure if this is how we do it??? Use this later on in gene::randomize() and solution::evaluate()
	popHistory.push_back(vector <solution *>());
	
	// Store pointers to all solutions to be used in next levels. 
	for(int i=0; i < (int) 2*model->_hostPopSize; i++)
	{
		
#ifdef DEBUGGING
		cout << "population::finalizeLevel() => popsorted[" << i << "] = " << endl;
		popsorted[i]->print(0);
#endif
		
		popHistory[lev+1].push_back(popsorted[i]);
		
#ifdef DEBUGGING
		cout << "population::finalizeLevel() => popHistory[" << lev+1 << "][" << i << "] = " << endl;
		popHistory[lev+1][i]->print(0);
#endif
		
	}
	
	// This is to store all solutions instead of only some of it.
	//popHistory[lev+1] = popsorted;
	
	// Another way to store all of popsorted using a STL command.
	//(popHistory.back())->insert((popHistory.back())->begin(), popsorted.begin(), popsorted.end());
	
	// Not sure if this is used when using a for loop in main() function to create population of different layers.
	//popsorted.clear();
	
	// Indexes of symbionts that are active and can be indexed by hosts
	symbsToUse.push_back(vector <int>());
	// Stores number of references to a symbiont
	symbsRefs.push_back(vector <int>());
	//symbsRefsPrev.push_back(vector <int>());
	
	
	// Indexes to solutions that can be used in next level.
	if(model->_symbsToUse == 0)
	{
		model->_numSymbsToUse = (int) 2*model->_hostPopSize;
		
		// If all solutions are used, create a vector from 0 to hostPopSize and pass it to next level.
		for(int i=0; i < (int) 2*model->_hostPopSize; i++)
		{
			symbsToUse[lev+1].push_back(i);
//			cout << "symb " << symbsToUse[lev+1][i] << " references: " << popHistory[lev+1][symbsToUse[lev+1][i]]->getRefs() << endl;
			symbsRefs[lev+1].push_back(popHistory[lev+1][symbsToUse[lev+1][i]]->getRefs());
//			symbsRefsPrev[lev+1].push_back(popHistory[lev+1][[symbsToUse[lev+1][i]]].getRefs());
		}
	}
	else if(model->_symbsToUse == 1)
	{
		model->_numSymbsToUse = 10; // model->_numUniqueSymbs = paretofront.get_uniqueIndexes().size())
	}
	else
	{
		model->_numSymbsToUse = (int) paretofront.size(); // or (int) sortedfronts[0].size()
		
		// If all PF set is used, (and we know that they are sorted on popsorted), create a vector 
		// from 0 to paretofront.size() and pass to next level.
		for(int i=0; i < (int)paretofront.size(); i++)
			symbsToUse[lev+1].push_back(i);
	}
	
	
	// Calculate the distances of symbionts in use (determined by 'symbToUse') from ALL points at the end of each level.
	if(level==0)
		for(int i=0; i < (int) 2*model->_hostPopSize; i++)
//			if(model->_doConnectivity)
				popHistory[lev+1][i]->calcDistances();
	else if(level==1)
	{
		// assigning _activePoints with all points for post-training performance measurements.
		model->_activePoints.clear();
//		model->_activePoints.resize(model->_dataSetSize);
		for(int i=0; i < model->_dataSetSize; i++)
			model->_activePoints.push_back(i);
		
		model->_pointPopSize = model->_activePoints.size();
		
#ifdef DEBUGGING
		cout << "population::finalizeLevel() => model->_activePoints: ";
		for(int i=0; i < model->_dataSetSize; i++)
			cout << model->_activePoints[i] << " ";
		cout << endl;
#endif
		
		for(int i=0; i < (int)symbsToUse[1].size(); i++)
//			if(model->_doConnectivity)
				popHistory[1][symbsToUse[1][i]]->calcDistances();
	}
	
	// Replacing the 3 working instructions above to include all levels instead checking each level manualy
//	for(int lev1=0; lev1 < model->_lev; lev1++)
//		for(int i=0; i < (int)symbsToUse[lev1].size(); i++)
//			if(model->_doConnectivity)
//				popHistory[lev1][symbsToUse[lev1][i]]->calcDistances();


//#ifdef DEBUGGING
	cout << "population::finalizeLevel() => popHistory has " << popHistory.size() << " levels now." << endl << endl;
	for(int i=0; i < (int)popHistory.size(); i++)
		cout << "population::finalizeLevel() => popHistory[" << i << "].size() = " << popHistory[i].size() << endl;
	cout << endl;
	
	for(int i=0; i < (int)symbsToUse.size(); i++)
	{
		cout << "population::finalizeLevel() => symbsToUse[" << i << "] has " << symbsToUse[i].size() << " indexes: ";
		for(int j=0; j < (int)symbsToUse[i].size(); j++)
			cout << symbsToUse[i][j] << " ";
		cout << endl;
	}
	cout << endl;
	
	for(int i=0; i < (int)symbsRefs.size(); i++)
	{
		cout << "population::finalizeLevel() => symbsRefs[" << i << "] has " << symbsRefs[i].size() << " indexes: ";
		for(int j=0; j < (int)symbsRefs[i].size(); j++)
			cout << symbsRefs[i][j] << " ";
		cout << endl;
	}
	cout << endl;
	
	for(int i=0; i < (int)popHistory.size(); i++)
	{
		cout << endl << "population::finalizeLevel() => popHistory[" << i << "] has " << symbsToUse[i].size() << " solutions: " << endl;
		for(int j=0; j < (int)symbsToUse[i].size(); j++)
		{
			cout << "popHistory[" << i << "][" << symbsToUse[i][j] << "]: " << endl;
			popHistory[i][symbsToUse[i][j]]->print(0);
		}
	}
	cout << endl;
//#endif
	
	cout << "population::finalizeLevel() exit." << endl;
} 

/***********************************************************************************************************/

int population::bestOnFitness()
{
	double bestFit = 1000000.0;
	int bestIdx = -1;
	for(size_t ind=0; ind < pop.size(); ind++)
	{
		//cout << "Printing ind (" << ind << ") " << endl;
		//pop[ind].print(0);
		if(pop[ind].getFitness() < bestFit)
		{
			bestFit = pop[ind].getFitness();
			//cout << "The ind's fitness is = " << endl;
			bestIdx = ind;
		}
		
	}
	//cout << "and the best ind is " << bestIdx << endl;
	//cout << "and it's fitness is " << pop[bestIdx].get_fitness() << endl;;

	return bestIdx;
}

/***********************************************************************************************************/

void population::calcCrowdingDistances()
{
	for(vector<front>::size_type i=0; i < sortedfronts.size(); i++)
		sortedfronts[i].calcCrowdingDistances();
}

/***********************************************************************************************************/

/*
front population::getParetoFront()
{
	paretofront = sortedfronts[0];
	//cout << "population::paretofront.size() = " << paretofront.size() << endl;
	
	// This is not required since uniqueSolutionsIndexes() is called from main()
	paretofront.uniqueSolutionsIndexes();
	
	return(sortedfronts[0]);
}
*/

/***********************************************************************************************************/

