
//#include "esc.hpp"
#include "types.hpp"
#include "misc.hpp"

#include <iostream>
#include <iomanip>
#include <vector>
#include <math.h>
#include <algorithm>

using namespace std;


// These five are used when symbionts have fitness evaluations
//extern vector <solution *> subspaceClusters;
//extern vector <int> attrsIndexed;
//extern vector <int> instsIndexed;
//extern int numAttrsIndexed;
//extern int numInstsIndexed;

extern esc *model;
extern vector < vector <solution *> > popHistory;

// sumDistances is used for shared fitness evaluation
extern vector <double> sumDistances;


/***********************************************************************************************************/

//double solution::static_sum = 0;

/***********************************************************************************************************/

solution::solution()//int l
//: level(l)
{
#ifdef DEBUGGING
	cout << "solution::solution() enter." << endl;
#endif
	
	level = model->_lev;
	
	sol = gene(level);
	//sol = gene();
	
	// Randomize can be moved to here.
	//randomize();
	
	// objectives are allocated at this point.
	if(level == 0 & model->_numObjectives0 == 0)
	{	
		//cout << "solution::solution(*solution) => Level0 -> No objectives!" << endl;
		objectives = vector <double> (2,-1);// intialize objectives vector with two -1 values
	}
	else if(level == 0 & level < model->_numLevels != 0)
		objectives = vector <double> (model->_numObjectives0);
	else if(level != 0 & level < model->_numLevels)
		objectives = vector <double> (model->_numObjectives1);
	else
		die(__FILE__, __FUNCTION__, __LINE__, "Wrong level value!");
	
	fitness = -1.0;
	
	frontnumber = 0;
	crowdingdistance = 0;
	
	refs = 0;
	attrPerSymb = 0.0;

	// Calculate the distance between this solution and each point.
	//if(level==0)
	//	calcDistances();
	
#ifdef DEBUGGING
	cout << "solution::solution() exit." << endl;
#endif
};

/***********************************************************************************************************/

solution::solution(solution *temp)//int l
//: level(l)
{
#ifdef DEBUGGING
	cout << "solution::solution(solution &) enter." << endl;
#endif
	
#ifdef DEBUGGING
	cout << "solution::solution(solution &) => The individual to be copied from: " << endl;
	temp->print(0);
#endif
	
	level = temp->getLevel();
	
	sol = gene(level);
	//sol = gene();

	sol.index1 = temp->getGene().index1;
	sol.index2 = temp->getGene().index2;

#ifdef DEBUGGING
	cout << "solution::solution(solution &) => parent solution index1:" << endl;
	for(int pha=0; pha < (int)temp->getGene().index1.size(); pha++)
		cout << temp->getGene().index1[pha] << ", ";
	cout << endl;
	cout << "solution::solution(solution &) => cloned solution index1:" << endl;
	for(int pha=0; pha < (int)sol.index1.size(); pha++)
		cout << sol.index1[pha] << ", ";
	cout << endl;
#endif
	
	// objectives are allocated at this point.
	if(level == 0 & model->_numObjectives0 == 0)
	{	
		//cout << "solution::solution(*solution) => Level0 -> No objectives!" << endl;
		objectives = vector <double> (2,-1);// intialize objectives vector with two -1 values
	}
	else if(level == 0 & level < model->_numLevels != 0)
		objectives = vector <double> (model->_numObjectives0);
	else if(level != 0 & level < model->_numLevels)
		objectives = vector <double> (model->_numObjectives1);
	else
		die(__FILE__, __FUNCTION__, __LINE__, "Wrong level value!");
	
	
	fitness = -1.0;
	
	frontnumber = 0;
	crowdingdistance = 0;
	
	refs = 0;
	
	attrs = temp->getAttrs();
	attrPerSymb = (double)attrs.size();

	// Calculate the distance between this solution and each point.
	//if(level==0)
	//	calcDistances();
	
#ifdef DEBUGGING
	cout << "solution::solution(solution &) => The new cloned individual: " << endl;
	temp->print(0);
#endif
	
#ifdef DEBUGGING
	cout << "solution::solution(solution &) exit." << endl;
#endif
};

/***********************************************************************************************************/

solution::~solution()
{
	//cout << "solution::~solution() => Level " << level << ", default solution destructor enters!" << endl;
}

/***********************************************************************************************************/

int solution::print(int t=0)
{
#ifdef DEBUGGING
	cout << "solution::print() enter." << endl;
#endif
	
	if(t == 0)
	{
		sol.print(t);

		cout << "level = " << level << ", Objectives:     ";
		if(level == 0 && model->_numObjectives == 0)
			cout << "No objectives for this individual!";
		else 
			for(int i=0; i < model->_numObjectives; i++)
			{
				//cout << "Obj" << i << " = " << objectives[i] << "    ";
				cout << objectives[i] << "    ";
			}
		cout << endl;
		
		cout << "FrontNumber : " << frontnumber << " CrowdingDistance : " << crowdingdistance << " References : " << refs << endl;	
		
		cout << "Attributes Accessed (" << attrs.size() << ", " << attrPerSymb << "): ";
		for(int i=0; i < (int)attrs.size(); i++)
			cout << attrs[i] << " ";
		cout << endl;
		/*
		cout << "Distances: ";
		for(int i=0; i < (int)distances.size(); i++)
			cout << distances[i] << " ";
		cout << endl;
		*/
	}
	
	else if(t == 1)
	{
		//cout << "Objectives:     ";
		for(int i=0; i < model->_numObjectives; i++)
		{
			//cout << "Obj" << i << " = " << objectives[i] << "    ";
			cout << objectives[i] << "    ";
		}
		cout << endl;
	}
		
#ifdef DEBUGGING
	cout << "solution::print() exit." << endl;
#endif
	
	return size();
}

/***********************************************************************************************************/

void solution::randomize()
{
#ifdef DEBUGGING
	cout << "solution::randomize() enter." << endl;
#endif
	
	sol.randomize();

#ifdef DEBUGGING
	cout << "solution::randomize() exit." << endl;
#endif
}

/***********************************************************************************************************/

bool solution::operator== (const solution& s1) const
{
	// Checks if two solutions have identical genes
	
	return (sol == s1.sol);
	//return sol.equals(s1.sol, level);
}

/***********************************************************************************************************/

vector<int> solution::evaluate(int gen)
{

#ifdef DEBUGGING
	cout << "solution::evaluate() enter." << endl;
#endif

	
	if(level == 0 && model->_numObjectives == 0)
	{
		return vector <int> (2,-1);
	}
	else if(level == 0 && model->_numObjectives != 0)
	{
		// Four objectives have been implemented for the first level individuals:
		// 1. cluster compactness (to be minimized),  2. exemplar count (to be maximized),
		// 3. exemplars shared between clusters (to be minimized), and 4. individual size (to be minimized).
		
		int attr, clus;
		double sum = 0;
		int num = 0;
		int numVisited = 0;
//		int notIndexedByPareto = 0;
		vector <int> visited(model->_dataSetSize, 0);
		vector <int> intersect;
		vector <int> newList;
		
#ifdef DEBUGGING
		cout << "solution::evaluate() => This solution has " << sol.index1.size() << " genes." << endl;
		sol.print(0);
#endif
		
		// Initializing the intersect list with instances belonging to the first 1D cluster
		attr = sol.index1[0];
		clus = sol.index2[0];
		
		for(int i=0; i < model->_dataSetSize; i++)
		{
			if(model->_attrWiseClusLabels[i][attr] == clus)
			{
				intersect.push_back(i);
			}
		}

/*
		// If compactness is calculated only over intersecting instances this piece of code can be used 
		// instead of the more expensive piece that follows.
		
 //#ifdef DEBUGGING
		cout << "index1[" << 0 << "](attr) = " << attr << ", index2[" << 0 << "](clus) = " << clus << endl;
		cout << "intersect list = ";
		for(int m=0; m < (int)intersect.size(); m++)
			cout << intersect[m] << " ";
		cout << endl;
 //#endif
 
		// Starting from the second index (i=1) since the first is used for initialization.
		for(int i=1; i < (int)sol.index1.size(); i++)
		{
			// If intersect is empty jump out of this loop.
			if(intersect.size() == 0)
				break;
			
			attr = sol.index1[i];
			clus = sol.index2[i];
			
			int k;
			
			for(int j=0; j < (int)intersect.size(); j++)
			{
				k = intersect[j];
				if(model->_attrWiseClusLabels[k][attr] != clus)
				{
					intersect.erase(intersect.begin()+j);
					j--;
				}
			}
			
		}
		
		// Calculating the sum of distances from intersect points to the cluster centroid.
		for(int i=0; i < (int)sol.index1.size(); i++)
		{
			attr = sol.index1[i];
			clus = sol.index2[i];
			int k;
			
			for(int j=0; j < (int)intersect.size(); j++)
			{
				k = intersect[j];
				sum += pow((model->_attrWiseClusCent[attr][clus] - model->_pointSet[k][attr]), 2);
			}
		}
*/
		
		
		for(int i=0; i < (int)sol.size(); i++)
		{
			attr = sol.index1[i];
			clus = sol.index2[i];
			
			newList.clear();

			for(int j=0; j < model->_dataSetSize; j++)
			{
				//cout << "model->_attrWiseClusLabels[" << attr << "][" << j << "] = " << model->_attrWiseClusLabels[attr][j] << endl;
				if(model->_attrWiseClusLabels[j][attr] == clus)
				{
					//cout << "instance " << j << " needs model->_attrWiseClusCent[" << attr << "][" << clus << "] (in evaluate()) = " << model->_attrWiseClusCent[attr][clus] << endl;
					newList.push_back(j);
					sum += pow((model->_attrWiseClusCent[attr][clus] - model->_pointSet[j][attr]), 2);
					num++;
					if(visited.at(j) == 0)
					{
						visited[j] = 1;
						numVisited++;
					}
				}
			}
			
#ifdef DEBUGGING
			cout << "The new intersect list with index1[" << i << "](attr) = " << attr << ", index2[" << i << "](clus) = " << clus << " => ";
			for(int m=0; m < (int)newList.size(); m++)
				cout << newList[m] << " ";
			cout << endl;
#endif
			
			if(i != 0) // index1[1], index1[2], ...
			{
				for(size_t k=0; k < intersect.size(); k++)
				{
					int flag = 0;
					// We don't need to start the second list from index 0, but instead we can start it from 
					// last index visited in previous iteration of the outer loop (for(i))
					for(size_t j=0; j < newList.size(); j++)
					{
						if(intersect.at(k) == newList.at(j))
						{
							flag = 1;
							break;
						}
					}
					if(flag == 0)
					{
						intersect.erase(intersect.begin()+k);
						k--;
					}
				}
				
#ifdef DEBUGGING
				cout << "Final intersect list after intersection with the list by index1[" << i << "](attr) = " << attr << ", index2[" << i << "](clus) = " << clus << " => ";
				for(int m=0; m < (int)intersect.size(); m++)
					cout << intersect[m] << " ";
				cout << endl;
#endif
				
			}
			
		}
		//cout << "index1.size() is: " << sol.index1.size() << endl;
		//cout << "intersect.size() is: " << intersect.size() << endl;
		//cout << "instsIndexed.size() is: " << numInstsIndexed << endl;
/*
		if(gen >= 0)
		{
			// intersection is finalized here
			for(int m=0; m < (int)intersect.size(); m++)
			{
				if(instsIndexed[intersect[m]] == 0)
				{
					//cout << "instance not indexed: " << intersect[m] << endl;
					notIndexedByPareto++;
				}
			}
		}
		else
		{
			notIndexedByPareto = 0;
		}
		//cout << "number of instances not indexed is: " << notIndexedByPareto << endl;
 */		

		// Assign values to the vectors storing the attributes and instances indexed by the solution
		attrs = sol.index1;
		insts = intersect;
		attrPerSymb = (double)attrs.size();
		
		// Remember the order of objectives is NOT important.
		
		// Normalized distortion values, division is used to normalize the distortion value, (minimize)
		objectives[0] = (double)sum / (double)sol.size();
		
		// Count of intersection of instances indexed by different (attr,clus) pairs of an individual (maximize). 
		// Instead of maximizing "intersect.size()" we minimize "model->_dataSetSize - intersect.size()"
		objectives[1] = (double)model->_dataSetSize - (double)intersect.size();
		
#ifdef DEBUGGING
		cout << "solution::evaluate() => Objectives for this individual: ";
		for(int i=0; i < model->_numObjectives; i++)
			cout << objectives[i] << " ";
		cout << endl;
#endif
		
		// Fitness value for solutions with level == 0 
		fitness = objectives[0];
		
#ifdef DEBUGGING
		cout << "solution::evaluate() exit." << endl;
#endif
		
		return intersect;
	}
	
	else if(level > 0 && level < model->_numLevels)
	{
		// Four objectives have been implemented for higher level individuals:
		// 1. compactness/distortion (to be minimized),  2. connectivity (to be minimized),
		// 3. cluster count (to be minimized), and 4. attribute count (to be minimized).
		
		double sum = 0;
		vector <int> subClstInstCnt((int)sol.index1.size(), 0);
		int subClstOffset = -1;
		int numUniqueAttrs = 0;
		int numOverallAttrs = 0;
		vector <int> attrsAccessed;
		attrsAccessed.clear();
		vector <int> indices;
		indices.clear();
		if(model->_labelsProvided || model->_doConnectivity)
			indices.resize(model->_dataSetSize, -1);
		else
			indices.resize(2, -1);
		
		// Reset number of instances assigned to each subspace cluster to 0
		sol.index2.clear();
		sol.index2.resize(sol.size(), 0);
		
#ifdef DEBUGGING
		cout << "solution::evaluate() => the host to be evaluated (" << size() << "):" << endl;
		print(0);
#endif
		
		vector <int> tempvec;
		vector <int> tempvec2;
		
		// Determining the attributes indexed by a host.
		for(int i=0; i < size(); i++)
		{
			// popHistory[level][index1[i]] is the solution pointer poiting to the i'th symbiont of current host.
			
			tempvec.clear();
			
			tempvec = (popHistory[level][sol.index1[i]])->getAttrs();
			
#ifdef DEBUGGING
			cout << "the host's " << i << "th symbiont (): " << sol.index1[i] << endl;
			popHistory[level][sol.index1[i]]->print(0);
#endif
			
#ifdef DEBUGGING
			cout << "symbiont " << i << "'s tempvec (" << tempvec.size() << "): ";
			for(int j=0; j < (int)tempvec.size(); j++)
				cout << tempvec[j] << " ";
			cout << endl;
#endif
			
			attrsAccessed.resize((int) (tempvec.size() + tempvec2.size()));
			
			merge(tempvec2.begin(), tempvec2.end(), tempvec.begin(), tempvec.end(), attrsAccessed.begin());
			
#ifdef DEBUGGING
			cout << "attrsAccessed after merge (" << attrsAccessed.size() << ") = ";
			for(int j=0; j < (int)attrsAccessed.size(); j++)
				cout << attrsAccessed[j] << " ";
			cout << endl;
#endif
			
			tempvec2.clear();
			tempvec2 = attrsAccessed;
			
#ifdef DEBUGGING
			cout << "tempvec2 (" << tempvec2.size() << "): ";
			for(int j=0; j < (int)tempvec2.size(); j++)
				cout << tempvec2[j] << " ";
			cout << endl;
#endif
			
		}

#ifdef DEBUGGING
		cout << "solution::evaluate() => attrsAccessed (" << attrsAccessed.size() << "):" << endl;
		for(int i=0; i < (int)attrsAccessed.size(); i++)
			cout << attrsAccessed[i] << " ";
		cout << endl;
#endif
		
		// Number of attributes accessed by a host.
		numOverallAttrs = (int) attrsAccessed.size();
		
		// attrs is simply all attributes accessed by the host.
		attrs = attrsAccessed;
		
		// Average number of attrs per symbionts of a host.
		attrPerSymb = (double)attrs.size() / (double)size();
		
		// Maybe a sort function is required here.
		//sort(attrsAccessed.begin(), attrsAccessed.end());
		
		// Number of unique attributes accessed by a host.
		vector <int>::iterator it;
		it = unique(attrs.begin(), attrs.end());
		attrs.resize(it - attrs.begin());
		numUniqueAttrs = (int) attrs.size();
		
#ifdef DEBUGGING
		cout << "solution::evaluate() => unique Attrs Accessed (" << attrs.size() << "):" << endl;
		for(int i=0; i < (int)attrs.size(); i++)
			cout << attrs[i] << " ";
		cout << endl;
#endif
		

		// Replacement code for Compactness calculation (look at previous versions of the code)
		// This piece of code only calculates the compactness, no sharing involved.
		
//		cout << "solution::evaluate() => gen = " << gen << " and model->_numGenerations = " << model->_numGenerations << endl;
		
//		cout << "solution::evaluate() => tempdist = ";
		
		if(gen == model->_numGenerations || !model->_doSubsampling) // subsampling condition added after changes in calcDistances()
		{
			for(int i=0; i < model->_dataSetSize; i++)
			{
				double dist = model->_infinity;
				subClstOffset = 0;
				int index = -1;
				double tempdist = 0;
				
				for(int j=0; j < (int)size(); j++)
				{
					tempdist = popHistory[level][sol.index1[j]]->getDistances()[i];
					
					if(tempdist == -1)
						cout << "solution::evaluate() => distance from instance " << i << " equals -1 (case 1)" << endl;
					else if(tempdist > 0 && tempdist < dist)
					{
						dist = tempdist;
						subClstOffset = j;
						index = sol.index1[j];
					}
					
				}
				
				// Storing the index of the symbiont which clusters i'th instance.
				if(model->_labelsProvided || model->_doConnectivity)
					indices[i] = index;
					//indices[model->_activePoints[i]] = index;
				
				// Adding the partial distance to the overall sum (objective 0 of the solution.)
				sum += dist;
				
				// Increase the number of instances assigned to the subClstOffset's symbiont.
				subClstInstCnt[subClstOffset]++;
				sol.index2[subClstOffset]++;
			}
//			cout << endl;
		}
		else
		{
			for(int i=0; i < model->_pointPopSize; i++)
			{
				double dist = model->_infinity;
				subClstOffset = 0;
				int index = -1;
				double tempdist = 0;
				
				for(int j=0; j < (int)size(); j++)
				{
					// distances is a vector with size of _pointPopSize and not _dataSetSize, 
					// therefore we use "getDistances()[i]" instead of "getDistances()[model->_activePoints[i]]"
					tempdist = popHistory[level][sol.index1[j]]->getDistances()[i];
				
//					cout << tempdist << " ";
					
					if(tempdist == -1)
						cout << "solution::evaluate() => distance from instance " << model->_activePoints[i] << " equals -1 (case 2)" << endl;
					else if(tempdist > 0 && tempdist < dist)
					{
						dist = tempdist;
						subClstOffset = j;
						index = sol.index1[j];
					}
					
				}
				
				// Storing the index of the symbiont which clusters i'th instance.
				if(model->_labelsProvided || model->_doConnectivity)
					//indices[i] = index;
					indices[model->_activePoints[i]] = index;
				
				// Adding the partial distance to the overall sum (objective 0 of the solution.)
				sum += dist;
				
				// Increase the number of instances assigned to the subClstOffset's symbiont.
				subClstInstCnt[subClstOffset]++;
				sol.index2[subClstOffset]++;
			}
		}

/*
		// Replace the above code to calculate SHARED fitness (compactness)
		for(int i=0; i < model->_dataSetSize; i++)
		{
			double dist = model->_infinity;
			double tempdist = 0;
			subClstOffset = 0;
			int index = -1;
			
			for(int j=0; j < (int)size(); j++)
			{
				tempdist = popHistory[level][sol.index1[j]]->getDistances()[i];
				
				if(tempdist < dist)
				{
					dist = tempdist;
					subClstOffset = j;
					index = sol.index1[j];
				}
				
			}
			
			// Stroing the index of the symbiont which clusters i'th instance.
			if(model->_labelsProvided || model->_doConnectivity)
				indices[i] = index;
			
			// Adding the partial distance to the overall sum (objective 0 of the solution.)
			sum += pow((dist / sumDistances[i]), 3);
			
			// Increase the number of instances assigned to the subClstOffset's symbiont.
			subClstInstCnt[subClstOffset]++;
			sol.index2[subClstOffset]++;
		}
*/
		
		// Check if any of subspace clusters have 0 instances assigned and apply gene::pruneSymbsWithZeroInsts()
		for(int j=0; j < (int)sol.size(); j++)
		{
			if(sol.index2[j]==0)
			{
				//cout << "solution::evaluate() => FOUND HOST THAT INDEXES SYMBIONTS WITH ZERO INSTANCES AFTER CALCULATING COMPACTNESS!!!" << endl;
				
				// We should re-evaluate the solution after pruning because cluster count and attribute count has definitely changed.
				//sol.pruneSymbsWithZeroInsts();
				//evaluate(gen);
				//break;
				// Or instead of this whole loop put:
				//if(sol.pruneSymbsWithZeroInsts())
				//	evaluate(gen);
			}
		}
		
		
		// Connectivity calculation of the host
		double conn = 0.0;
		if(model->_doConnectivity)
		{
			// Print the host for which connectivity is to be calculated.
			//cout << "solution::evaluate()->connectivity => print the host for which connectivity is to be calculated: " << endl;
			//print(0);
			
			//int nn_size = max((int)0.01*model->_dataSetSize+1, 10);
			int nn_size = 10;
			// nn is a map with the key value being the distance between instance model->_pointSet[i] and instance j
			// and the mapped value is j.
			map < double, long > nn;
			map < double, long >::iterator it;
			map < double, long >::iterator it2;
			
			gene tempgene;
			
			// If consider_all_instances=0, only _activePoint are considered for determining Nearest neighbours of an instance,
			// If consider_all_instances=1, all data set instances are considered for determining Nearest neighbours of an instance,
			int consider_all_instances = 1;
			
			int tempLoopCounter;
			
			if(consider_all_instances == 1)
				tempLoopCounter = model->_dataSetSize;
			else if(consider_all_instances == 0)
				tempLoopCounter = model->_pointPopSize;
			else
				die(__FILE__, __FUNCTION__, __LINE__, "Wrong consider_all_instanecs value!");
			
			
			for(int i=0; i < model->_pointPopSize; i++)
			{
				nn.clear();
				
				tempgene = popHistory[level][indices[model->_activePoints[i]]]->getGene();
				
				//cout << "instance " << i << " is " << model->_activePoints[i] << " and the Symb associated is " << indices[model->_activePoints[i]] << " which is: " << endl;
				//popHistory[1][indices[model->_activePoints[i]]]->print(0);
				
				// calculating the distance between instance model->_pointSet[i] and other instances
				for(int j=0; j < tempLoopCounter; j++)
				{
					
					//cout << "distance between instance " << model->_activePoints[i] << " and " << model->_activePoints[j] << " is ";
					double tempdist = 0.0;
					for(int k=0; k < (int)tempgene.index1.size(); k++)
					{
						int attr = tempgene.index1[k];
						////cout << "attr = " << attr << ", ";
						////cout << "point1 = " << model->_pointSet[model->_activePoints[i]][attr] << ", point2 = " << model->_pointSet[model->_activePoints[j]][attr] << ", ";
						
						if(consider_all_instances == 1)
							tempdist += pow((model->_pointSet[model->_activePoints[i]][attr] - model->_pointSet[j][attr]), 2);
						else if(consider_all_instances == 0)
							tempdist += pow((model->_pointSet[model->_activePoints[i]][attr] - model->_pointSet[model->_activePoints[j]][attr]), 2);
						else
							die(__FILE__, __FUNCTION__, __LINE__, "Wrong consider_all_instanecs value!");
						
						////cout << " tempdist = " << tempdist << endl;
					}
					//cout << tempdist << endl;
					
					//cout << "nn.size() before insert = " << nn.size() << " and nn = ";
					//for(map < double, long >::iterator it_temp=nn.begin(); it_temp != nn.end(); it_temp++)
					//	cout << it_temp->first << ", " << it_temp->second << "; ";
					//cout << endl;
						
					if((int)nn.size() < nn_size)
					{
						if(consider_all_instances == 1)
							nn.insert(map < double, long > :: value_type(tempdist, j));
						else if(consider_all_instances == 0)
							nn.insert(map < double, long > :: value_type(tempdist, model->_activePoints[j]));
						else
							die(__FILE__, __FUNCTION__, __LINE__, "Wrong consider_all_instanecs value!");
					}
					else if(tempdist < (--it2)->first)
					{
						it = nn.end();
						nn.erase(--it);
						
						if(consider_all_instances == 1)
							nn.insert(map < double, long > :: value_type(tempdist, j));
						else if(consider_all_instances == 0)
							nn.insert(map < double, long > :: value_type(tempdist, model->_activePoints[j]));
						else
							die(__FILE__, __FUNCTION__, __LINE__, "Wrong consider_all_instanecs value!");
					}

					//cout << "nn.size() after insert = " << nn.size() << " and nn = ";
					//for(map < double, long >::iterator it_temp=nn.begin(); it_temp != nn.end(); it_temp++)
						//cout << it_temp->first << ", " << it_temp->second << "; ";
					//cout << endl;
					
					it2 = nn.end();
					//cout << "nn.end()->first = " << (--it2)->first << endl;
				} // for(int j=0; j < tempLoopCounter; j++)
				
				
				// Printing the NN list of activePoint i
				//cout << "calculating connectivity now:" << endl;
				//cout << "nn.size() after insert = " << nn.size() << " and nn = ";
				//for(map < double, long >::iterator it_temp=nn.begin(); it_temp != nn.end(); it_temp++)
				//	cout << it_temp->first << ", " << it_temp->second << "; ";
				//cout << endl;
				
				
				// NN list of instance i is sorted in nn_list and nn_dist.
				// calculate instance i's connectivity now
				int j = 0;
				
#ifdef DEBUGGING
				cout << "solution::evaluate() => the host to be evaluated:" << endl;
				print(0);
				for(int p=0; p<(int)size(); p++)
					popHistory[level][sol.index1[p]]->print(0);
				cout << endl;
				
				cout << "NN list for instance " << i << ": ";
				for(map < double, long >::iterator it_temp=nn.begin(); it_temp != nn.end(); it_temp++)
					cout << it_temp->second << ", ";
				cout << endl << endl;
#endif
					
				for(map < double, long >::iterator it_temp=nn.begin(); it_temp != nn.end(); it_temp++)
				{
#ifdef DEBUGGING
					cout << "solution::evaluate() => ";
					cout << "i=" << i << ", index[" << model->_activePoints[i] << "] = " << indices[model->_activePoints[i]];
					cout << ", j=" << j << ", index[" << it_temp->second << "] = " << indices[it_temp->second] << endl;
#endif
					
					// If the NN instance is not clustered by current host, cluster it now cuz we need to know if 
					// it is in the same cluster as instance i
					if(indices[it_temp->second] == -1)
					{
						double dist = model->_infinity;
						subClstOffset = 0;
						int index = -1;
						double tempdist = 0;

						vector <int> index1;
						vector <int> index2;
						/*
						double dist2 = model->_infinity;
						int subClstOffset2 = -1;
						int index_2 = -1;
						*/
						
						for(int m=0; m < (int)size(); m++)
						{
#ifdef DEBUGGING
							cout << "solution::evaluate() => distances after calling getDistances() for symbiont " << sol.index1[m] << ": ";
							vector <double> tempdist3 = popHistory[level][sol.index1[m]]->getDistances();
							for(int l=0; l < (int)tempdist3.size(); l++)
								cout << tempdist3[l] << ", ";
							cout << endl;
#endif
							
							tempdist = popHistory[level][sol.index1[m]]->getDistances()[it_temp->second];
							//cout << "tempdist=" << tempdist << endl;
							
							//============
							// If the distance of current instance and current symbiont is never calculated
							//============
							if(tempdist == -1)
							{
								index1.clear();
								double tempdist2 = 0.0;
								int attr, clus;
								
								index2.clear();
								
								index1 = (popHistory[level][sol.index1[m]])->getGene().index1;
								index2 = (popHistory[level][sol.index1[m]])->getGene().index2;
								
#ifdef DEBUGGING
								cout << "index1: ";
								for(int k=0; k < (int)index1.size(); k++)
									cout << index1[k] << " ";
								cout << endl;
								cout << "index2: ";
								for(int k=0; k < (int)index2.size(); k++)
									cout << index2[k] << " ";
								cout << endl;
#endif
								
								for(int k=0; k < (int) index1.size(); k++)
								{
									attr = index1[k];
									clus = index2[k];
									tempdist2 += pow((model->_pointSet[it_temp->second][attr] - model->_attrWiseClusCent[attr][clus]), 2);
									
#ifdef DEBUGGING
									cout << "attr = " << attr << ", clus = " << clus;
									cout << ", dataset value[" << it_temp->second << "][" << attr << "]=" << model->_pointSet[it_temp->second][attr];
									cout << ", cluster centroid[" << attr << "][" << clus << "]=" << model->_attrWiseClusCent[attr][clus];
									cout << ", and accumulated tempdist2=" << tempdist2 << endl;
#endif
								}
								
								// Normalize the distance the way it is in calcDistance() function, either by divinding by
								// N or sqrt(N)
								tempdist2 = sqrt(tempdist2) / (double)index1.size();
								
#ifdef DEBUGGING
								cout << "distance from instance " << it_temp->second << " to subspace cluster (tempdist2)=" << sol.index1[m] << " is " << tempdist2 << endl;
#endif
								/*
								if((tempdist2 < dist2) || (tempdist2==dist2 && (int)index1.size()<(int)popHistory[level][sol.index1[subClstOffset2]]->size()))
								{
									dist2 = tempdist2;
									subClstOffset2 = m;
									index_2 = sol.index1[m];
								}
								*/
								// Once the distance between an instance and a symbiont is calculated put it in tempdist
								tempdist = tempdist2;
							} // if(tempdist == -1)
							//============
							// Done calculating the distance between current instance and symbiont
							//============
							
							
							if(tempdist == -1)
								cout << "solution::evaluate() => distance equals -1 (case indices[" << it_temp->second << "] = -1) => RE-CODE HERE!" << endl;
							else if(tempdist > 0 && tempdist < dist)
							{
								dist = tempdist;
								subClstOffset = m;
								index = sol.index1[m];
//								cout << "dist=" << dist << ", subClstOffset=" << subClstOffset << ", index=" << index << ", ";
							}
//							cout << endl;
							
						} // for(int m=0; m < (int)size(); m++)
						
//						cout << "dist=" << dist << ", subClstOffset=" << subClstOffset << ", index=" << index << ", ";
						// Storing the index of the symbiont which clusters instance it_temp->second
						indices[it_temp->second] = index;
						
					} // if(indices[it_temp->second] == -1)
					
					if(indices[model->_activePoints[i]] != indices[it_temp->second])
					{
						conn += double(1.0) / (double(j)+1.0);
//						cout << " => conn so far is = " << conn;
					}
//					cout << endl;
					
					j++;
				} // for(map < double, long >::iterator it_temp=nn.begin(); it_temp != nn.end(); it_temp++)
			} // for(int i=0; i < model->_pointPopSize; i++)
			
//			cout << " and the connectivity of this host is: " << conn << endl;
		} // if(model->_doConnectivity)
		else
		{
			conn = 1.0;
		}
		
		//cout << "numObjectives1 = " << model->_numObjectives1 << endl;
		//cout << "distortion = " << sum << " and connectivity = " << conn << endl;
		// print ind before changing the objective values
		//cout << "Print ind before changing the objective values." << endl;
		//print(0);
		
		// Sum of distances for all solutions.
		//static_sum += sum;
		//cout << "solution::evaluate() => static_sum = " << static_sum << endl;
		
		if(model->_doDistortion) // We could also take the square root of this distance, since eucleadian formula
			objectives[0] = sum; // nearest neighbour distortion value or compactness
		else
			objectives[0] = 1.0; // no use of distortion objective
		
		if(model->_numObjectives1 > 1 && model->_doConnectivity)
			objectives[1] = conn; // connectivity value of the solution
		else if(model->_numObjectives1 > 1 && model->_doAttrCount)
			objectives[1] = (double)numUniqueAttrs; // or (double)attrsAccessed.size();
		else if(model->_numObjectives1 > 1 && model->_doClusterCount)
			objectives[1] = (double)sol.size(); // length of an indivisual or number of subspace clusters it indexes (minimize);
		else
			objectives[1] = 1.0; // no use of the second objective
		
		
#ifdef DEBUGGING
		// Print the host after changing the objective values
		cout << "Print the host after changing the objective values." << endl;
		print(0);
#endif
		
#ifdef DEBUGGING
		cout << "solution::evaluate() exit." << endl;
#endif
		
		if(gen == model->_numGenerations && model->_labelsProvided)
			return indices;
		else
			return vector <int> (2, -1);
		
	}
	
	else
		die(__FILE__, __FUNCTION__, __LINE__, "Wrong level value!");
	
}

/***********************************************************************************************************/

void solution::crossover(solution *far, solution *mor)
{
	sol.crossover(far->sol, mor->sol);
}

/***********************************************************************************************************/

void solution::mutation()
{
	sol.mutation();
}

/***********************************************************************************************************/

/*
vector <int> solution::getSCindexes()
{
	vector <int> sol.index1(sol.size());
//	sol.index1 = (int *)malloc(sol.index1.size()*sizeof(int));

	if(sol.index1 == NULL)
		die(__FILE__, __FUNCTION__, __LINE__, "Can't allocate memory for sol.index1!");
	
	for(int i=0; i < (int)sol.size(); i++)
		sc_list[i] = sol.index1[i];
	
	return sc_list;
}
*/

/***********************************************************************************************************/

/*
vector <int> solution::setClusterLabels()
{

	vector <int> clusters(model->_dataSetSize, -1);
	
	if(clusters == NULL)
		die(__FILE__, __FUNCTION__, __LINE__, "Can't allocate memory for clusters!");

	
	vector <int> clusters = evaluate(model->_numGenerations);
	
	return clusters;
}
*/

/***********************************************************************************************************/

/* Function to calculate cluster to class distribution (confusion matrix) and performance measures */
//int performance_measures(vector<int> clusters or int* clusters)
// do_print determines if confusion matrix is printed, do_print2 determines if performance measures (acc, f1, ...) are printed

void solution::performance_measures(int do_print, int do_print2)
{
	// Determining true number of clusters. Note that true num of clusters might be different from ind's length, 
	// since some clusters might have no instances assigned to them, however after adding prune2() function 
	// all clusters have at least one instance assigned which means ind's length equals true number of clusters.
	if(sol.pruneSymbsWithZeroInsts())
	{
//		evaluate(model->_gen);
//		cout << "solution::performance_measures() => solution after prunning:" << endl; 
//		print(0);
	}
	
	// "clusters" stores labels assigned to points by the clustering solution at hand.
	vector <int> clusters = evaluate(model->_numGenerations1);
	cout << "solution::performance_measures() => solution after evaluating with all instances:" << endl;
	print(0);

	// num_clusters equals number of subspace clusters in clustering solution at hand.
	int num_clusters = size();
	
//#ifdef DEBUGGING
	cout << "solution::performance_measures() => clusters.size() = " << clusters.size() << endl;
	for(int i=0; i < (int)clusters.size(); i++)
		cout << clusters[i] << " ";
	cout << endl;
	
	cout << "solution::performance_measures() => num_clusters = " << num_clusters << endl;
//#endif
	
	// classes stores the true class labels of points.
	vector <int> classes(model->_dataSetSize, -1);
	classes.clear();
	for(int i=0; i < model->_dataSetSize; i++)
		classes.push_back((int)model->_labelSet[i]);
	
	// num_classes equals number of true classes.
	int num_classes = (int) model->_uniqueLabels.size();
	
#ifdef DEBUGGING
	cout << "solution::performance_measures() => classes.size() = " << classes.size() << endl;
	for(int i=0; i < (int)classes.size(); i++)
		cout << classes[i] << " ";
	cout << endl;

	cout << "solution::performance_measures() => num_classes = " << num_classes << endl;
#endif
	
	/*
	// This piece of code is redundant, determining the number of clusters, 
	// assuming that there are clusters with zero points which need to be pruned.
	vector <int> uniques(model->_dataSetSize, -1);
	
	uniques[0] = clusters[0];
	num_clusters++;
	
	for(int m=0; m < model->_dataSetSize; m++)
	{
		int flag = 0;
		for(int n=0; n < num_clusters; n++)
		{
			if(clusters[m] == uniques[n])
			{
				flag = 1;
				break;
			}
		}
		if(flag==0)
		{
			uniques[num_clusters] = clusters[m];
			//cout << uniques[num_clusters] << endl;
			num_clusters++;
		}
	}
	*/
	
	// Making the confusion matrix.
	int confusion[num_clusters][num_classes];

	for(int i=0; i < num_clusters; i++)
		for(int j=0; j < num_classes; j++)
			confusion[i][j] = 0;
	
	for(int i=0; i < model->_dataSetSize; i++)
		for(int j=0; j < num_clusters; j++)
			if(clusters[i] == sol.index1[j]) // uniques[j]
			{
				//confusion[j][classes[i]-1]++; // THIS LINE NEEDS MODIFICATION
				confusion[j][model->getLabelIndex(classes[i])]++; // THIS LINE NEEDS MODIFICATION
				//printf("cluster %d matched with class %d, %d times\n", clusters[i], classes[i], confusion[j][classes[i]]);
			}
	
	
	// =======================================
#ifdef DEBUGGING
	// temporary printing the confusion matrix
	cout << endl << "printing confusion matrix right after calculating it" << endl;
	for(int i=0; i < num_clusters; i++)
	{
		for(int j=0; j < num_classes; j++)
			cout << confusion[i][j] << " ";
		cout << endl;
	}
	cout << endl;
#endif
	// =======================================
	
	
	// Summing rows and columns of confusion matrix.
	int sum = 0;
	int class_sum[num_classes];
	int cluster_sum[num_clusters];
	
	for(int p=0; p < num_clusters; p++)
	{
		sum = 0;
		
		for(int q=0; q < num_classes; q++)
			sum += confusion[p][q];
		
		cluster_sum[p] = sum;
	}
	
	for(int q=0; q < num_classes; q++)
	{
		sum = 0;
		
		for(int p=0; p < num_clusters; p++)
			sum += confusion[p][q];

		class_sum[q] = sum;
	}

	
	// =======================================
#ifdef DEBUGGING
	// temporary printing the confusion matrix
	cout << "printing confusion matrix right after calculating cluster-sums and class-sums" << endl;
	for(int i=0; i < num_clusters; i++)
	{
		for(int j=0; j < num_classes; j++)
			cout << confusion[i][j] << " ";
		cout << cluster_sum[i] << endl;
	}
	for(int i=0; i < num_classes; i++)
		cout << class_sum[i] << " ";
	cout << endl << endl;
#endif
	// =======================================
	
	
	//	cluster2class array keeps the class associated with each cluster.
	//	accuracy is calculated by summing the maximum number in each cluster and dividing by model->_dataSetSize
	
	//int cluster2class[num_clusters];
	vector <int> cluster2class(num_clusters);
	double precision[num_clusters];
	double recall[num_clusters];
	double f_value[num_clusters];
	double coverage[num_clusters];
	double purity[num_clusters];
	double overall_accuracy = 0;
	double overall_f_value = 0;
	double overall_purity = 0;
	double f_value_sum = 0;
	double purity_sum = 0;
	int tmpSum = 0;
	int tmpIdx = -1;
	int tmpCnt = 0;
	double tmpPerc = 0;
	
	for(int p=0; p < num_clusters; p++)
	{
		tmpIdx = -1;
		tmpCnt = 0;
		tmpPerc = 0;
		
		for(int q=0; q < num_classes; q++)
		{
			if((double)confusion[p][q]/(double)class_sum[q] > tmpPerc) // when CLASS PERCENTAGE is to be checked
			//if((double)confusion[p][q] > tmpCnt) // when EXEMPLAR COUNT are to be checked
			{
				tmpCnt = confusion[p][q];
				tmpPerc = (double)confusion[p][q] / (double)class_sum[q];
				tmpIdx = q;
			}
		}
		
		tmpSum += tmpCnt;
		
		cluster2class[p] = (int)model->_uniqueLabels[tmpIdx];
		//cout << "solution::performance_measures() => cluster " << p << " is mapped to class " << cluster2class[p] << endl;
		
		precision[p] = (double)confusion[p][tmpIdx] / (double)cluster_sum[p];
		recall[p] = (double)confusion[p][tmpIdx] / (double)class_sum[tmpIdx];
		
		f_value[p] = (2*precision[p]*recall[p]) / (precision[p]+recall[p]);
		
		coverage[p] = (double)cluster_sum[p] / (double)model->_dataSetSize;
		purity[p] = (double)tmpCnt / (double)cluster_sum[p];
		
		overall_f_value += coverage[p] * f_value[p]; // weighted f measure
		overall_purity += coverage[p] * purity[p]; // weighted purity
		f_value_sum += f_value[p]; // sum of f measures
		purity_sum += purity[p]; // sum of purities
	}
	
	overall_accuracy = (double)tmpSum / (double)model->_dataSetSize;
	

	// determining the unique number of clusters 
	vector <int> clusters_uniques = cluster2class;
	vector <int>::iterator it;
	sort(clusters_uniques.begin(), clusters_uniques.end());
	it = unique(clusters_uniques.begin(), clusters_uniques.end());
	clusters_uniques.resize(it - clusters_uniques.begin());
	int num_clusters_uniques = (int)clusters_uniques.size();
	
	
	// =======================================
#ifdef DEBUGGING
	cout << "cluster2class = ";
	for(int i=0; i < (int)cluster2class.size(); i++)
		cout << cluster2class[i] << " ";
	cout << endl;

	cout << "clusters_uniques = ";
	for(int i=0; i < (int)clusters_uniques.size(); i++)
		cout << clusters_uniques[i] << " ";
	cout << endl << endl;
#endif
	// =======================================
	
	
	// Shrink the confusion matrix to a (num_clusters_uniques X num_classes) matrix by merging clusters mapped ot the same class
	// confusion2 matrix is used for merging clusters into num_classes clusters hierarchically and then determining performance
//	int confusion2[min(num_classes, num_clusters)][num_classes];
	int confusion2[num_clusters_uniques][num_classes];
	
	for(int i=0; i < num_clusters_uniques; i++)//min(num_classes, num_clusters)
		for(int j=0; j < num_classes; j++)
			confusion2[i][j] = 0;
	
	vector<int> cluster2class2(num_clusters_uniques, -1);
	
	for(int i=0, m=0; i < num_clusters; i++, m++)
	{
		if(find(cluster2class2.begin(), cluster2class2.end(), cluster2class[i]) == cluster2class2.end())
			for(int k=0; k < num_classes; k++)
				confusion2[m][k] += confusion[i][k];
		else
			m--;

		
		if(find(cluster2class2.begin(), cluster2class2.end(), cluster2class[i]) == cluster2class2.end())
			for(int j = i+1; j < num_clusters; j++)
			{
				if(cluster2class[j] == cluster2class[i])
					for(int k=0; k < num_classes; k++)
						confusion2[m][k] += confusion[j][k];
			}
		
		//if(i <= num_clusters_uniques)
			if(find(cluster2class2.begin(), cluster2class2.end(), cluster2class[i]) == cluster2class2.end())
				//cluster2class2.push_back(cluster2class[i]);
				cluster2class2[m] = cluster2class[i];

	}
	
	
	// =======================================
#ifdef DEBUGGING
	// temporary printing the confusion matrix
	cout << "printing confusion matrix right after condensing the confusion matrix" << endl;
	cout << "cluster2class = ";
	for(int i=0; i < (int)cluster2class.size(); i++)
		cout << cluster2class[i] << " ";
	cout << endl;
	cout << "clusters_uniques = ";
	for(int i=0; i < (int)clusters_uniques.size(); i++)
		cout << clusters_uniques[i] << " ";
	cout << endl;
	cout << "cluster2class2 = ";
	for(int i=0; i < (int)cluster2class2.size(); i++)
		cout << cluster2class2[i] << " ";
	cout << endl << endl;
#endif
	// =======================================
	
	
	// Summing rows and columns of new confusion matrix.
	int sum2 = 0;
	int class_sum2[num_classes];
	int cluster_sum2[num_clusters_uniques];//min(num_classes,num_clusters) or num_clusters
	
	for(int p=0; p < num_clusters_uniques; p++)//min(num_classes,num_clusters) or num_classes
	{
		sum2 = 0;
		
		for(int q=0; q < num_classes; q++)
			sum2 += confusion2[p][q];

		cluster_sum2[p] = sum2;
	}
	
	for(int q=0; q < num_classes; q++)
	{
		sum2 = 0;
		
		for(int p=0; p < num_clusters_uniques; p++)//min(num_classes,num_clusters) or num_classes
			sum2 += confusion2[p][q];

		class_sum2[q] = sum2;
	}
	
	
	// =======================================
#ifdef DEBUGGING
	// temporary printing the confusion matrix
	cout << "printing confusion matrix right after calculating the cluster-sums and class-sums of new condensed matrix" << endl;
	for(int i=0; i < num_clusters; i++)
	{
		for(int j=0; j < num_classes; j++)
			cout << confusion[i][j] << " ";
		cout << cluster_sum[i] << endl;
	}
	for(int i=0; i < num_classes; i++)
		cout << class_sum[i] << " ";
	cout << endl << endl;
#endif
	// =======================================
	
	
	// Calculating detection rate for each class, then calculating class-wise detection rate
	//double detection[num_classes];
	vector<double> detection(num_classes);
	double detection_sum = 0;
	
	 
	for(int q=0, m=0; q < num_classes; q++)
	{
		if(find(cluster2class2.begin(), cluster2class2.end(), (int)model->_uniqueLabels[q]) != cluster2class2.end())
		{
			int idx = 0;
			for(int j=0; j < (int)cluster2class2.size(); j++)
			{
				if((int)model->_uniqueLabels[q] == cluster2class2[j])
				{
					idx = j;
					break;
				}
			}
			
			detection[q] = (double)confusion2[idx][q] / (double)class_sum2[q];//model->getLabelIndex(cluster2class2[q])
			detection_sum += detection[q];
			m++;
		}
		else
			detection[q] = 0;

	}
	
	
	// =======================================
#ifdef DEBUGGING
	// temporary printing the confusion matrix
	cout << "printing confusion matrix calculating the detection rates" << endl;
	for(int i=0; i < num_clusters; i++)
	{
		for(int j=0; j < num_classes; j++)
			cout << confusion[i][j] << " ";
		cout << cluster_sum[i] << endl;
	}
	for(int i=0; i < num_classes; i++)
		cout << class_sum[i] << " ";
	cout << endl << endl;
#endif
	// =======================================
	
	
	// PRINTING STARTS HERE
	if(do_print)
	{
		// printing the original confusion matrix
		printf("\nPrinting the ORIGINAL confusion matrix");
		printf("\nClasses:   ");
		for(int p=0; p < num_classes; p++)
		{
			// here determine wether class labels start with 0 or 1 while printing based on "FIRST_CLASS_LABEL"
			printf("%5d           ", (int)model->_uniqueLabels[p]);
			//printf("%5d           ", p + FIRST_CLASS_LABEL);
		}
		printf("      Sum   Mapped class");
		printf("\n");
		
		printf("           ");
		for(int p=0; p < num_classes; p++)
		{
			printf("----------------");
		}
		printf("----\n");
		
		for(int p=0; p < num_clusters; p++)
		{
			printf("cluster %2d:", sol.index1[p]);//uniques[p]
			for(int q=0; q < num_classes; q++)
			{
				printf("%5d (%1.4lf)  ", confusion[p][q], ((double)confusion[p][q]/(double)class_sum[q]));
			}
			printf("   |%5d        %2d\n", cluster_sum[p], cluster2class[p]);
		}
		
		printf("           ");
		for(int q=0; q < num_classes; q++)
		{
			printf("----------------");
		}
		printf("----\n");
		
		printf("Sum        ");
		for(int q=0; q < num_classes; q++)
		{
			printf("%5d           ", class_sum[q]);
		}
		printf("\n");
		
		// printing the merged confusion matrix
		printf("\nPrinting the CONDENSED confusion matrix");
		printf("\nClasses:   ");
		for(int p=0; p < num_classes; p++)
		{
			// here determine wether class labels start with 0 or 1 while printing based on "FIRST_CLASS_LABEL"
			printf("%5d           ", (int)model->_uniqueLabels[p]);
			//printf("%5d           ", p + FIRST_CLASS_LABEL);
		}
		printf("      Sum   Mapped class");
		printf("\n");
		
		printf("           ");
		for(int p=0; p < num_classes; p++)
		{
			printf("----------------");
		}
		printf("----\n");
		
		for(int p=0; p < num_clusters_uniques; p++)//min(num_classes, num_clusters) or num_classes
		{
			printf("cluster %2d:", p);//sol.index1[p]
			for(int q=0; q < num_classes; q++)
			{
				printf("%5d (%1.4lf)  ", confusion2[p][q], ((double)confusion2[p][q]/(double)class_sum2[q]));
			}
			printf("   |%5d        %2d\n", cluster_sum2[p], cluster2class2[p]);
		}
		
		printf("           ");
		for(int q=0; q < num_classes; q++)
		{
			printf("----------------");
		}
		printf("----\n");
		
		printf("Sum        ");
		for(int q=0; q < num_classes; q++)
		{
			printf("%5d           ", class_sum2[q]);
		}
		printf("\n");
		printf("\n");
		
		// printing detection rates
		printf("Classes:  ");
		printf("   0");
		for(int q=1; q < num_classes; q++)
		{
			printf("%8d", q);
		}
		printf("\n");
		printf("          ");
		for(int q=0; q < num_classes; q++)
		{
			printf("--------");
		}
		printf("\n");
		
		//	printing detection rates
		printf("Detection: ");
		for(int q=0; q < num_classes; q++)
		{
			printf("%1.4lf  ", detection[q]);
		}
		printf("\n");
		printf("\n");
		
		// Printing cluster performance measures
		printf("Clusters:  ");
		printf("   0");
		for(int p=1; p < num_clusters; p++)
		{
			printf("%8d", p);
		}
		printf("\n");
		printf("          ");
		for(int p=0; p < num_clusters; p++)
		{
			printf("--------");
		}
		printf("\n");
		
		//	printing precision
		printf("Precision: ");
		for(int p=0; p < num_clusters; p++)
		{
			printf("%1.4lf  ", precision[p]);
		}
		printf("\n");
		//	printing recall
		printf("Recall:    ");
		for(int p=0; p < num_clusters; p++)
		{
			printf("%1.4lf  ", recall[p]);
		}
		printf("\n");
		//	printing f_value
		printf("F_value:   ");
		for(int p=0; p < num_clusters; p++)
		{
			printf("%1.4lf  ", f_value[p]);
		}
		printf("\n");
		//	printing coverage
		printf("Coverage:  ");
		for(int p=0; p < num_clusters; p++)
		{
			printf("%1.4lf  ", coverage[p]);
		}
		printf("\n");
		//	printing purity
		printf("Purity:    ");
		for(int p=0; p < num_clusters; p++)
		{
			printf("%1.4lf  ", purity[p]);
		}
		printf("\n\n");
	}
	
	if(do_print2)
	{
		cout << "Detection Rate = " << (double)detection_sum / (double)num_classes << endl;
		cout << "Overall Accuracy = " << overall_accuracy << endl;
		cout << "Average F_measure = " << (double)f_value_sum / (double)num_clusters << endl;
		cout << "Average Purity = " << (double)purity_sum / (double)num_clusters << endl;
		cout << "Weighted F_measure = " << overall_f_value << endl;
		cout << "Weighted Purity = " << overall_purity << endl;
	}
	
	cout << endl;
	
	return;
}

/***********************************************************************************************************/

void solution::updateAttrs()
{
	if(level == 0)
	{
#ifdef DEBUGGING
		cout << "solution::updateAttrs() => attrs: ";
		for(int i=0; i < (int)attrs.size(); i++)
			cout << attrs[i] << " ";
		cout << endl;
		
		cout << "solution::updateAttrs() => sol.index: ";
		for(int i=0; i < (int)sol.index1.size(); i++)
			cout << sol.index1[i] << " ";
		cout << endl;
#endif
		
		attrs = sol.index1;
		attrPerSymb = (double)attrs.size();
		
	}
	
	else if(level != 0 & level < model->_numLevels)
	{
		cout << "solution::updateAttrs() not implemented for solutions of level 1 or higher." << endl;
	}
	
	else
		die(__FILE__, __FUNCTION__, __LINE__, "Wrong level value!");
}

/***********************************************************************************************************/

void solution::calcDistances()
{
	// The distance between  an instance and a symbiont/SCC should be normalized by the attribute count of the symbiont.
	// This function is only defined for level 0 solutions (symbionts), and not higher level solutions (hosts.)
	// If subsampling is not requested, or if this is the last generation of the ESC distances from all instances are calculated,
	// otherwise only distances from activePoints are calculated.
	
#ifdef DEBUGGING
	cout << "solution::calcDistances() => enter" << endl;
#endif

#ifdef DEBUGGING
	cout << "solution::calcDistances() => The new cloned symbiont after mutation:" << endl;
	print(0);
#endif
	
	int attr, clus;//, tempLoopCounter;
	double dist;
	
	// This piece was used when there was a distiction before two cases of using subsampling and not using it.
//	if(model->_doSubsampling && model->_gen < model->_numGenerations)
//		tempLoopCounter = model->_pointPopSize;
//	else
//		tempLoopCounter = model->_dataSetSize;
	
		if((model->_doSubsampling==0) || (model->_gen==model->_numGenerations))
		{
			//cout << "solution::calcDistances() => calculating distances for all points (case 1)" << endl;
			
			distances.clear();
			distances.resize(model->_dataSetSize, -1);
			
			for(int i=0; i < (int)model->_dataSetSize; i++)
			{
				dist = 0;
				
				for(int j=0; j < (int)size(); j++)
				{
					attr = getGene().index1[j];
					clus = getGene().index2[j];
					
					dist += pow((model->_pointSet[i][attr] - model->_attrWiseClusCent[attr][clus]), 2); 
					
					//cout << "solution::calcDistances() => attr = " << attr << ", clus = " << clus << ", and dist = " << dist << endl;
				}
				
				// Euclidean distance normalized by the number of supporting attributes, or its squared root.
				distances[i] = sqrt(dist) / (double)size();//sqrt(double(size());
				
				//cout << "solution::calcDistances() => dist = " << dist << ", sqrt(dist) = " << sqrt(dist) << ", size() = " << size() << ", sqrt(dist)/size() = " << distances[i] << endl;
			}
		}
		else if((model->_doSubsampling) && !(model->_gen==model->_numGenerations) && (model->_doConnectivity))
		{
			//cout << "solution::calcDistances() => calculating distances for activeSet points (2)" << endl;
			
			distances.clear();
			distances.resize(model->_dataSetSize, -1);
			
			for(int i=0; i < (int)model->_activePoints.size(); i++)
			{
				dist = 0;
				
				for(int j=0; j < (int)size(); j++)
				{
					attr = getGene().index1[j];
					clus = getGene().index2[j];
					
					dist += pow((model->_pointSet[model->_activePoints[i]][attr] - model->_attrWiseClusCent[attr][clus]), 2); 
					
					//cout << "solution::calcDistances() => attr = " << attr << ", clus = " << clus << ", and dist = " << dist << endl;
				}
				
				// Euclidean distance normalized by the number of supporting attributes, or its squared root.
				distances[model->_activePoints[i]] = sqrt(dist) / (double)size();//sqrt(double(size());
				
				//cout << "solution::calcDistances() => dist = " << dist << ", sqrt(dist) = " << sqrt(dist) << ", size() = " << size() << ", sqrt(dist)/size() = " << distances[i] << endl;
			}
		}
		else
		{
			//cout << "solution::calcDistances() => calculating distances for activeSet points (case 3)" << endl;
			
			distances.clear();
			distances.resize(model->_pointPopSize, -1);
			
			for(int i=0; i < (int)model->_activePoints.size(); i++)
			{
				dist = 0;
				
				for(int j=0; j < (int)size(); j++)
				{
					attr = getGene().index1[j];
					clus = getGene().index2[j];
					
					dist += pow((model->_pointSet[model->_activePoints[i]][attr] - model->_attrWiseClusCent[attr][clus]), 2); 
					
					//cout << "solution::calcDistances() => attr = " << attr << ", clus = " << clus << ", and dist = " << dist << endl;
				}
				
				// Euclidean distance normalized by the number of supporting attributes, or its squared root.
				distances[i] = sqrt(dist) / (double)size();//sqrt(double(size());
				
				//cout << "solution::calcDistances() => dist = " << dist << ", sqrt(dist) = " << sqrt(dist) << ", size() = " << size() << ", sqrt(dist)/size() = " << distances[i] << endl;
			}
#ifdef DEBUGGING
			cout << "solution::calcDistances() => distances for activeSet points (case 3): " << endl;
			for(int k = 0; k < (int)distances.size(); k++)
				cout << "activePoint[" << model->_activePoints[k] << "]'s distance = " << distances[k] << ", ";
			cout << endl;
#endif
		}
	
	// This piece of code is replaced by the piece above it
/*
	for(int i=0; i < (int)model->_activePoints.size(); i++)
	{
		dist = 0;
		
		for(int j=0; j < (int)size(); j++)
		{
			attr = getGene().index1[j];
			clus = getGene().index2[j];
			
			// This piece was used when there was a distiction before two cases of using subsampling and not using it.
//			if(model->_doSubsampling && model->_gen < model->_numGenerations)
//				dist += pow((model->_pointSet[model->_activePoints[i]][attr] - model->_attrWiseClusCent[attr][clus]), 2); 
//			else
//				dist += pow((model->_pointSet[i][attr] - model->_attrWiseClusCent[attr][clus]), 2); 

			dist += pow((model->_pointSet[model->_activePoints[i]][attr] - model->_attrWiseClusCent[attr][clus]), 2); 
			
			//cout << "solution::calcDistances() => attr = " << attr << ", clus = " << clus << ", and dist = " << dist << endl;
		}
		
		// Euclidean distance normalized by the number of supporting attributes, or its squared root.
//		distances[i] = sqrt(dist) / (double)size();//sqrt(double(size());
		distances[model->_activePoints[i]] = sqrt(dist) / (double)size();//sqrt(double(size());
		
		//cout << "solution::calcDistances() => dist = " << dist << ", sqrt(dist) = " << sqrt(dist) << ", size() = " << size() << ", sqrt(dist)/size() = " << distances[i] << endl;
		
	}
*/ 

	
#ifdef DEBUGGING
	cout << "solution::calcDistances() => distances.size() = " << distances.size() << endl;
	for(int j=0; j < (int)distances.size(); j++)
		cout << distances[j] << " ";
	cout << endl;
#endif
	
#ifdef DEBUGGING
	cout << "solution::calcDistances() => exit" << endl;
#endif
	
}

/***********************************************************************************************************/

