
//#include "esc.hpp"
#include "types.hpp"

#include <iostream>
#include <algorithm>
#include <limits.h>
#include <float.h>
#include <vector>
#include <iterator>

extern esc *model;

/***********************************************************************************************************/

size_t currentobj;

void setCurrentObj(size_t c)
{
	currentobj = c;
}

/***********************************************************************************************************/

// used to sort on a specific objective
bool cmp_1(const solution *s1, const solution *s2)
{
	return(s1->getObj(currentobj) < s2->getObj(currentobj));
	//return (*s1 < *s2);
}

/***********************************************************************************************************/

bool cmp_cd(const solution *s1, const solution *s2)
{
	return(s1->cmpOnCD(*s2));
}

/***********************************************************************************************************/

// checks if one solution dominates another
bool solution::operator>> (const solution& s1) const
{ 
	bool better = false;
	
	for(vector<double>::const_iterator i=objectives.begin(), j=s1.objectives.begin(); i != objectives.end(); i++, j++)
	{ 
		if(*i > *j) // all objectives are to be minimized.
			return(false);
		if(*i < *j)
			better = true;
	}
	
	return(better);
}

/***********************************************************************************************************/

// checks if one solutions has better frontnumber/crowddistance than another (This is not the domination function)
bool solution::operator<< (const solution& s1) const
{
	return( frontnumber<s1.frontnumber || (frontnumber==s1.frontnumber && crowdingdistance>s1.crowdingdistance) );
}

/***********************************************************************************************************/

// checks if one solution dominates another in objective currentobj
bool solution::operator< (const solution& s1) const
{
	return(objectives[currentobj] < s1.objectives[currentobj]);
}

/***********************************************************************************************************/

// sorts a front on the objective currentobj
void front::sortOnObjective(int o)
{  
	setCurrentObj(o);
#if ATTEMPT_IDENTICAL
	stable_sort(indexes.begin(),indexes.end(),cmp_1);
#else
	sort(indexes.begin(),indexes.end(),cmp_1);
#endif
}

/***********************************************************************************************************/

// sorts a front on crowdingdistances
void front::sortOnCD()
{
#if ATTEMPT_IDENTICAL
	stable_sort(indexes.begin(),indexes.end(),cmp_cd);
#else
	sort(indexes.begin(),indexes.end(),cmp_cd);
#endif
}

/***********************************************************************************************************/

#if SORTING==SLOWSORTING

// This is the original slow sorting algorithm (the one termed 'fast' by Deb)
void population::sortOnFronts()
{
	sortedfronts=vector<front>();
	
	// delete previous information
	for(vector<solution>::size_type i=0;i<pop.size();i++)
	{
		pop[i].domcount=0;
		pop[i].dominates.resize(0); 
	}
	
	
	// compare every solution to all other solutions
	for(vector<solution>::size_type i=0; i < pop.size(); i++)
	{
		for(vector<solution>::size_type j=i+1; j < pop.size(); j++)
		{
			if(pop[i] >> pop[j])
			{
				pop[i].dominates.push_back(&pop[j]);
				pop[j].domcount++;
			}
			else
			{ 
				if(pop[j] >> pop[i])
				{
					pop[j].dominates.push_back(&pop[i]);
					pop[i].domcount++;
				}
			}
		}
		
		if(pop[i].domcount == 0)
		{
			if(sortedfronts.size() == 0)
				sortedfronts.push_back(front(&pop[i],1));
			else
				sortedfronts[0].push_back(&pop[i]);
		}
	}
	
	// build the fronts
	unsigned int nfront=0;
	
	while(sortedfronts.size() > nfront)
	{
		front tmpfront;
		bool tfinit = false;
		
		for(vector<solution *>::size_type i=0; i < sortedfronts[nfront].size(); i++)
		{
			for(vector<solution *>::iterator iter=sortedfronts[nfront][i].dominates.begin(); iter != sortedfronts[nfront][i].dominates.end(); iter++)
			{
				(*iter)->domcount--;
				
				if((*iter)->domcount == 0)
				{
					if(!tfinit)
					{
						tmpfront = front(*iter,nfront+2);
						tfinit = true;
					}
					else
						tmpfront.push_back(*iter);
				}
			}
		}
		
		if(tfinit)
		{
			sortedfronts.push_back(front());
			sortedfronts[nfront+1].swap(tmpfront);
			tfinit = false;
		}
		
		nfront++;
	}
	
	unsigned int s=0;
	unsigned int i=0;
	unsigned int p=0;
	
#if ATTEMPT_IDENTICAL
	while(s < doublepopsize) 
#else
		while(s < popsize) 
#endif
		{
			sortedfronts[i].calcCrowdingDistances();
			
			if(s+sortedfronts[i].size() > popsize)
			{
				front backup = sortedfronts[i];
				
				sortedfronts[i].sortOnCD();
				
				for(unsigned int j=0; j < sortedfronts[i].size(); j++)
					popsorted[p++] = &sortedfronts[i][j];
				
				sortedfronts[i] = backup; // restore sorting on objectives
			}
			else
				for(unsigned int j=0; j < sortedfronts[i].size(); j++)
					popsorted[p++] = &sortedfronts[i][j];
			
			s += sortedfronts[i++].size();
		}
	
#if !ATTEMPT_IDENTICAL
	while(s < doublepopsize)
	{
		for(unsigned int j=0; j < sortedfronts[i].size(); j++)
			popsorted[p++] = &sortedfronts[i][j];
		
		s += sortedfronts[i++].size();
	}
#endif

	// Added by A.V.
	paretofront = sortedfronts[0];
}

#endif

/***********************************************************************************************************/

void front::calcCrowdingDistances()
{  
	for(vector<solution *>::size_type i=0; i < indexes.size(); i++) 
		(*indexes[i]).resetCrowdDist();
	
	for(int o=0; o < model->_numObjectives; o++)
	{
		double minimum,maximum;  
		
		//cout << "sorting:" << endl;
		//print();
		
		sortOnObjective(o);
		
		minimum = (*indexes[0]).getObj(o);
		maximum = (*indexes[indexes.size()-1]).getObj(o);
		
		// model->_numObjectives+1 corresponds to infinite, since the sum of the
		// other contributions to crowdingdistance can never be more than
		// the number of objectives.
		(*indexes[0]).addtoCrowdDist(model->_numObjectives+1);
		
		if(indexes.size() > 1) 
			(*indexes[indexes.size()-1]).addtoCrowdDist(model->_numObjectives+1);
		
		if(minimum != maximum)
			for(vector<solution *>::size_type i=1; i < indexes.size()-1; i++) 
				(*indexes[i]).addtoCrowdDist( ((*indexes[i+1]).getObj(o)-(*indexes[i-1]).getObj(o))/(maximum-minimum) );
	}
}

/***********************************************************************************************************/

