
//#include "esc.hpp"
#include "types.hpp"

#include <algorithm>
#include <vector>
#include <stdlib.h>
#include <float.h>

#define NOSPLIT DBL_MAX

double split_limit;
unsigned int split_obj;

/***********************************************************************************************************/

bool split_cmp(solution *sol)
{
	return(sol->getObj(split_obj)<split_limit);
}

/***********************************************************************************************************/

void checksplit(vector<solution *> *L, vector<solution *> *H, unsigned int M, unsigned int indexL, unsigned int indexH, char mess[])
{
	double maxlow=-DBL_MAX,minhigh=DBL_MAX;
	unsigned int i;
	
	if( (*L).size()>0)
	{
		for(i=0;i<indexL;i++)
			if((*L)[i]->getObj(M)>maxlow)
				maxlow=(*L)[i]->getObj(M);
		
		while(i<(*L).size())
		{
			if(minhigh>(*L)[i]->getObj(M))
				minhigh=(*L)[i]->getObj(M);
			i++;
		}
		
		if(H!=NULL && (*H).size()>0)
		{
			for(i=0;i<indexH;i++) 
				if((*H)[i]->getObj(M)>maxlow)
					maxlow=(*H)[i]->getObj(M);
			
			while(i<(*H).size())
			{
				if(minhigh>(*H)[i]->getObj(M))
					minhigh=(*H)[i]->getObj(M);
				i++;
			}
		}
		
		if(maxlow>=minhigh)
		{
			cout << "Mismatch after split:"<< endl;
			
			cout << "maxlow:" << maxlow << "   minhigh:" << minhigh << endl;
			
			cout << "L low:";
			for(i=0;i<indexL;i++) 
				cout << " " << (*L)[i]->getObj(M);
			cout << endl << "L high:";
			while(i<(*L).size())
				cout << " " << (*L)[i++]->getObj(M);
			
			if(H!=NULL)
			{
				cout << endl << "H low:";
				for(i=0;i<indexH;i++) 
					cout << " " << (*H)[i]->getObj(M);
				cout << endl << "H high:";
				while(i<(*H).size())
					cout << " " << (*H)[i++]->getObj(M);
			}
			
			cout << endl << "cuts: " << indexL << "  " << indexH;
			cout << endl << mess << endl;
			exit(0);
		}
	}
}

/***********************************************************************************************************/

double find_splitpoint(vector<solution *>::iterator start, vector<solution *>::iterator slut)
{
	vector<solution *>::iterator search;
	
	search=start+rand()%(slut-start);
	vector<solution *>::iterator search_start=search;
	
	while(search!=slut && (*(search))->getObj(split_obj)==split_limit)
		search++;
	
	if(search==slut || (*(search))->getObj(split_obj)==split_limit)
		search=start;
	
	while(search!=search_start && (*(search))->getObj(split_obj)==split_limit) 
		search++;
	
	if((*(search))->getObj(split_obj)==split_limit)
		return(NOSPLIT);
	else
		return( (*(search))->getObj(split_obj) );
}

/***********************************************************************************************************/

bool splitA(vector<solution *> &V, unsigned int M, unsigned int &index)
{
	vector<solution *>::iterator start, slut,mid,target;
	
	start=V.begin();
	slut=V.end();
	
	target= V.begin()+V.size()/2;
	
	// start accepting non-midpoint splits at size 4, split must be within
	// 12.5 percent of the midpoint.
	int tolerance= (V.end()-V.begin())/4;  
	
	unsigned int inx=rand()%V.size();
	
	split_limit = (*(V.begin()+inx))->getObj(M);
	split_obj= M;
	
	mid = partition(start,slut,split_cmp);
	
	while(abs(mid-target)>tolerance)
	{
		
		if(mid<target)
			start=mid;
		else
			slut=mid;
		
		split_limit=find_splitpoint(start,slut);
		
		if(split_limit==NOSPLIT) {
			if(start==V.begin() && slut==V.end())
				return (false);
			else
			{	
				if(start-V.begin()>V.end()-slut)
					index=start-V.begin();
				else
					index=slut-V.begin();
				return(true);
			}
		}
		mid = partition(start,slut,split_cmp);
	}
	
	index=mid-V.begin();
	
	return(true);
}

/***********************************************************************************************************/

bool splitB(vector<solution *> &L, vector<solution *> &H, unsigned int M, unsigned int &indexL, unsigned int &indexH) 
{
	//  cout << " splitB: " << L.size() << "  " << H.size() << endl;
	
	vector<solution *>::iterator startL, slutL,midL;
	vector<solution *>::iterator startH, slutH,midH;
	int totalsize= L.size()+H.size();
	
	startL=L.begin();
	slutL=L.end();
	
	startH=H.begin();
	slutH=H.end();
	
	if(L.size()>H.size())
	{
		unsigned int inx=rand()%L.size();
		split_limit = (*(L.begin()+inx))->getObj(M);
	}
	else
	{
		unsigned int inx=rand()%H.size();
		split_limit = (*(H.begin()+inx))->getObj(M);
	}
	
	split_obj= M;
	midL = partition(startL,slutL,split_cmp);
	midH = partition(startH,slutH,split_cmp);
	
	while( (midL-L.begin()+midH-H.begin())<totalsize/4 || 
		  (L.end()-midL+H.end()-midH)<totalsize/4 )
	{
		
		if( (midL-L.begin()+midH-H.begin())<totalsize/4)
		{
			// move splitpoints upwards
			startL=midL;
			startH=midH;
		}
		else
		{
			// move splitpoints downwards
			slutL=midL;
			slutH=midH;
		}
		
		double newsplit_limit;
		
		// find new split_limit within one of the intervals
		if(L.size()>H.size() && startL!=slutL || startH==slutH)
		{
			newsplit_limit = find_splitpoint(startL,slutL);
			if(newsplit_limit==NOSPLIT && startH!=slutH)
				newsplit_limit = find_splitpoint(startH,slutH); 
		}
		else
		{
			newsplit_limit = find_splitpoint(startH,slutH);
			if(newsplit_limit==NOSPLIT && startL!=slutL)
				newsplit_limit = find_splitpoint(startL,slutL);
		}
		
		// no further splitting possible?
		if(newsplit_limit==NOSPLIT)
		{
			if(startL==L.begin() && slutL==L.end() &&
			   startH==H.begin() && slutH==H.end())
				return(false);/* no splitting was performed: all values
							   identical */
			else
			{
				if(startL-L.begin()+startH-H.begin()>
				   L.end()-slutL+H.end()-slutH)
				{
					indexL=startL-L.begin();
					indexH=startH-H.begin();
				}
				else
				{
					indexL=slutL-L.begin();
					indexH=slutH-H.begin();
				}
				return(true); // we did split
			}
		}
		else
		{
			// make yet another partition.
			split_limit=newsplit_limit;
			midL = partition(startL,slutL,split_cmp);
			midH = partition(startH,slutH,split_cmp);
		}
	}
	indexL=midL-L.begin();
	indexH=midH-H.begin();
	//  cout << "splitret at end" << indexL << " " << indexH << endl;
	
	return(true);
}

/***********************************************************************************************************/

