
//#include "esc.hpp"
#include "types.hpp"
#include "misc.hpp"

#include <iostream>
#include <iomanip>
//#include <fstream>
#include <vector>
#include <math.h>
#include <algorithm>

using namespace std;

// my declarations
//extern int crossoverGenerated;
//extern int mutationGenerated;

extern esc *model;
extern vector < vector <solution *> > popHistory;
extern vector < vector <int> > symbsToUse;
extern vector < vector <int> > symbsRefs;


/***********************************************************************************************************/

gene::gene(int l)
: level(l)
{
//	level = model->_lev;
	
#ifdef DEBUGGING
	cout << "gene::gene(int) enter." << endl;
#endif
	
	int num_gene = max(model->_minAttributes, model->_minSymbionts);
	
#ifdef DEBUGGING
	cout << "gene::gene(int) => level = " << level << endl;
	cout << "gene::gene(int) => _minAttributes = " << model->_minAttributes << ", _maxAttributes = " << model->_maxAttributes << ", and _dataSetDim = " << model->_dataSetDim << endl;
#endif
	
	if(level == 0)
		num_gene = (int) (model->_minAttributes + drand48() * (min(model->_maxAttributes, model->_dataSetDim) - model->_minAttributes));
	else if(level > 0 && level < model->_numLevels)
		num_gene = (int) (model->_minSymbionts + drand48() * (min(model->_maxSymbionts, (int)symbsToUse[level].size()) - model->_minSymbionts));
	else
		die(__FILE__, __FUNCTION__, __LINE__, "Wrong level value!");
	
	index1 = vector <int> (num_gene);
	index2 = vector <int> (num_gene, -1);
	
#ifdef DEBUGGING
	cout << "gene::gene(int) => index1.size() = " << index1.size() << endl;
#endif

#ifdef DEBUGGING
	cout << "gene::gene(int) exit." << endl;
#endif
}

/***********************************************************************************************************/

int gene::print(int t=0)
{
//	cout << "gene::print() enter." << endl;

	cout << "level = " << level << endl;
	cout << "index1 (" << index1.size() << "): ";
	for(size_t i=0; i < index1.size(); i++)
		cout << setw(4) << index1[i] << ",";
	cout << endl;
	
	cout << "index2 (" << index2.size() << "): ";
	for(size_t i=0; i < index2.size(); i++)
		cout << setw(4) << index2[i] << ",";
	cout << endl;
	
	/*
	if(level == 0)
	{
		if(t == 0)
		{
//			cout << "The individual has " << index1.size() << " genes." << endl;
			
			cout << "Attribute Index (" << index1.size() << "): ";
			for(size_t i=0; i < index1.size(); i++)
				cout << setw(4) << index1[i] << ",";
			cout << endl;
			
			cout << "Centroid Index (" << index1.size() << "): ";
			for(size_t i=0; i < index1.size(); i++)
				cout << setw(4) << index2[i] << ",";
			cout << endl;
		}
		
		else if(t == 1)
		{
			cout << "The individual has " << index1.size() << " genes." << endl;
			
			cout << "Attribute Index: ";
			for(size_t i=0; i < index1.size(); i++)
				cout << setw(4) << index1[i] << ",";
			cout << endl;
			
			cout << "Centroid Index: ";
			for(size_t i=0; i < index1.size(); i++)
				cout << setw(4) << model->_attrWiseClusCent[index1[i]][index2[i]] << ",";
			cout << endl;
		}
		
		else if(t == 2)
		{
			for(size_t i=0; i < index1.size(); i++)
				cout << setw(4) << index1[i] << ",";
			cout << endl;
			
			for(size_t i=0; i < index1.size(); i++)
				cout << setw(4) << index2[i] << ",";
			cout << endl;
		}
		
		else if(t == 4)
		{
			for(size_t i=0; i < index1.size(); i++)
				cout << setw(4) << index1[i] << ",";
			cout << endl;
			
			for(size_t i=0; i < index1.size(); i++)
				cout << setw(4) << model->_attrWiseClusCent[index1[i]][index2[i]] << ",";
			cout << endl;
		}
	}
	
	else if(level > 0 && level < model->_numLevels)
	{
		if(t == 0)
		{
			cout << "The individual has " << size() << " genes." << endl;
			
			cout << "Symbionts Index: ";
			for(size_t i=0; i < index1.size(); i++)
				cout << setw(4) << index1[i] << ",";
			cout << endl;
		}
		
		else if(t == 1)
		{
			for(size_t i=0; i < index1.size(); i++)
				cout << setw(4) << index1[i];
			cout << endl;
		}
	}
	
	else
		die(__FILE__, __FUNCTION__, __LINE__, "Wrong level value!");

	*/
	
//	cout << "gene::print() exit." << endl;
	
	return (int)index1.size();
}

/***********************************************************************************************************/

void gene::randomize()
{
#ifdef DEBUGGING
	cout << "gene::randomize() enter." << endl;
#endif
	
	if(level == 0)
	{
		// Randomizing symbionts with level == 0
		
		int j, k;
		
#ifdef DEBUGGING
		cout << "gene::randomize() => Gene size = " << index1.size() << endl;
		print(0);
#endif
		
		int new_num = (int) (model->_minAttributes + drand48() * (min(model->_maxAttributes, model->_dataSetSize) - model->_minAttributes));
		
		index1.resize(new_num);
		index2.resize(new_num);
		
#ifdef DEBUGGING
		cout << "gene::randomize() => new_num = " << new_num << endl;
		print(0);
#endif
		
		vector <int> visited(model->_dataSetDim, 0);
		
		for(int i=0; i < new_num; i++)
		{
			do
			{
				// Generate a number between 0 and _dataSetDim-1 including the bounds.
				j = (int) (drand48() * model->_dataSetDim);
#ifdef DEBUGGING
				cout << "attr = " << j << " " << visited[j] << " " << model->_noneligibleAttr[j] << " ";
#endif
			}while((visited[j] == 1) || (model->_noneligibleAttr[j] == 1));
			
			// Assigning the value for index1 vector
			index1[i] = j;
			visited[j] = 1;
			
			do
			{
				// Generate a number between 0 and _maxClusPerAttr-1, including bounds.
				k = (int) (drand48() * model->_maxClusPerAttr) % model->_numClusPerAttr[index1[i]];
#ifdef DEBUGGING
				cout << "clus = " << k << " " << model->_noneligibleClus[j][k] << " ";
#endif
			}while((int)model->_noneligibleClus[j][k] == 1);
			
			// Assigning the value for index2 vector
			index2[i] = k;
			
#ifdef DEBUGGING
			cout << "attr" << index1[i] << "clus" << index2[i] << " ";
#endif
		}
		
	}

	else if(level > 0 && level < model->_numLevels)
	{
		// Randomizing inds with level != 0
		
		int j;
		
#ifdef DEBUGGING
		cout << "gene::randomize() => size = " << index1.size() << endl;
		print(0);
#endif
		
		int new_num = (int) (model->_minSymbionts + drand48() * (min(model->_maxSymbionts, (int)symbsToUse[level].size()) - model->_minSymbionts));
		
		index1.resize(new_num);
		index2.resize(new_num, -1);
		
#ifdef DEBUGGING
		cout << "gene::randomize() => new_num = " << new_num << endl;
		print(0);
#endif
		
		vector <int> visited((int)symbsToUse[level].size(), 0);
		
#ifdef DEBUGGING
		cout << "gene::randomize() => symbsToUse[level].size() = " << symbsToUse[level].size() << endl;
		for(int i=0; i < (int)symbsToUse[level].size(); i++)
			cout << symbsToUse[level][i] << " ";
		cout << endl;
#endif
		
		for(int i=0; i < new_num; i++)
		{
			do
			{
				// Generate a number between 0 and (_numSymbToUse - 1) including bounds.
				j = (int) (drand48() * symbsToUse[level].size());
				
#ifdef DEBUGGING
				cout << "symb = " << j << " visited[" << j << "]" << visited[j] << " ";
#endif
				
			}while(visited[j] == 1);
			
#ifdef DEBUGGING
			cout << "gene::randomize() => symbsToUse[" << level << "][" << j << "] = " << symbsToUse[level][j] << endl;
#endif
			
			index1[i] = symbsToUse[level][j];
			visited[j] = 1;
			popHistory[level][index1[i]]->increfs();
			symbsRefs[level][index1[i]]++;
			
#ifdef DEBUGGING
			cout << "gene::randomize() => symb added = " << index1[i] << endl;
#endif
		}
		
		
		
		/*
		if(model->_symbsToUse == 0)
		{
			// Indexing all individuals of lower level population.
			
			vector <int> visited(2 * model->_hostPopSize, 0);
			
			for(int i=0; i < new_num; i++)
			{
				do
				{
					// A number between 0 and (2 * _hostPopSize - 1) including bounds.
					j = (int) (drand48() * 2 * model->_hostPopSize);
#ifdef DEBUGGING
					cout << "symb = " << j << " " << visited[j] << " ";
#endif
				}while(visited[j] == 1);
				
				index1[i] = j;
				visited[j] = 1;
				
#ifdef DEBUGGING
				cout << "symb added = " << index1[i] << endl;
#endif
			}
		}
		
		else if(model->_symbsToUse == 1)
		{
			
//			// Indexing only unique individuals on pareto front of lower level population.
//			
//			//cout << "bestFront.get_uniqueIndexes().size() = " << bestFront.get_uniqueIndexes().size() << endl;
//			vector <int> visited(bestFront.get_uniqueIndexes().size(), 0);
//			//cout << "bestFront.get_uniqueIndexes().size() = " << bestFront.get_uniqueIndexes().size() << endl;
//			//cout << "visited.size() = " << visited.size() << endl;
//			
//			for(int i=0; i < new_num; i++)
//			{
//				do
//				{
//					// A number between 0 and # of unique individuals including the bounds.
//					j = (int) (drand48() * bestFront.get_uniqueIndexes().size()); 
//					//cout << "Random number is: " << j << endl;
//				}while(visited[j] == 1);
//				
//				index1[i] = bestFront.get_uniqueIndexes()[j];
//				visited[j] = 1;
//			}
			
		}
		
		else
		{
			// Indexing all individuals on pareto front of lower level population.
			
			cout << "gene::randomize() => paretofront.size() = " << paretofront.size() << endl;
			
			vector <int> visited(paretofront.size(), 0);
			
			for(int i=0; i < new_num; i++)
			{
				do
				{
					// A number between 0 and paretofront.size()-1 including the bounds.
					j = (int) (drand48() * paretofront.size());
#ifdef DEBUGGING
					cout << "symb = " << j << " " << visited[j] << " ";
#endif
				}while(visited[j] == 1);
				
				// *******************************
				// SOME MODIFICATIONS NEEDED HERE.
				// *******************************
				index1[i] = paretofront[j]; // paretofront[j] is a pointer to a solution, not the integer index to a solution.
				visited[j] = 1;
				
#ifdef DEBUGGING
				cout << "symb added = " << index1[i] << endl;
#endif
			}
		}
		*/
		
		
		
	}
	
	else
		die(__FILE__, __FUNCTION__, __LINE__, "Wrong level value!");

#ifdef DEBUGGING
	cout << "gene::randomize() => Individual before prune, sort and eligibility test:" << endl;
	print(0);
#endif
	
	// Make sure no attribute is indexed more than once.
	pruneDuplicates();
	
	// Sort the gene ascending.
	sort();
	
	// Check to see if noneligible 1D cluster centroids are used while randomization.
	isEligible();
	
#ifdef DEBUGGING
	cout << "gene::randomize() => Individual after prune, sort and eligibility test:" << endl;
	print(0);
#endif
	
#ifdef DEBUGGING
	cout << "gene::randomize() exit." << endl;
#endif
}

/***********************************************************************************************************/

void gene::pruneDuplicates()
{
#ifdef DEBUGGING
	cout << "gene::pruneDuplicates() enter." << endl;
#endif
	
	if(((int)index1.size() == 2) && (index1[0] == index1[1]))
	{
		cout << "gene::pruneDuplicates() => individual found with duplicate genes:" << endl;
		print(0);
		//die(__FILE__, __FUNCTION__, __LINE__, "A gene is indexing an attribute twice!");
	}
	
	// Something is remaining to be done in this function,
	// if there are two consecutive repetition of an existing attribute, the algorithm cannot detect the second one
	// e.g. 9, 4, 7, 9, 9, 3, 5 => third 9 cannot be removed
	for(int i = 0; i < (int)index1.size(); i++)
	{
		for(int j=i+1; j < (int)index1.size(); j++)
		{
			if(index1[i] == index1[j])
			{
				
				index1.erase(index1.begin()+j);
				index2.erase(index2.begin()+j);
				j--;
				if(level != 0)
					popHistory[level][index1[i]]->decrefs();
			}
		}
	}

#ifdef DEBUGGING
	cout << "gene::pruneDuplicates() exit." << endl;
#endif
}

/***********************************************************************************************************/

bool gene::pruneSymbsWithZeroInsts()
{
#ifdef DEBUGGING
	cout << "gene::pruneSymbsWithZeroInsts() enter." << endl;
#endif
	
	bool isModified = false;
	
	// If there is a host is using a symbiont with 0 instances assigned to it.
	for(int i=0; i < (int)size(); i++)
	{
		if(index2[i] == 0)
		{
			isModified = true;
			
			if(level != 0)	
				popHistory[level][index1[i]]->decrefs();
			
			index1.erase(index1.begin()+i);
			index2.erase(index2.begin()+i);
			
			i--;
		}
	}

#ifdef DEBUGGING
	cout << "gene::pruneSymbsWithZeroInsts() exit." << endl;
#endif
	
	return modified;
}

/***********************************************************************************************************/

void gene::sort()
{
#ifdef DEBUGGING
	cout << "gene::sort() enter." << endl;
#endif
	
	int tmp;
	
	if(index1.size() > 1)
	{
		for(int i=0; i < (int)index1.size(); i++)
		{
			for(int j=i+1; j < (int)index1.size(); j++)
			{
				if(index1[i] > index1[j])
				{
					tmp = index1[i];
					index1[i] = index1[j];
					index1[j] = tmp;
					
					tmp = index2[i];
					index2[i] = index2[j];
					index2[j] = tmp;
				}
			}
		}
	}

#ifdef DEBUGGING
	cout << "gene::sort() exit." << endl;
#endif
}

/***********************************************************************************************************/

bool gene::isEligible()
{
	// Checks if the individual uses noneligible attributes or 1-D cluster centroids.
	
#ifdef DEBUGGING
	cout << "gene::isEligible() enter." << endl;
#endif
	
	if(level == 0)
	{
		for(int i=0; i < (int)index1.size(); i++)
		{
			if(model->_noneligibleAttr[index1[i]] == 1)
			{
				cout << "gene::isEligible() ERROR => this ind uses noneligible attribute " << index1[i] << endl;
				return false;
			}
			else if(model->_noneligibleClus[index1[i]][index2[i]] == 1)
			{
				cout << "gene::isEligible() ERROR => this ind uses attribute " << index1[i] << " with noneligible cluster centroid " << index2[i] << endl;
				return false;
			}
		}
		
#ifdef DEBUGGING
		cout << "gene::isEligible() exit." << endl;
#endif
		
		return true;
	}
	
	else if(level > 0 && level < model->_numLevels)
	{
#ifdef DEBUGGING
		cout << "gene::isEligible() exit." << endl;
#endif
		return true;
	}
	
	else
		die(__FILE__, __FUNCTION__, __LINE__, "Wrong level value!");

}

/***********************************************************************************************************/

bool gene::operator== (const gene& g1) const
{
	// Checks if two solutions have identical genes.
	// This function assumes that both genes are pruned (no duplicate values) and sorted.
	if(index1.size() != g1.index1.size())
		return false;
	
	if(level == 0)
	{
		if(index1 != g1.index1 || index2 != g1.index2)
			return false;
		else
			return true;
	}

	else if(level > 0 && level < model->_numLevels)
	{
		if(index1 != g1.index1)
			return false;
		else
			return true;
	}
	
	else
		die(__FILE__, __FUNCTION__, __LINE__, "Wrong level value!");
}

/***********************************************************************************************************/

/*
bool gene::equals(const gene& g1) const
{
	// Checks if two genes have identical genes, a substitue for the == operator overload.
	// This function assumes that both genes are pruned (no duplicate values) and sorted.

	cout << "gene::equals() enter." << endl;
	
	if(index1.size() != g1.index1.size())
		return false;
	
	if(level == 0)
	{
		for(size_t i=0; i < index1.size(); i++)
		{
			if((index1[i] != g1.index1[i]) || (index2[i] != g1.index2[i]))
				return false;
		}
		
	}
	
	else if(level > 0 && level < model->_numLevels)
	{
		for(size_t i=0; i < index1.size(); i++)
		{
			if(index1[i] != g1.index1[i])
				return false;
		}
		
	}
	
	else
		die(__FILE__, __FUNCTION__, __LINE__, "Wrong level value!");

	return true;
}
*/

/***********************************************************************************************************/

void gene::crossover( gene& p1,  gene& p2)
{
	if(level == 0)
	{
		crossover_level0_1point(p1, p2);
		//crossover_level0_exchange(p1->sol, p2->sol); 
	}
	
	else if(level > 0 && level < model->_numLevels)
	{
		crossover_level1_1point(p1, p2);
	}
	
	else
		die(__FILE__, __FUNCTION__, __LINE__, "Wrong level value!");
}

/***********************************************************************************************************/

void gene::crossover_level0_1point( gene& p1,  gene& p2)
{
	// Crossover operator for individual level == 0
	// Simple one-point crossover
	// Crossover probability has been applied in BreedGeneration function
	
#ifdef DEBUGGING
	cout << "gene::crossover_type0_1point() enter." << endl;
#endif
	
#ifdef DEBUGGING
	// Printing both parents
	cout << "gene::crossover_type0_1point() => Father is:" << endl;
	p1.print(0);
	cout << "gene::crossover_type0_1point() => Mother is:" << endl;
	p2.print(0);
#endif
	
	// If both parents are the same return one of them as the new child.
	if(p1 == p2)
	{
		//cout << "gene::crossover_type0_1point() => Both parents are the same return one of them as the new child!" << endl;

		index1 = p1.index1;
		index2 = p1.index2;
		return;
	}
	
	// Special case, if both parent are equal of size 2.
	if((int)p1.size()==2 && (int)p2.size()==2)
	{
		index1.resize(p1.size());
		index2.resize(index1.size());
		
		if((p1.index1[0] != p2.index1[1]) && (p1.index1[1] != p2.index1[0]))
		{
			if(drand48() < 0.5)
			{
				//cout << "gene::crossover_type1_1point() => Case 1" << endl;
				
				index1[0] = p1.index1[0];
				index2[0] = p1.index2[0];
				index1[1] = p2.index1[1];
				index2[1] = p2.index2[1];
			}
			else
			{
				//cout << "gene::crossover_type1_1point() => Case 2" << endl;
				
				index1[0] = p2.index1[0];
				index2[0] = p2.index2[0];
				index1[1] = p1.index1[1];
				index2[1] = p1.index2[1];
			}
		}
		
		else if(p1.index1[0] != p2.index1[1])
		{
			//cout << "gene::crossover_type1_1point() => Case 3" << endl;
			
			index1[0] = p1.index1[0];
			index2[0] = p1.index2[0];
			index1[1] = p2.index1[1];
			index2[1] = p2.index2[1];
		}
		
		else if(p1.index1[1] != p2.index1[0])
		{
			//cout << "gene::crossover_type1_1point() => Case 4" << endl;
			
			index1[0] = p2.index1[0];
			index2[0] = p2.index2[0];
			index1[1] = p1.index1[1];
			index2[1] = p1.index2[1];
		}
		
		else // both parents are the same
		{
			index1 = p1.index1;
			index2 = p1.index2;
		}
		
		// Sort the host.
		sort();
		
		//Prune the host.
		pruneDuplicates();
		
#ifdef DEBUGGING
		cout << "gene::crossover_type0_1point() => The new individual has " << index1.size() << " genes." << endl;
		for(int kh=0; kh < (int)index1.size(); kh++)
			cout << index1[kh] << ", ";
		cout << endl;
#endif
		
		return;
		
	}
	
	int startgene, componentcross;
	vector <int> :: iterator it, it2;
	
	startgene = (int) (drand48() * 2);

	// Generate a number between 1 and min(p1.size(),p2.size())-1 including the bounds
	componentcross = (int) (drand48() * (min(p1.size(), p2.size())-1) + 1);
	
#ifdef DEBUGGING
	cout << "gene::crossover_type0_1point() => Starting parent: " << startgene << " and crossover point: " << componentcross << endl;
#endif
	
	if(startgene == 0)
	{
		index1.resize(p2.size());
		index2.resize(p2.size());
		
		it = copy(p1.index1.begin(), p1.index1.begin()+componentcross, index1.begin());
		copy(p2.index1.begin()+componentcross, p2.index1.end(), index1.begin()+componentcross);
		it2 = copy(p1.index2.begin(), p1.index2.begin()+componentcross, index2.begin());
		copy(p2.index2.begin()+componentcross, p2.index2.end(), index2.begin()+componentcross);
		
	}
	
	else if(startgene == 1)
	{
		index1.resize(p1.size());
		index2.resize(p1.size());
		
		it = copy(p2.index1.begin(), p2.index1.begin()+componentcross, index1.begin());
		copy(p1.index1.begin()+componentcross, p1.index1.end(), index1.begin()+componentcross);
		it2 = copy(p2.index2.begin(), p2.index2.begin()+componentcross, index2.begin());
		copy(p1.index2.begin()+componentcross, p1.index2.end(), index2.begin()+componentcross);
		
	}
	
	else
		die(__FILE__, __FUNCTION__, __LINE__, "Wrong startgene value!!! startgene != 0 && startgene != 1");
	
	// Sort the new indiviual.
	sort();
	
	// Prune the new individual.
	pruneDuplicates();
	
	// Individuals with less than _minAttributes attributes should be fixed.
	// As the first treatment change the order of the parents in producing the children.
	if((int)index1.size() < model->_minAttributes)
	{
		
#ifdef DEBUGGING
		cout << "gene::crossover_type0_1point() => The new individual has " << index1.size();
		cout << " genes, therefore reversing the crosover order!" << endl;
		for(int kh=0; kh < (int)index1.size(); kh++)
			cout << index1[kh] << ", ";
		cout << endl;
		for(int kh=0; kh < (int)index2.size(); kh++)
			cout << index2[kh] << ", ";
		cout << endl;
#endif
		
		if(startgene == 0) 
		{
			index1.resize(p1.size());
			index2.resize(p1.size());
			
			it = copy(p2.index1.begin(), p2.index1.begin()+componentcross, index1.begin());
			copy(p1.index1.begin()+componentcross, p1.index1.end(), index1.begin()+componentcross);
			it2 = copy(p2.index2.begin(), p2.index2.begin()+componentcross, index2.begin());
			copy(p1.index2.begin()+componentcross, p1.index2.end(), index2.begin()+componentcross);
		}
		
		else
		{
			index1.resize(p2.size());
			index2.resize(p2.size());
			
			it = copy(p1.index1.begin(), p1.index1.begin()+componentcross, index1.begin());
			copy(p2.index1.begin()+componentcross, p2.index1.end(), index1.begin()+componentcross);
			it2 = copy(p1.index2.begin(), p1.index2.begin()+componentcross, index2.begin());
			copy(p2.index2.begin()+componentcross, p2.index2.end(), index2.begin()+componentcross);
		}
		
		// Sort the new indiviual.
		sort();
		
		// Prune the new individual.
		pruneDuplicates();
		
	}
	
	// The second treatment is to replace the new individual with the shorter parent.
	if((int)index1.size() < model->_minAttributes)
	{
		
#ifdef DEBUGGING
		cout << "gene::crossover_type0_1point() => The new individual has " << index1.size();
		cout << " genes, and still needs to be fixed!" << endl;
		for(int kh=0; kh < (int)index1.size(); kh++)
			cout << index1[kh] << ", ";
		cout << endl;
		for(int kh=0; kh < (int)index2.size(); kh++)
			cout << index2[kh] << ", ";
		cout << endl;
#endif
		
		// If new host is shorter than 2, replace it with shorter parent
		if((int)p1.size() <= (int)p2.size()) 
		{
			index1.resize(p1.size());
			index2.resize(p1.size());
			index1 = p1.index1;
			index2 = p1.index2;
		}
		else
		{
			index1.resize(p2.size());
			index2.resize(p2.size());
			index1 = p2.index1;
			index2 = p2.index2;
		}
	}
	
	if((int)index1.size() < model->_minAttributes)
		die(__FILE__, __FUNCTION__, __LINE__, "The individual's size < _minAttributes after crossover!");
	
	
#ifdef DEBUGGING
	cout << "gene::crossover_type0_1point() exit." << endl;
#endif
}

/***********************************************************************************************************/

void gene::crossover_level0_exchange( gene& p1,  gene& p2)
{
	// crossover operator for gene_type_1
	// finds a common attr between two inds and swap the clst indxes between two inds 
	// crossover probability has been applied in BreedGeneration function
	
	// If both parents are the same return one of them as the new child.
	if(p1 == p2)
	{
		//cout << "gene::crossover_type0_exchange() => Both parents are the same return one of them as the new child!" << endl;
		index1 = p1.index1;
		index2 = p1.index2;
		return;
	}
	
	int flag = 0;
	int rand1;
	int temp;
	
	vector <int> b1(p1.index1.size());
	
	// shuffle the indices - permutation of 1 to index1.size()-1 
	for(vector <int>::size_type i=0; i < b1.size(); i++)
	{
		b1[i] = i;
	}
	
	for(vector<int>::size_type i=0; i < b1.size(); i++)
	{
		rand1 = (int) (i + (drand48() * (b1.size() - i)));
		temp = b1[rand1];
		b1[rand1] = b1[i];
		b1[i] = temp;
	}
	
	index1.resize(p1.index1.size());
	index2.resize(p1.index2.size());
	index1 = p1.index1;
	index2 = p1.index2;
	
	for(vector<int>::size_type i=0; i < p1.index1.size(); i++)
	{
		for(vector<int>::size_type j=0; j < p2.index1.size(); j++)
		{
			if(p1.index1[b1[i]] == p2.index1[j] && p1.index2[b1[i]] != p2.index2[j])
			{
				index2[b1[i]] = p2.index2[j];
				//child1->index2[b1[i]] = parent2->index2[j];
				//child2->index2[j] = parent1->index2[b1[i]];
				if(index2[b1[i]] >= model->_numClusPerAttr[index1[b1[i]]])
				{
					cout << "Wrong cluster index " << index2[b1[i]] << " generated in Crossover operation for attribute " << index1[b1[i]] << endl;
					exit(1);
				}
				flag = 1;
				break;
			}
		}
		if(flag == 1)
		{
			break;
		}
	}
	
}

/***********************************************************************************************************/

/*
void gene::crossover_exchange_fixed_inds(const gene& p1, const gene& p2)
{

	// crossover probability has been applied in BreedGeneration function
	int flag = 0;
	int rand1;
	
	// find a common attr between two inds
	// and swap the clst indxes between two inds 
	vector<int> b1(index1.size());
	
	// shuffle the indices - permutation of 1 to index1.size()-1 
	for(vector<int>::size_type i=0; i < b1.size(); i++)
	{
		b1[i] = i;
	}
	
	for (vector<int>::size_type i=0; i < b1.size(); i++)
	{
		rand1 = ran1int(i, b1.size()-1);
		temp = b1[rand1];
		b1[rand1] = b1[i];
		b1[i] = temp;
	}

		index1 = p1.index1;
		index2 = p1.index2;
		
		for(vector<int>::size_type i=0; i < b1.size(); i++)
		{
			for(vector<int>::size_type j=0; j < b1.size(); j++)//j=0; j < parent2->n_gene; j++
			{
				if(p1.index1[b1[i]] == p2.index1[j] && p1.index2[b1[i]] != p2.index2[j])
				{
					index2[b1[i]] = p2.index2[j];
//					child1->index2[b1[i]] = parent2->index2[j];
//					child2->index2[j] = parent1->index2[b1[i]];
					if(index2[b1[i]] >= _numClusPerAttr[index1[b1[i]]])
					{
						cout << "Wrong cluster index " << index2[b1[i]] << " generated in Crossover operation for attribute " << index1[b1[i]] << endl;
						exit(1);
					}
					flag = 1;
					break;
				}
			}
			if(flag == 1)
			{
				break;
			}
		}
}
*/

/***********************************************************************************************************/

void gene::crossover_level1_1point( gene& p1,  gene& p2)
{
	// Crossover operator for individual level != 0
	// Simple one-point crossover
	// Crossover probability has been applied in BreedGeneration function
	
#ifdef DEBUGGING
	cout << "gene::crossover_type1_1point() enter." << endl;
#endif
	
#ifdef DEBUGGING
	// Printing both parents
	cout << "gene::crossover_type0_1point() => Father is:" << endl;
	p1.print(0);
	cout << "gene::crossover_type0_1point() => Mother is:" << endl;
	p2.print(0);
#endif
	
	// If both parents are the same return one of them as the new child.
	if(p1 == p2)
	{
		//cout << "gene::crossover_type1_1point() => Both parents are the same return one of them as the new child!" << endl;
		
		index1 = p1.index1;
		index2 = p1.index2;
		return;
	}
	
	// Special case, if both parent are equal of size 2.
	if((int)p1.size()==2 && (int)p2.size()==2)
	{
		index1.resize(p1.size());
		index2.resize(index1.size());

		if((p1.index1[0] != p2.index1[1]) && (p1.index1[1] != p2.index1[0]))
		{
			if(drand48() < 0.5)
			{
				//cout << "gene::crossover_type1_1point() => Case 1" << endl;
				
				index1[0] = p1.index1[0];
				index1[1] = p2.index1[1];
			}
			else
			{
				//cout << "gene::crossover_type1_1point() => Case 2" << endl;
				
				index1[0] = p2.index1[0];
				index1[1] = p1.index1[1];
			}
		}
		
		else if(p1.index1[0] != p2.index1[1])
		{
			//cout << "gene::crossover_type1_1point() => Case 3" << endl;
			
			index1[0] = p1.index1[0];
			index1[1] = p2.index1[1];
		}
		
		else if(p1.index1[1] != p2.index1[0])
		{
			//cout << "gene::crossover_type1_1point() => Case 4" << endl;
			
			index1[0] = p2.index1[0];
			index1[1] = p1.index1[1];
		}
		
		else // both parents are the same
		{
			index1 = p1.index1;
		}
		
		// Sort the host.
		sort();
		
		//Prune the host.
		pruneDuplicates();
		
#ifdef DEBUGGING
		cout << "gene::crossover_type1_1point() => The new individual has " << index1.size() << " genes." << endl;
		for(int kh=0; kh < (int)index1.size(); kh++)
			cout << index1[kh] << ", ";
		cout << endl;
#endif
		
		return;
		
	}
	
	int componentcross, startgene;
	vector <int> :: iterator it, it2;
	
	startgene = (int) (drand48() * 2);
	
	// Generate a number between 1 and min(p1.size(),p2.size())-1 including the bounds
	componentcross = (int) (drand48() * (min(p1.size(), p2.size())-1) + 1);
	
	//cout << "gene::crossover_type1_1point() => Starting parent: " << startgene << " and crossover point: " << componentcross << endl;
	
	if(startgene == 0) 
	{
		index1.resize(p2.size());
		index2.resize(p2.size());
		
		it = copy(p1.index1.begin(), p1.index1.begin()+componentcross, index1.begin());
		copy(p2.index1.begin()+componentcross, p2.index1.end(), index1.begin()+componentcross);
		
		/*
		for(int i=0; i <= componentcross; i++)
			index1[i] = p1.index1[i];
		
		for(int i=componentcross+1; i < (int)index1.size(); i++)
			index1[i] = p2.index1[i];
		*/
	}
	
	else if(startgene == 1)
	{
		index1.resize(p1.size());
		index2.resize(p1.size());
		
		it = copy(p2.index1.begin(), p2.index1.begin()+componentcross, index1.begin());
		copy(p1.index1.begin()+componentcross, p1.index1.end(), index1.begin()+componentcross);

		/*
		for(int i=0; i <= componentcross; i++)
			index1[i] = p2.index1[i];
		
		for(int i=componentcross+1; i < (int)index1.size(); i++)
			index1[i] = p1.index1[i];
		*/
	}
	
	else
		die(__FILE__, __FUNCTION__, __LINE__, "Wrong startgene value!!! startgene != 0 && startgene != 1");
	
	
	// Sort the new indiviual.
	sort();
	
	// Prune the new individual.
	pruneDuplicates();
	
	// Individuals with less than _minSymbionts symbionts should be fixed.
	// As the first treatment change the order of the parents in producing the children.
	if((int)index1.size() < model->_minSymbionts)
	{
#ifdef DEBUGGING
		cout << "gene::crossover_type1_1point() => The new individual has " << index1.size();
		cout << " genes, therefore reversing the crosover order!" << endl;
		for(int kh=0; kh < (int)index1.size(); kh++)
			cout << index1[kh] << ", ";
		cout << endl;
#endif
		
		if(startgene == 0) 
		{
			index1.resize(p1.size());
			index2.resize(p1.size());
			
			it = copy(p2.index1.begin(), p2.index1.begin()+componentcross, index1.begin());
			copy(p1.index1.begin()+componentcross, p1.index1.end(), index1.begin()+componentcross);
		}
		
		else
		{
			index1.resize(p2.size());
			index2.resize(p2.size());
			
			it = copy(p1.index1.begin(), p1.index1.begin()+componentcross, index1.begin());
			copy(p2.index1.begin()+componentcross, p2.index1.end(), index1.begin()+componentcross);
		}
		
		// Sort the new indiviual.
		sort();
		
		// Prune the new individual.
		pruneDuplicates();
		
	}
	
	// The second treatment is to replace the new individual with the shorter parent.
	if((int)index1.size() < model->_minSymbionts)
	{
		
#ifdef DEBUGGING
		cout << "gene::crossover_type1_1point() => The new individual has " << index1.size();
		cout << " genes, and still needs to be fixed!" << endl;
		for(int kh=0; kh < (int)index1.size(); kh++)
			cout << index1[kh] << ", ";
		cout << endl;
#endif
		
		// If new host is shorter than 2, replace it with shorter parent
		if((int)p1.size() <= (int)p2.size()) 
		{
			index1.resize(p1.size());
			index2.resize(p1.size());
			index1 = p1.index1;
		}
		else
		{
			index1.resize(p2.size());
			index2.resize(p2.size());
			index1 = p2.index1;
		}
	}
	
	if((int)index1.size() < model->_minSymbionts)
		die(__FILE__, __FUNCTION__, __LINE__, "The individual's size < model->_minSymbionts after crossover!");
	
#ifdef DEBUGGING
	cout << "gene::crossover_type1_1point() => The new individual has " << index1.size() << " genes." << endl;
	for(int kh=0; kh < (int)index1.size(); kh++)
		cout << index1[kh] << ", ";
	cout << endl;
#endif
	
#ifdef DEBUGGING
	cout << "gene::crossover_type1_1point() exit." << endl;
#endif
}

/***********************************************************************************************************/

void gene::mutation()
{
	// This function is changed so that if the ind's size is equal to minimum value, only use a mutation to increase the size
	// and if the ind's size is equal to the maximum value, only use a mutation to decrease the size of the individual.
	
	double p;

#ifdef DEBUGGING
	cout << "gene::mutation() => temprand = " << temprand << endl;
#endif
	
	// This piece of code applies atmoic mutations to the indivdual one after the other.
	if(level == 0)
	{
		for(p = 1.0; drand48() < p && size() > min(model->_minAttributes, model->_dataSetDim); p = p * model->_pad)
		{
			//cout << "gene::mutation() => removing an attribute" << endl;
			mutation_level0_removeAttr();
			//print(0);
		}
		
		for(p = 1.0; drand48() < p && size() < min(model->_maxAttributes, model->_dataSetDim); p = p * model->_paa)
		{
			//cout << "gene::mutation() => adding an attribute" << endl;
			mutation_level0_addAttr();
			//print(0);
		}
		
		/*
		for(p = 1.0; drand48() < p; p = p * model->_plc)
		{
			//cout << "gene::mutation() => replacing 1D centroid of an attribute" << endl;
			mutation_level0_replaceClus();
			//print(0);
		}
		*/
	}
	
	else if(level > 0 && level < model->_numLevels)
	{
		for(p = 1.0; drand48() < p && size() > min(model->_minSymbionts, (int)symbsToUse[level].size()); p = p * model->_psd)
		{
			//cout << "gene::mutation() => removing a symbiont" << endl;
			mutation_level1_removeSymb();
			//print(0);
		}
		
		for(p = 1.0; drand48() < p && size() < min(model->_maxSymbionts, (int)symbsToUse[level].size()); p = p * model->_psa)
		{
			//cout << "gene::mutation() => adding a symbiont" << endl;
			mutation_level1_addSymb();
			//print(0);
		}
		
		/*
		for(p = 1.0; drand48() < p && size() > min(model->_minSymbionts, (int)symbsToUse[level].size()); p = p * model->_pss)
		{
			//cout << "gene::mutation() => removing a symbiont" << endl;
			mutation_level1_replaceSymb();
			//print(0);
		}
		*/
	}
	
	else
		die(__FILE__, __FUNCTION__, __LINE__, "Wrong level value!");
	
/*	
	// This piece of code only applies one mutation to the indivdual.
 
	if(level == 0)
	{
		if(size() == min(model->_minAttributes, model->_dataSetDim))
		{
			mutation_level0_addAttr();
		}
		
		else if(size() == min(model->_maxAttributes, model->_dataSetDim))
		{
			mutation_level0_removeAttr();
		}
		
		else
		{
			if(temprand < 0.25)
				mutation_level0_replaceAttr();
			else if(temprand >= 0.25 && temprand < 0.5)
				mutation_level0_replaceClus();
			else if(temprand >= 0.5 && temprand < 0.75)
				mutation_level0_removeAttr();
			else
				mutation_level0_addAttr();
		}
	}
	
	else if(level > 0 && level < model->_numLevels)
	{
		if(size() == min(model->_minSymbionts, (int)symbsToUse[level].size()))
		{
			mutation_level1_addSymb();
		}
		
		else if(size() == min(model->_maxSymbionts, (int)symbsToUse[level].size()))
		{
			mutation_level1_removeSymb();
		}
		
		else
		{
			if(temprand < 0.33)
				mutation_level1_replaceSymb();
			else if(temprand >= 0.33 && temprand < 0.66)
				mutation_level1_removeSymb();
			else
				mutation_level1_addSymb();
		}
	}
	
	else
		die(__FILE__, __FUNCTION__, __LINE__, "Wrong level value!");
*/
}		
		
/***********************************************************************************************************/

void gene::mutation_level0_removeAttr()
{
	// Mutation operator for individual level == 0
	// Removes an attribute from symbiont.
	// Mutation probability has been applied in BreedGeneration function.
	
#ifdef DEBUGGING
	cout << "gene::mutation_type0_removeAttr() enter." << endl;
#endif
	
#ifdef DEBUGGING
	cout << "gene::mutation_type0_removeAttr() => The symbiont to be mutated by removal: " << endl;
	print(0);
#endif
	
	if((int)index1.size() > min(model->_minAttributes, model->_dataSetDim))
	{
		int gene_indx;
		
		gene_indx = (int) (drand48() * index1.size());
		
#ifdef DEBUGGING
		cout << "gene::mutation_type0_removeAttr() => The gene index to be removed is: " << gene_indx << " and its attribute is: ";
		cout << index1[gene_indx] << endl;
#endif
		
		index1.erase(index1.begin() + gene_indx);
		index2.erase(index2.begin() + gene_indx);
	}
	
	else
	{
		cout << "gene::mutation_type0_removeAttr() => The ind is too small, hence nothing is removed. Return the parent." << endl;
		return;
	}
	
#ifdef DEBUGGING
	cout << "gene::mutation_type0_removeAttr() exit." << endl;
#endif
}

/***********************************************************************************************************/

void gene::mutation_level0_addAttr()
{
	// Mutation operator for individual level == 0
	// Adds an attribute which doesn't exist in symbiont and selects one of its cluster centroids.
	// Mutation probability has been applied in BreedGeneration function.
	
#ifdef DEBUGGING
	cout << "gene::mutation_type0_addAttr() enter." << endl;
#endif
	
#ifdef DEBUGGING
	cout << "gene::mutation_type0_addAttr() => The symbiont to be mutated by addition: " << endl;
	print(0);
#endif
	
	if((int)index1.size() < min(model->_maxAttributes, model->_dataSetDim))
	{
		int new_attr, new_clus, max_tries;
		double temp_rand, accumulative_prob;
		vector <int> :: iterator it, it2;
		
		max_tries = 0;
		
		// Find an attribute which doesn't exist in symbiont.
		do
		{
			if(model->_useAttrProbLrng)
			{
				temp_rand = drand48();
				accumulative_prob = model->_attrProbs[0];
				for(int i=0; i < model->_dataSetDim; i++)
				{
					if(temp_rand <= accumulative_prob)
					{
						new_attr = i;
						break;
					}
					
					accumulative_prob += model->_attrProbs[i+1];
				}
			}
			else
			{
				// replaced by piece of code above to account for the attribute probabilitie dynamically.
				new_attr = (int) (drand48() * model->_dataSetDim);
			}
#ifdef DEBUGGING
			cout << "gene::mutation_type0_addAttr() => The new attribute which might not be used: " << new_attr << endl;
#endif
			
			max_tries++;
			if(max_tries > model->_maxAttempts)
				return;
			
		}while((find(index1.begin(), index1.end(), new_attr) != index1.end()) || ((int)model->_noneligibleAttr[new_attr] == 1));
		
		max_tries = 0;
		
		// Find a cluster centroid within the attribute just selected.
		do
		{
			// Generate a number between 0 and _maxClusPerAttr-1, including bounds.
			new_clus = (int) (drand48() * model->_maxClusPerAttr) % model->_numClusPerAttr[new_attr];
			
#ifdef DEBUGGING
			cout << "gene::mutation_type0_addAttr() => The new cluster centroid which might not be used: " << new_clus << endl;
#endif
			
			max_tries++;
			if(max_tries > model->_maxAttempts)
				return;
			
		}while((int)model->_noneligibleClus[new_attr][new_clus] == 1);
		
		int done = 0;
		
		// Find the appropriate position for the new attribute.
		for(it = index1.begin(), it2 = index2.begin(); it < index1.end(); it++, it2++)
		{
			if(new_attr < (*it))
			{
				index1.insert(it, new_attr);
				index2.insert(it2, new_clus);
				done = 1;
				break;
			}
		}
		
		// If the new_attr > all attributes indexed by symbiont.
		if(!done)
		{
			//cout << "gene::mutation_type0_addAttr() => This attr is greater than any attr indexed by ind." << endl;
			
			index1.push_back(new_attr);
			index2.push_back(new_clus);
		}
		
#ifdef DEBUGGING
		cout << "gene::mutation_type0_addAttr() => new attr added is " << new_attr;
		cout << " and the new cluster centroid is " << new_clus << endl;
#endif
		
		// No need to call pruneDuplicates() since duplicate attributes were not allowed in this function.
		// No need to sort because it is taken care of in the current function. 
		// sort();
	}
	
	else
	{
		cout << "gene::mutation_type0_addAttr() => The ind is too large, hence nothing is added. Return the parent." << endl;
		return;
	}
	
#ifdef DEBUGGING
	cout << "gene::mutation_type0_addAttr() exit." << endl;
#endif
}

/***********************************************************************************************************/

void gene::mutation_level0_replaceAttr()
{
	// Mutation operator for individual level == 0
	// Replaces an attribute with one which doesn't exist in individual and selects one of its cluster centroids.
	// Mutation probability has been applied in BreedGeneration function.
	
#ifdef DEBUGGING
	cout << "gene::mutation_type0_replaceAttr() enter." << endl;
#endif
	
#ifdef DEBUGGING
	cout << "gene::mutation_type0_replaceAttr() => The symbiont to be mutated by replacing an atribute: " << endl;
	print(0);
#endif
	
	int gene_indx = (int) (drand48() * index1.size());
	
	int max_tries;
	
	//cout << "gene::mutation_type0_replaceAttr() => The gene index to be changed is: " << gene_indx;
	//cout << " and its attribute is: " << index1[gene_indx] << endl;
	
	if((int)index1.size() != model->_dataSetDim)
	{
		int new_attr, new_clus;
		
		max_tries = 0;
		
		// Find an attribute which doesn't exist in symbiont.
		do
		{
			// Generate a number between 0 and _dataSetDim, including bounds.
			new_attr = (int) (drand48() * model->_dataSetDim);
			
#ifdef DEBUGGING
			cout << "gene::mutation_type0_replaceAttr() => The new attribute which might not be used: " << new_attr << endl;
#endif
			
			max_tries++;
			if(max_tries > model->_maxAttempts)
				return;
			
		}while((find(index1.begin(), index1.end(), new_attr) != index1.end()) || ((int)model->_noneligibleAttr[new_attr] == 1));
		
		max_tries = 0;
		
		// Find a cluster centroid within the attribute just selected.
		do
		{
			// Generate a number between 0 and _maxClusPerAttr-1, including bounds.
			new_clus = (int) (drand48() * model->_maxClusPerAttr) % model->_numClusPerAttr[new_attr];
			
#ifdef DEBUGGING
			cout << "gene::mutation_type0_replaceAttr() => The new cluster centroid which might not be used: " << new_clus << endl;
#endif
			
			max_tries++;
			if(max_tries > model->_maxAttempts)
				return;
			
		}while((int)model->_noneligibleClus[new_attr][new_clus] == 1);
		
		index1[gene_indx] = new_attr;
		index2[gene_indx] = new_clus;
		
#ifdef DEBUGGING
		cout << "gene::mutation_type0_replaceAttr() => new attr of gene " << gene_indx << " is replaced with " << index1[gene_indx];
		cout << " and the new cluster centroid is " << index2[gene_indx] << endl;
#endif
		
		// No need to call pruneDuplicates() since duplicate attributes were not allowed in this function.
		// Sort the ind. 
		sort();
	}
	
	else
	{
		cout << "gene::mutation_type0_replaceAttr() => The ind is using all attributes, hence no replace is possible. Return the parent." << endl;
		return;
	}
	
#ifdef DEBUGGING
	cout << "gene::mutation_type0_replaceAttr() exit." << endl;
#endif
}

/***********************************************************************************************************/

void gene::mutation_level0_replaceClus()
{
	// Mutation operator for individual level == 0
	// Replaces an attribute with one which doesn't exist in individual and selects one of its cluster centroids.
	// Mutation probability has been applied in BreedGeneration function.
	
#ifdef DEBUGGING
	cout << "gene::mutation_type0_replaceClus() enter." << endl;
#endif
	
#ifdef DEBUGGING
	cout << "gene::mutation_type0_replaceClus() => The symbiont to be mutated by replacing a cluster centroid: " << endl;
	print(0);
#endif
	
	int gene_indx = (int) (drand48() * index1.size());
	
#ifdef DEBUGGING
	cout << "gene::mutation_type0_replaceClus() => The gene index to be replaced is: " << gene_indx;
	cout << ", its attribute is: " << index1[gene_indx] << " and its cluster centroid is: " << index2[gene_indx] << endl;
#endif
	
	int new_clus, max_tries;
	
	max_tries = 0;
	
	// Find a cluster centroid within the attribute just selected.
	do
	{
		// Generate a number between 0 and _maxClusPerAttr-1, including bounds.
		new_clus = (int) (drand48() * model->_maxClusPerAttr) % model->_numClusPerAttr[index1[gene_indx]];
		
#ifdef DEBUGGING
		cout << "gene::mutation_type0_replaceClus() => The new cluster centroid which might not be used: " << new_clus << endl;
#endif
		
		max_tries++;
		if(max_tries > model->_maxAttempts)
			return;
		
	}while(((int)model->_noneligibleClus[index1[gene_indx]][new_clus] == 1) || (new_clus == index2[gene_indx]));
	
	index2[gene_indx] = new_clus;
	
#ifdef DEBUGGING
	cout << "gene::mutation_type0_replaceClus() => the attr of gene " << gene_indx << " is " << index1[gene_indx];
	cout << " and the new cluster centroid is replaced with " << index2[gene_indx] << endl;
#endif
	
#ifdef DEBUGGING
	cout << "gene::mutation_type0_replaceClus() exit." << endl;
#endif
}

/***********************************************************************************************************/

void gene::mutation_level1_removeSymb()
{
	// Mutation operator for individual level != 0
	// Removes a symbiont from the host.
	// Mutation probability has been applied in BreedGeneration function.
	
#ifdef DEBUGGING
	cout << "gene::mutation_type1_removeSymb() enter." << endl;
#endif
	
#ifdef DEBUGGING
	cout << "gene::mutation_type1_removeSymb() => The host to be mutated by removal: " << endl;
	print(0);
#endif
	
	if((int)index1.size() > min(model->_minSymbionts, (int)symbsToUse[level].size()))
	{
		int gene_indx;
		
		gene_indx = (int) (drand48() * index1.size());
		
#ifdef DEBUGGING
		cout << "gene::mutation_type1_removeSymb() => The gene index to be removed is: " << gene_indx << " and its symbiont is: ";
		cout << index1[gene_indx] << endl;
#endif
		
		//popHistory[level][index1[gene_indx]]->decrefs();
		index1.erase(index1.begin() + gene_indx);
		index2.erase(index2.begin() + gene_indx);
	}
	
	else
	{
		cout << "gene::mutation_type1_removeSymb() => The ind is too small, hence nothing is removed. Return the parent." << endl;
		return;
	}
	
#ifdef DEBUGGING
	cout << "gene::mutation_type1_removeSymb() exit." << endl;
#endif
}

/***********************************************************************************************************/

void gene::mutation_level1_addSymb()
{
	// Mutation operator for individual level != 0
	// Adds a symbiont which doesn't exist in host.
	// Mutation probability has been applied in BreedGeneration function.
	
#ifdef DEBUGGING
	cout << "gene::mutation_type1_addSymb() enter." << endl;
#endif
	
#ifdef DEBUGGING
	cout << "gene::mutation_type1_addSymb() => The host to be mutated by addition: " << endl;
	print(0);
#endif
	
	if((int)index1.size() < min(model->_maxSymbionts, (int)symbsToUse[level].size()))
	{
		int temp, new_symb, max_tries;
		vector <int> :: iterator it, it2;
		
		max_tries = 0;
		
		// Find a symbiont which doesn't exist in the host.
		do
		{
			temp = (int) (drand48() * (int)symbsToUse[level].size());
			new_symb = symbsToUse[level][temp];
			
#ifdef DEBUGGING
			cout << "gene::mutation_type1_addSymb() => The new attribute which might not be used: " << new_symb << endl;
#endif
			
			max_tries++;
			if(max_tries > model->_maxAttempts)
				return;
			
		}while(find(index1.begin(), index1.end(), new_symb) != index1.end());
		
		int done = 0;
		
		// Find the appropriate position for the new attribute.
		for(it = index1.begin(), it2 = index2.begin(); it < index1.end(); it++, it2++)
		{
			if(new_symb < (*it))
			{
				index1.insert(it, new_symb);
				index2.insert(it2, 0);
				//popHistory[level][new_symb]->increfs();
				done = 1;
				break;
			}
		}
		
		// If the new_symb > all symbionts indexed by host.
		if(!done)
		{
			//cout << "gene::mutation_type1_addSymb() => This symb is greater than any symb indexed by host." << endl;
			
			index1.push_back(new_symb);
			index2.push_back(0);
			//popHistory[level][new_symb]->increfs();
		}
		
#ifdef DEBUGGING
		cout << "gene::mutation_type1_addSymb() => new symb added is " << new_symb << endl;
#endif
		
		// No need to call pruneDuplicates() since duplicate symbionts were not allowed in this function.
		// No need to sort because it is taken care of in the current function. 
		// sort();
	}
	
	else
	{
		cout << "gene::mutation_type1_addSymb() => The ind is too large, hence nothing is added. Return the parent." << endl;
		return;
	}
	
#ifdef DEBUGGING
	cout << "gene::mutation_type1_addAttr() exit." << endl;
#endif
}

/***********************************************************************************************************/

void gene::mutation_level1_replaceSymb()
{
	// Mutation operator for individual level != 0
	// Replaces a symbiont with one which doesn't exist in the host.
	// Mutation probability has been applied in BreedGeneration function.
	
#ifdef DEBUGGING
	cout << "gene::mutation_type1_replaceSymb() enter." << endl;
#endif
	
#ifdef DEBUGGING
	cout << "gene::mutation_type1_replaceSymb() => The host to be mutated by replacing a symbiont: " << endl;
	print(0);
#endif
	
	if((int)index1.size() != (int)symbsToUse[level].size())
	{
		int gene_indx, gene_val;
		
		// If maximum number of clusters are used
		if((int)index1.size() == (int)symbsToUse[level].size())
			return;
		
		gene_indx = (int) (drand48() * index1.size());
		
#ifdef DEBUGGING
		cout << "gene::mutation_type1_replaceSymb() => Index of gene to mutate is " << gene_indx << endl;
#endif
		
		int max_tries = 0;
		
		do
		{
			gene_val = (int) (drand48() * (int)symbsToUse[level].size());
			
			max_tries++;
			if(max_tries > model->_maxAttempts)
				return;
			
		}while(find(index1.begin(), index1.end(), symbsToUse[level][gene_val]) != index1.end());
		
		//cout << "gene::mutation_type1_replaceSymb() => Value of gene " << gene_indx << " is changed from " << index1[gene_indx];
		//popHistory[level][index1[gene_indx]]->decrefs();
		
		index1[gene_indx] = symbsToUse[level][gene_val];
		
		//popHistory[level][index1[gene_indx]]->increfs();
		//cout << " to " << index1[gene_indx] << endl;
		
		// Sort the host after replacing on of its symbionts.
		sort();
	}
	
	else
	{
		cout << "gene::mutation_type1_removeSymb() => The ind is using all symbionts available, hence no replace is possible. Return the parent." << endl;
		return;
	}
	
#ifdef DEBUGGING
	cout << "gene::mutation_type1_replaceSymb() exit." << endl;
#endif
}

/***********************************************************************************************************/

